@@ -7,7 +7,8 @@
 set -Eeuo pipefail
 
 # Configuration:
-HACKAGE_INDEX_STATE="2023-10-05T11:38:51Z"
+# N.B. You may want to also update the index-state in hadrian/cabal.project.
+HACKAGE_INDEX_STATE="2024-10-30T22:56:00Z"
 MIN_HAPPY_VERSION="1.20"
 MIN_ALEX_VERSION="3.2.6"
 
@@ -403,6 +404,12 @@ function configure() {
     else
     args+=("--disable-numa")
   fi
+  if [[ -n ${HAPPY:-} ]]; then
+    args+=("HAPPY=$HAPPY")
+  fi
+  if [[ -n ${ALEX:-} ]]; then
+    args+=("ALEX=$ALEX")
+  fi
 
   start_section "configuring"
   # See https://stackoverflow.com/questions/7577052 for a rationale for the
@@ -411,8 +418,6 @@ function configure() {
     --enable-tarballs-autodownload \
     "${args[@]+"${args[@]}"}" \
     GHC="$GHC" \
-    HAPPY="$HAPPY" \
-    ALEX="$ALEX" \
     || ( cat config.log; fail "configure failed" )
   end_section "configuring"
 }
@@ -427,7 +432,9 @@ function push_perf_notes() {
     return
   fi
 
-  if [ -n "${CROSS_TARGET:-}" ] && [ "${CROSS_EMULATOR:-}" != "js-emulator" ]; then
+  # TODO: Remove this check, see #25299
+  # It is easy to forget to update this when testing a new cross platform
+  if [[ -n "${CROSS_TARGET:-}" ]] && [[ "${CROSS_TARGET:-}" != *"javascript"* ]] && [[ "${CROSS_TARGET:-}" != *"wasm"* ]]; then
     info "Can't test cross-compiled build."
     return
   fi
@@ -556,7 +563,7 @@ function install_bindist() {
 
       run ${CONFIGURE_WRAPPER:-} ./configure \
           --prefix="$instdir" \
-          "${args[@]+"${args[@]}"}"
+          "${args[@]+"${args[@]}"}" || fail "bindist configure failed"
       make_install_destdir "$TOP"/destdir "$instdir"
       # And check the `--info` of the installed compiler, sometimes useful in CI log.
       "$instdir"/bin/ghc --info
@@ -675,12 +682,25 @@ function test_hadrian() {
 }
 
 function summarise_hi_files() {
-  for iface in $(find . -type f -name "*.hi" | sort); do echo "$iface  $($HC --show-iface $iface | grep "  ABI hash:")"; done | tee $OUT/abis
-  for iface in $(find . -type f -name "*.hi" | sort); do echo "$iface  $($HC --show-iface $iface | grep "  interface hash:")"; done | tee $OUT/interfaces
-  for iface in $(find . -type f -name "*.hi" | sort); do
-      fname="$OUT/$(dirname $iface)"
-      mkdir -p $fname
-      $HC --show-iface $iface > "$OUT/$iface"
+  hi_files=$(find . -type f -name "*.hi" | sort)
+  for iface in $hi_files; do echo "$iface  $($HC --show-iface "$iface" | grep "  ABI hash:")"; done | tee $OUT/abis
+  for iface in $hi_files; do echo "$iface  $($HC --show-iface "$iface" | grep "  interface hash:")"; done | tee $OUT/interfaces
+  for iface in $hi_files; do
+      fname="$OUT/$(dirname "$iface")"
+      mkdir -p "$fname"
+      $HC --show-iface "$iface" > "$OUT/$iface"
+  done
+}
+
+function summarise_o_files() {
+  OBJDUMP=$(if test "$(uname)" == "Darwin"; then echo "objdump -m"; else echo "objdump"; fi)
+  o_files=$(find . -type f -name "*.o" | sort)
+  for o in $o_files; do
+      fname="$OUT/objs/$(dirname "$o")"
+      mkdir -p "$fname"
+      # To later compare object dumps except for the first line which prints the file path
+      $OBJDUMP --all-headers "$o" | tail -n+2 > "$OUT/objs/$o.all-headers"
+      $OBJDUMP --disassemble-all "$o" | tail -n+2 > "$OUT/objs/$o.disassemble-all"
   done
 }
 
@@ -695,11 +715,13 @@ function cabal_abi_test() {
 
   start_section "Cabal test: $OUT"
   mkdir -p "$OUT"
-  run "$HC" \
+  "$HC" \
     -hidir tmp -odir tmp -fforce-recomp -haddock \
     -iCabal/Cabal/src -XNoPolyKinds Distribution.Simple -j"$cores" \
-    "$@" 2>&1 | tee $OUT/log
+    -fobject-determinism \
+    "$@" 2>&1 | sed '1d' | tee $OUT/log
   summarise_hi_files
+  summarise_o_files
   popd
   end_section "Cabal test: $OUT"
 }
@@ -744,12 +766,40 @@ function check_interfaces(){
      for line in $(echo "$difference" | tr ' ' '\n' | grep ".hi" | sort | uniq); do
        diff "$1/$line" "$2/$line"
      done
-     fail "$3"
+     fail "$4"
   fi
 }
 
+function check_objects(){
+
+  # Big fast check
+  if diff -r "$1" "$2"
+  then
+    echo "Objects are the same"
+  else
+    echo "--------------------------------------------------------------------------------"
+    echo "Comparing all objects (1. headers, 2. disassembly). Stopping at first failure..."
+    echo "--------------------------------------------------------------------------------"
+
+    pushd "$1" >/dev/null
+    OBJ_DUMPS=$(find . -type f -name "*.all-headers" -or -name "*.disassemble-all")
+    popd >/dev/null
+
+    for dump in $OBJ_DUMPS
+    do
+      if diff "$1/$dump" "$2/$dump"
+      then
+        fail "Mismatched object: $dump"
+      fi
+    done
+
+    fail "Some objects are mismatched, but theres no diff with --all-headers or --disassemble-all. Perhaps try objdump -s"
+  fi
+
+}
+
 function abi_test() {
-  for i in {1..20}; do info "iteration $i"; run_abi_test; done
+  for i in {1..10}; do info "iteration $i"; run_abi_test; done
 }
 
 function run_abi_test() {
@@ -757,10 +807,12 @@ function run_abi_test() {
     fail "HC not set"
   fi
   mkdir -p out
-  OUT="$PWD/out/run1" DIR=$(mktemp -d XXXX-looooooooong) cabal_abi_test -O0
-  OUT="$PWD/out/run2" DIR=$(mktemp -d XXXX-short) cabal_abi_test -O0
+  OUT="$PWD/out/run1" DIR=$(mktemp -d XXXX-looooooooong) cabal_abi_test -O1 -haddock
+  # Count uniques in reverse one of the runs to get more non-determinism exposed
+  OUT="$PWD/out/run2" DIR=$(mktemp -d XXXX-short) cabal_abi_test -O1 -haddock -dunique-increment=-1 -dinitial-unique=16777215
   check_interfaces out/run1 out/run2 abis "Mismatched ABI hash"
   check_interfaces out/run1 out/run2 interfaces "Mismatched interface hashes"
+  check_objects out/run1 out/run2
 }
 
 function save_test_output() {
