@@ -65,7 +65,7 @@ eprint(f"Supported platforms: {job_mapping.keys()}")
 class Artifact(NamedTuple):
     job_name: str
     download_name: str
-    output_name: Optional[str]
+    output_name: str
     subdir: str
     anchor_name: str
 
@@ -96,17 +96,20 @@ def darwin(arch):
 windowsArtifact = PlatformSpec ( 'x86_64-windows'
                                , 'ghc-{version}-x86_64-unknown-mingw32' )
 
+def centos(n, arch='x86_64'):
+    return linux_platform(arch, "{arch}-linux-centos{n}".format(n=n,arch=arch))
+
 def fedora(n, arch='x86_64'):
-    return linux_platform(arch, "x86_64-linux-fedora{n}".format(n=n))
+    return linux_platform(arch, "{arch}-linux-fedora{n}".format(n=n,arch=arch))
 
 def alpine(n, arch='x86_64'):
-    return linux_platform(arch, "x86_64-linux-alpine{n}".format(n=n))
+    return linux_platform(arch, "{arch}-linux-alpine{n}".format(n=n,arch=arch))
 
 def rocky(n, arch='x86_64'):
-    return linux_platform(arch, "x86_64-linux-rocky{n}".format(n=n))
+    return linux_platform(arch, "{arch}-linux-rocky{n}".format(n=n,arch=arch))
 
 def ubuntu(n, arch='x86_64'):
-    return linux_platform(arch, "x86_64-linux-ubuntu{n}".format(n=n))
+    return linux_platform(arch, "{arch}-linux-ubuntu{n}".format(n=n,arch=arch))
 
 def linux_platform(arch, opsys):
     return PlatformSpec( opsys, 'ghc-{version}-{arch}-unknown-linux'.format(version="{version}", arch=arch) )
@@ -165,6 +168,11 @@ def mk_one_metadata(release_mode, version, job_map, artifact):
           , "dlSubdir": artifact.subdir.format(version=version)
           , "dlHash" : h }
 
+    # Only add dlOutput if it is inconsistent with the filename inferred from the URL
+    output = artifact.output_name.format(version=version)
+    if Path(urlparse(final_url).path).name != output:
+        res["dlOutput"] = output
+
     eprint(res)
 
     # add the uri to the anchor name cache so we can lookup an anchor for this uri
@@ -173,7 +181,7 @@ def mk_one_metadata(release_mode, version, job_map, artifact):
 
 # Turns a platform into an Artifact respecting pipeline_type
 # Looks up the right job to use from the .gitlab/jobs-metadata.json file
-def mk_from_platform(release_mode, pipeline_type, platform):
+def mk_from_platform(pipeline_type, platform):
     info = job_mapping[platform.name][pipeline_type]
     eprint(f"From {platform.name} / {pipeline_type} selecting {info['name']}")
     return Artifact(info['name']
@@ -182,30 +190,34 @@ def mk_from_platform(release_mode, pipeline_type, platform):
                    , platform.subdir
                    , f"ghc{{version}}-{platform.name}")
 
+
 # Generate the new metadata for a specific GHC mode etc
-def mk_new_yaml(release_mode, version, date, pipeline_type, job_map) -> object:
+def mk_new_yaml(release_mode, version, date, pipeline_type, job_map):
     def mk(platform):
         eprint("\n=== " + platform.name + " " + ('=' * (75 - len(platform.name))))
-        return mk_one_metadata(release_mode, version, job_map, mk_from_platform(release_mode, pipeline_type, platform))
+        return mk_one_metadata(release_mode, version, job_map, mk_from_platform(pipeline_type, platform))
 
     # Here are all the bindists we can distribute
     ubuntu1804 = mk(ubuntu("18_04"))
     ubuntu2004 = mk(ubuntu("20_04"))
+    ubuntu2204 = mk(ubuntu("22_04"))
     rocky8 = mk(rocky("8"))
+    centos7 = mk(centos(7))
     fedora33 = mk(fedora(33))
     darwin_x86 = mk(darwin("x86_64"))
     darwin_arm64 = mk(darwin("aarch64"))
     windows = mk(windowsArtifact)
     alpine3_12 = mk(alpine("3_12"))
-    alpine3_18 = mk(alpine("3_18"))
+    alpine3_20 = mk(alpine("3_20"))
     alpine3_18_arm64 = mk(alpine("3_18", arch='aarch64'))
     deb9 = mk(debian(9, "x86_64"))
     deb10 = mk(debian(10, "x86_64"))
     deb11 = mk(debian(11, "x86_64"))
     deb12 = mk(debian(12, "x86_64"))
     deb10_arm64 = mk(debian(10, "aarch64"))
-    deb11_arm64 = mk(debian(11, "aarch64"))
+    deb12_arm64 = mk(debian(12, "aarch64"))
     deb10_i386 = mk(debian(10, "i386"))
+    deb12_i386 = mk(debian(12, "i386"))
 
     source = mk_one_metadata(release_mode, version, job_map, source_artifact)
     test = mk_one_metadata(release_mode, version, job_map, test_artifact)
@@ -216,28 +228,34 @@ def mk_new_yaml(release_mode, version, date, pipeline_type, job_map) -> object:
     a64 = { "Linux_Debian": { "< 10": deb9
                            , "( >= 10 && < 11 )": deb10
                            , "( >= 11 && < 12 )": deb11
-                           , ">= 11": deb12
+                           , ">= 12": deb12
                            , "unknown_versioning": deb11 }
           , "Linux_Ubuntu" : { "unknown_versioning": ubuntu2004
                              , "( >= 16 && < 18 )": deb9
-                             , "( >= 18 && < 19 )": ubuntu1804 }
+                             , "( >= 18 && < 19 )": ubuntu1804
+                             , "( >= 19 && < 21 )": ubuntu2004
+                             , "( >= 21 )": ubuntu2204
+                             }
           , "Linux_Mint"   : { "< 20": ubuntu1804
                              , ">= 20": ubuntu2004
                              , "unknown_versioning": ubuntu2004 }
-          , "Linux_Fedora"  : { ">= 33": fedora33 }
-          , "Linux_RedHat"  : { "< 9": rocky8
-                              , ">= 9": fedora33
-                              , "unknown_versioning": fedora33 }
+          , "Linux_CentOS"  : { "( >= 7 && < 8 )" : centos7
+                              , "unknown_versioning" : centos7  }
+          , "Linux_Fedora"  : { ">= 33": fedora33
+                              , "unknown_versioning": centos7 }
+          , "Linux_RedHat"  : { "unknown_versioning": fedora33 }
           , "Linux_UnknownLinux" : { "unknown_versioning": rocky8 }
           , "Darwin" : { "unknown_versioning" : darwin_x86 }
           , "Windows" : { "unknown_versioning" :  windows }
-          , "Linux_Alpine" : { "( >= 3.12 && < 3.18 )": alpine3_12
-                             , ">= 3.18": alpine3_18
+          , "Linux_Alpine" : { "( >= 3.12 && < 3.20 )": alpine3_12
+                             , ">= 3.20": alpine3_20
                              , "unknown_versioning": alpine3_12 }
 
           }
 
-    a32 = { "Linux_Debian": { "unknown_versioning": deb10_i386 }
+    a32 = { "Linux_Debian": { "( >= 10 && < 12 )": deb10_i386
+                            , ">= 12": deb12_i386
+                            , "unknown_versioning": deb10_i386 }
           , "Linux_Ubuntu": { "unknown_versioning": deb10_i386 }
           , "Linux_Mint" : { "unknown_versioning": deb10_i386 }
           , "Linux_UnknownLinux" : { "unknown_versioning": deb10_i386 }
@@ -245,8 +263,8 @@ def mk_new_yaml(release_mode, version, date, pipeline_type, job_map) -> object:
 
     arm64 = { "Linux_UnknownLinux": { "unknown_versioning": deb10_arm64 }
             , "Linux_Alpine" : { "unknown_versioning": alpine3_18_arm64 }
-            , "Linux_Debian": { "( >= 10 && < 11 )": deb10_arm64
-                              , "( >= 11 && < 12 )": deb11_arm64
+            , "Linux_Debian": { "( >= 10 && < 12 )": deb10_arm64
+                              , "( >= 12 )": deb12_arm64
                               , "unknown_versioning": deb10_arm64
                               }
             , "Darwin": { "unknown_versioning": darwin_arm64 }
