@@ -137,12 +137,9 @@ cpsTop logger platform cfg dus proc =
       dump Opt_D_dump_cmm_sp "Layout Stack" g
 
       ----------- Sink and inline assignments  --------------------------------
-      (g, dus) <- {-# SCC "sink" #-} -- See Note [Sinking after stack layout]
-           if cmmOptSink cfg
-              then pure $ runUniqueDSM dus $ cmmSink cfg g
-              else return (g, dus)
-      dump Opt_D_dump_cmm_sink "Sink assignments" g
-
+      g <- {-# SCC "sink" #-} -- See Note [Sinking after stack layout]
+           condPass (cmmOptSink cfg) (cmmSink platform) g
+                    Opt_D_dump_cmm_sink "Sink assignments"
 
       ------------- CAF analysis ----------------------------------------------
       let cafEnv = {-# SCC "cafAnal" #-} cafAnal platform call_pps l g
