@@ -3,9 +3,11 @@ module GHC.Cmm.Type
     , b8, b16, b32, b64, b128, b256, b512, f32, f64, bWord, bHalfWord, gcWord
     , cInt
     , cmmBits, cmmFloat
-    , typeWidth, cmmEqType, cmmEqType_ignoring_ptrhood
+    , typeWidth, setCmmTypeWidth
+    , cmmEqType, cmmCompatType
     , isFloatType, isGcPtrType, isBitsType
-    , isWord32, isWord64, isFloat64, isFloat32
+    , isWordAny, isWord32, isWord64
+    , isFloat64, isFloat32
 
     , Width(..)
     , widthInBits, widthInBytes, widthInLog, widthFromBytes
@@ -25,6 +27,8 @@ module GHC.Cmm.Type
     , cmmVec
     , vecLength, vecElemType
     , isVecType
+
+    , DoAlignSanitisation
    )
 where
 
@@ -32,7 +36,6 @@ where
 import GHC.Prelude
 
 import GHC.Platform
-import GHC.Data.FastString
 import GHC.Utils.Outputable
 import GHC.Utils.Panic
 
@@ -84,26 +87,35 @@ instance Outputable CmmCat where
 cmmEqType :: CmmType -> CmmType -> Bool -- Exact equality
 cmmEqType (CmmType c1 w1) (CmmType c2 w2) = c1==c2 && w1==w2
 
-cmmEqType_ignoring_ptrhood :: CmmType -> CmmType -> Bool
-  -- This equality is temporary; used in CmmLint
-  -- but the RTS files are not yet well-typed wrt pointers
-cmmEqType_ignoring_ptrhood (CmmType c1 w1) (CmmType c2 w2)
-   = c1 `weak_eq` c2 && w1==w2
+-- | A weaker notion of equality of 'CmmType's than 'cmmEqType',
+-- used (only) in Cmm Lint.
+--
+-- Why "weaker"? Because:
+--
+--  - we don't distinguish GcPtr vs NonGcPtr, because the the RTS files
+--    are not yet well-typed wrt pointers,
+--  - for vectors, we only compare the widths, because in practice things like
+--    X86 xmm registers support different types of data (e.g. 4xf32, 2xf64, 2xu64 etc).
+cmmCompatType :: CmmType -> CmmType -> Bool
+cmmCompatType (CmmType c1 w1) (CmmType c2 w2)
+   = c1 `weak_eq` c2 && w1 == w2
    where
      weak_eq :: CmmCat -> CmmCat -> Bool
-     FloatCat         `weak_eq` FloatCat         = True
-     FloatCat         `weak_eq` _other           = False
-     _other           `weak_eq` FloatCat         = False
-     (VecCat l1 cat1) `weak_eq` (VecCat l2 cat2) = l1 == l2
-                                                   && cat1 `weak_eq` cat2
-     (VecCat {})      `weak_eq` _other           = False
-     _other           `weak_eq` (VecCat {})      = False
-     _word1           `weak_eq` _word2           = True        -- Ignores GcPtr
+     FloatCat    `weak_eq` FloatCat    = True
+     FloatCat    `weak_eq` _other      = False
+     _other      `weak_eq` FloatCat    = False
+     (VecCat {}) `weak_eq` (VecCat {}) = True  -- only compare overall width
+     (VecCat {}) `weak_eq` _other      = False
+     _other      `weak_eq` (VecCat {}) = False
+     _word1      `weak_eq` _word2      = True  -- Ignores GcPtr
 
 --- Simple operations on CmmType -----
 typeWidth :: CmmType -> Width
 typeWidth (CmmType _ w) = w
 
+setCmmTypeWidth :: Width -> CmmType -> CmmType
+setCmmTypeWidth w (CmmType c _) = CmmType c w
+
 cmmBits, cmmFloat :: Width -> CmmType
 cmmBits  = CmmType BitsCat
 cmmFloat = CmmType FloatCat
@@ -145,10 +157,15 @@ isGcPtrType _other               = False
 isBitsType (CmmType BitsCat _) = True
 isBitsType _                   = False
 
-isWord32, isWord64, isFloat32, isFloat64 :: CmmType -> Bool
+isWordAny, isWord32, isWord64,
+  isFloat32, isFloat64 :: CmmType -> Bool
 -- isWord64 is true of 64-bit non-floats (both gc-ptrs and otherwise)
 -- isFloat32 and 64 are obvious
 
+isWordAny (CmmType BitsCat  _) = True
+isWordAny (CmmType GcPtrCat _) = True
+isWordAny _other               = False
+
 isWord64 (CmmType BitsCat  W64) = True
 isWord64 (CmmType GcPtrCat W64) = True
 isWord64 _other                 = False
@@ -167,18 +184,18 @@ isFloat64 _other                 = False
 --              Width
 -----------------------------------------------------------------------------
 
-data Width   = W8 | W16 | W32 | W64
-             | W128
-             | W256
-             | W512
-             deriving (Eq, Ord, Show)
+data Width
+  = W8
+  | W16
+  | W32
+  | W64
+  | W128
+  | W256
+  | W512
+  deriving (Eq, Ord, Show)
 
 instance Outputable Width where
-   ppr rep = ptext (mrStr rep)
-
-mrStr :: Width -> PtrString
-mrStr = sLit . show
-
+   ppr rep = text (show rep)
 
 -------- Common Widths  ------------
 
@@ -357,6 +374,14 @@ data ForeignHint
         -- Used to give extra per-argument or per-result
         -- information needed by foreign calling conventions
 
+instance Outputable ForeignHint where
+  ppr NoHint     = empty
+  ppr SignedHint = quotes(text "signed")
+--  ppr AddrHint   = quotes(text "address")
+-- Temp Jan08
+  ppr AddrHint   = (text "PtrHint")
+
+
 -------------------------------------------------------------------------
 
 -- These don't really belong here, but I don't know where is best to
@@ -466,3 +491,6 @@ C calling convention rather early on in the compiler).  However, given
 this, the cons outweigh the pros.
 
 -}
+
+-- | is @-falignment-sanitisation@ enabled?
+type DoAlignSanitisation = Bool
