@@ -60,7 +60,7 @@ import GHC.CmmToAsm.Config
 import GHC.CmmToAsm.Types
 
 
-import GHC.Cmm.Dataflow.Collections
+import GHC.Cmm.Dataflow.Label
 import GHC.Cmm
 import GHC.Cmm.CLabel
 import GHC.Cmm.Utils (cmmLoadBWord)
@@ -132,11 +132,21 @@ cmmMakeDynamicReference config referenceKind lbl
               addImport symbolPtr
               return $ cmmMakePicReference config symbolPtr
 
+        AccessViaSymbolPtr | ArchRISCV64 <- platformArch platform -> do
+              let symbolPtr = mkDynamicLinkerLabel SymbolPtr lbl
+              addImport symbolPtr
+              return $ cmmMakePicReference config symbolPtr
+
         AccessViaSymbolPtr -> do
               let symbolPtr = mkDynamicLinkerLabel SymbolPtr lbl
               addImport symbolPtr
               return $ cmmLoadBWord platform (cmmMakePicReference config symbolPtr)
 
+        -- On wasm, always preserve the original CLabel, the backends
+        -- will handle dynamic references properly
+        AccessDirectly | ArchWasm32 <- platformArch platform ->
+              pure $ CmmLit $ CmmLabel lbl
+
         AccessDirectly -> case referenceKind of
                 -- for data, we might have to make some calculations:
               DataReference -> return $ cmmMakePicReference config lbl
@@ -164,9 +174,13 @@ cmmMakePicReference config lbl
   | ArchAArch64 <- platformArch platform
   = CmmLit $ CmmLabel lbl
 
+  -- as on AArch64, there's no pic base register.
+  | ArchRISCV64 <- platformArch platform
+  = CmmLit $ CmmLabel lbl
+
   | OSAIX <- platformOS platform
   = CmmMachOp (MO_Add W32)
-          [ CmmReg (CmmGlobal PicBaseReg)
+          [ CmmReg (CmmGlobal $ GlobalRegUse PicBaseReg (bWord platform))
           , CmmLit $ picRelative (wordWidth platform)
                           (platformArch platform)
                           (platformOS   platform)
@@ -175,7 +189,7 @@ cmmMakePicReference config lbl
   -- both ABI versions default to medium code model
   | ArchPPC_64 _ <- platformArch platform
   = CmmMachOp (MO_Add W32) -- code model medium
-          [ CmmReg (CmmGlobal PicBaseReg)
+          [ CmmReg (CmmGlobal $ GlobalRegUse PicBaseReg (bWord platform))
           , CmmLit $ picRelative (wordWidth platform)
                           (platformArch platform)
                           (platformOS   platform)
@@ -184,7 +198,7 @@ cmmMakePicReference config lbl
   | (ncgPIC config || ncgExternalDynamicRefs config)
       && absoluteLabel lbl
   = CmmMachOp (MO_Add (wordWidth platform))
-          [ CmmReg (CmmGlobal PicBaseReg)
+          [ CmmReg (CmmGlobal $ GlobalRegUse PicBaseReg (bWord platform))
           , CmmLit $ picRelative (wordWidth platform)
                           (platformArch platform)
                           (platformOS   platform)
@@ -210,6 +224,13 @@ absoluteLabel lbl
 -- pointers, code stubs and GOT offsets look like is located in the
 -- module CLabel.
 
+-- | Helper to check whether the data resides in a DLL or not, see @labelDynamic@
+ncgLabelDynamic :: NCGConfig -> CLabel -> Bool
+ncgLabelDynamic config = labelDynamic (ncgThisModule config)
+                                      (ncgPlatform config)
+                                      (ncgExternalDynamicRefs config)
+
+
 -- We have to decide which labels need to be accessed
 -- indirectly or via a piece of stub code.
 data LabelAccessStyle
@@ -248,7 +269,7 @@ howToAccessLabel config _arch OSMinGW32 _kind lbl
 
         -- If the target symbol is in another PE we need to access it via the
         --      appropriate __imp_SYMBOL pointer.
-        | labelDynamic config lbl
+        | ncgLabelDynamic config lbl
         = AccessViaSymbolPtr
 
         -- Target symbol is in the same PE as the caller, so just access it directly.
@@ -263,7 +284,7 @@ howToAccessLabel config ArchAArch64 _os _kind lbl
         | not (ncgExternalDynamicRefs config)
         = AccessDirectly
 
-        | labelDynamic config lbl
+        | ncgLabelDynamic config lbl
         = AccessViaSymbolPtr
 
         | otherwise
@@ -280,44 +301,31 @@ howToAccessLabel config ArchAArch64 _os _kind lbl
 --
 howToAccessLabel config arch OSDarwin DataReference lbl
         -- data access to a dynamic library goes via a symbol pointer
-        | labelDynamic config lbl
+        | ncgLabelDynamic config lbl
         = AccessViaSymbolPtr
 
         -- when generating PIC code, all cross-module data references must
         -- must go via a symbol pointer, too, because the assembler
         -- cannot generate code for a label difference where one
-        -- label is undefined. Doesn't apply t x86_64.
-        -- Unfortunately, we don't know whether it's cross-module,
-        -- so we do it for all externally visible labels.
-        -- This is a slight waste of time and space, but otherwise
-        -- we'd need to pass the current Module all the way in to
-        -- this function.
+        -- label is undefined. Doesn't apply to x86_64 (why?).
         | arch /= ArchX86_64
-        , ncgPIC config && externallyVisibleCLabel lbl
+        , not (isLocalCLabel (ncgThisModule config) lbl)
+        , ncgPIC config
+        , externallyVisibleCLabel lbl
         = AccessViaSymbolPtr
 
         | otherwise
         = AccessDirectly
 
-howToAccessLabel config arch OSDarwin JumpReference lbl
+howToAccessLabel config _ OSDarwin JumpReference lbl
         -- dyld code stubs don't work for tailcalls because the
         -- stack alignment is only right for regular calls.
         -- Therefore, we have to go via a symbol pointer:
-        | arch == ArchX86 || arch == ArchX86_64 || arch == ArchAArch64
-        , labelDynamic config lbl
+        | ncgLabelDynamic config lbl
         = AccessViaSymbolPtr
 
 
-howToAccessLabel config arch OSDarwin _kind lbl
-        -- Code stubs are the usual method of choice for imported code;
-        -- not needed on x86_64 because Apple's new linker, ld64, generates
-        -- them automatically, neither on Aarch64 (arm64).
-        | arch /= ArchX86_64
-        , arch /= ArchAArch64
-        , labelDynamic config lbl
-        = AccessViaStub
-
-        | otherwise
+howToAccessLabel _ _ OSDarwin _ _
         = AccessDirectly
 
 ----------------------------------------------------------------------------
@@ -366,7 +374,7 @@ howToAccessLabel config arch os DataReference lbl
         | osElfTarget os
         = case () of
             -- A dynamic label needs to be accessed via a symbol pointer.
-          _ | labelDynamic config lbl
+          _ | ncgLabelDynamic config lbl
             -> AccessViaSymbolPtr
 
             -- For PowerPC32 -fPIC, we have to access even static data
@@ -394,21 +402,27 @@ howToAccessLabel config arch os DataReference lbl
 
 howToAccessLabel config arch os CallReference lbl
         | osElfTarget os
-        , labelDynamic config lbl && not (ncgPIC config)
+        , ncgLabelDynamic config lbl
+        , not (ncgPIC config)
         = AccessDirectly
 
         | osElfTarget os
         , arch /= ArchX86
-        , labelDynamic config lbl
+        , ncgLabelDynamic config lbl
         , ncgPIC config
         = AccessViaStub
 
 howToAccessLabel config _arch os _kind lbl
         | osElfTarget os
-        = if labelDynamic config lbl
+        = if ncgLabelDynamic config lbl
             then AccessViaSymbolPtr
             else AccessDirectly
 
+-- On wasm, always keep the original CLabel and let the backend decide
+-- how to handle dynamic references
+howToAccessLabel _ ArchWasm32 _ _ _
+        = AccessDirectly
+
 -- all other platforms
 howToAccessLabel config _arch _os _kind _lbl
         | not (ncgPIC config)
@@ -518,7 +532,7 @@ gotLabel
         -- HACK: this label isn't really foreign
         = mkForeignLabel
                 (fsLit ".LCTOC1")
-                Nothing ForeignLabelInThisPackage IsData
+                ForeignLabelInThisPackage IsData
 
 
 
@@ -527,23 +541,13 @@ gotLabel
 --
 -- We don't need to declare any offset tables.
 -- However, for PIC on x86, we need a small helper function.
-pprGotDeclaration :: NCGConfig -> SDoc
+pprGotDeclaration :: NCGConfig -> HDoc
 pprGotDeclaration config = case (arch,os) of
-   (ArchX86, OSDarwin)
-        | ncgPIC config
-        -> vcat [
-                text ".section __TEXT,__textcoal_nt,coalesced,no_toc",
-                text ".weak_definition ___i686.get_pc_thunk.ax",
-                text ".private_extern ___i686.get_pc_thunk.ax",
-                text "___i686.get_pc_thunk.ax:",
-                text "\tmovl (%esp), %eax",
-                text "\tret" ]
-
    (_, OSDarwin) -> empty
 
    -- Emit XCOFF TOC section
    (_, OSAIX)
-        -> vcat $ [ text ".toc"
+        -> lines_ [ text ".toc"
                   , text ".tc ghc_toc_table[TC],.LCTOC1"
                   , text ".csect ghc_toc_table[RW]"
                     -- See Note [.LCTOC1 in PPC PIC code]
@@ -553,16 +557,16 @@ pprGotDeclaration config = case (arch,os) of
 
    -- PPC 64 ELF v1 needs a Table Of Contents (TOC)
    (ArchPPC_64 ELF_V1, _)
-        -> text ".section \".toc\",\"aw\""
+        -> line $ text ".section \".toc\",\"aw\""
 
    -- In ELF v2 we also need to tell the assembler that we want ABI
    -- version 2. This would normally be done at the top of the file
    -- right after a file directive, but I could not figure out how
    -- to do that.
    (ArchPPC_64 ELF_V2, _)
-        -> vcat [ text ".abiversion 2",
-                  text ".section \".toc\",\"aw\""
-                ]
+        -> lines_ [ text ".abiversion 2",
+                    text ".section \".toc\",\"aw\""
+                  ]
 
    (arch, os)
         | osElfTarget os
@@ -572,7 +576,7 @@ pprGotDeclaration config = case (arch,os) of
 
         | osElfTarget os
         , arch /= ArchPPC_64 ELF_V1 && arch /= ArchPPC_64 ELF_V2
-        -> vcat [
+        -> lines_ [
                 -- See Note [.LCTOC1 in PPC PIC code]
                 text ".section \".got2\",\"aw\"",
                 text ".LCTOC1 = .+32768" ]
@@ -590,60 +594,8 @@ pprGotDeclaration config = case (arch,os) of
 -- and one for non-PIC.
 --
 
-pprImportedSymbol :: NCGConfig -> CLabel -> SDoc
+pprImportedSymbol :: NCGConfig -> CLabel -> HDoc
 pprImportedSymbol config importedLbl = case (arch,os) of
-   (ArchX86, OSDarwin)
-        | Just (CodeStub, lbl) <- dynamicLinkerLabelInfo importedLbl
-        -> if not pic
-             then
-              vcat [
-                  text ".symbol_stub",
-                  text "L" <> ppr_lbl lbl <> ptext (sLit "$stub:"),
-                      text "\t.indirect_symbol" <+> ppr_lbl lbl,
-                      text "\tjmp *L" <> ppr_lbl lbl
-                          <> text "$lazy_ptr",
-                  text "L" <> ppr_lbl lbl
-                      <> text "$stub_binder:",
-                      text "\tpushl $L" <> ppr_lbl lbl
-                          <> text "$lazy_ptr",
-                      text "\tjmp dyld_stub_binding_helper"
-              ]
-             else
-              vcat [
-                  text ".section __TEXT,__picsymbolstub2,"
-                      <> text "symbol_stubs,pure_instructions,25",
-                  text "L" <> ppr_lbl lbl <> ptext (sLit "$stub:"),
-                      text "\t.indirect_symbol" <+> ppr_lbl lbl,
-                      text "\tcall ___i686.get_pc_thunk.ax",
-                  text "1:",
-                      text "\tmovl L" <> ppr_lbl lbl
-                          <> text "$lazy_ptr-1b(%eax),%edx",
-                      text "\tjmp *%edx",
-                  text "L" <> ppr_lbl lbl
-                      <> text "$stub_binder:",
-                      text "\tlea L" <> ppr_lbl lbl
-                          <> text "$lazy_ptr-1b(%eax),%eax",
-                      text "\tpushl %eax",
-                      text "\tjmp dyld_stub_binding_helper"
-              ]
-           $+$ vcat [        text ".section __DATA, __la_sym_ptr"
-                    <> (if pic then int 2 else int 3)
-                    <> text ",lazy_symbol_pointers",
-                text "L" <> ppr_lbl lbl <> ptext (sLit "$lazy_ptr:"),
-                    text "\t.indirect_symbol" <+> ppr_lbl lbl,
-                    text "\t.long L" <> ppr_lbl lbl
-                    <> text "$stub_binder"]
-
-        | Just (SymbolPtr, lbl) <- dynamicLinkerLabelInfo importedLbl
-        -> vcat [
-                text ".non_lazy_symbol_pointer",
-                char 'L' <> ppr_lbl lbl <> text "$non_lazy_ptr:",
-                text "\t.indirect_symbol" <+> ppr_lbl lbl,
-                text "\t.long\t0"]
-
-        | otherwise
-        -> empty
-
    (ArchAArch64, OSDarwin)
         -> empty
 
@@ -662,7 +614,7 @@ pprImportedSymbol config importedLbl = case (arch,os) of
 
    (_, OSAIX) -> case dynamicLinkerLabelInfo importedLbl of
             Just (SymbolPtr, lbl)
-              -> vcat [
+              -> lines_ [
                    text "LC.." <> ppr_lbl lbl <> char ':',
                    text "\t.long" <+> ppr_lbl lbl ]
             _ -> empty
@@ -700,7 +652,7 @@ pprImportedSymbol config importedLbl = case (arch,os) of
         | osElfTarget os
         -> case dynamicLinkerLabelInfo importedLbl of
             Just (SymbolPtr, lbl)
-              -> vcat [
+              -> lines_ [
                    text ".LC_" <> ppr_lbl lbl <> char ':',
                    text "\t.quad" <+> ppr_lbl lbl ]
             _ -> empty
@@ -709,14 +661,14 @@ pprImportedSymbol config importedLbl = case (arch,os) of
      -> case dynamicLinkerLabelInfo importedLbl of
             Just (SymbolPtr, lbl)
               -> let symbolSize = case ncgWordWidth config of
-                         W32 -> sLit "\t.long"
-                         W64 -> sLit "\t.quad"
+                         W32 -> text "\t.long"
+                         W64 -> text "\t.quad"
                          _ -> panic "Unknown wordRep in pprImportedSymbol"
 
-                 in vcat [
+                 in lines_ [
                       text ".section \".got2\", \"aw\"",
                       text ".LC_" <> ppr_lbl lbl <> char ':',
-                      ptext symbolSize <+> ppr_lbl lbl ]
+                      symbolSize <+> ppr_lbl lbl ]
 
             -- PLT code stubs are generated automatically by the dynamic linker.
             _ -> empty
@@ -724,10 +676,10 @@ pprImportedSymbol config importedLbl = case (arch,os) of
    _ -> panic "PIC.pprImportedSymbol: no match"
  where
    platform = ncgPlatform config
-   ppr_lbl  = pprCLabel     platform AsmStyle
+   ppr_lbl :: CLabel -> HLine
+   ppr_lbl  = pprAsmLabel   platform
    arch     = platformArch  platform
    os       = platformOS    platform
-   pic      = ncgPIC config
 
 --------------------------------------------------------------------------------
 -- Generate code to calculate the address that should be put in the
@@ -833,11 +785,11 @@ initializePicBase_ppc _ _ _ _
 -- (See PprMach.hs)
 
 initializePicBase_x86
-        :: Arch -> OS -> Reg
+        :: OS -> Reg
         -> [NatCmmDecl (Alignment, RawCmmStatics) X86.Instr]
         -> NatM [NatCmmDecl (Alignment, RawCmmStatics) X86.Instr]
 
-initializePicBase_x86 ArchX86 os picReg
+initializePicBase_x86 os picReg
         (CmmProc info lab live (ListGraph blocks) : statics)
     | osElfTarget os
     = return (CmmProc info lab live (ListGraph blocks') : statics)
@@ -855,12 +807,12 @@ initializePicBase_x86 ArchX86 os picReg
           fetchGOT (BasicBlock bID insns) =
              BasicBlock bID (X86.FETCHGOT picReg : insns)
 
-initializePicBase_x86 ArchX86 OSDarwin picReg
+initializePicBase_x86 OSDarwin picReg
         (CmmProc info lab live (ListGraph (entry:blocks)) : statics)
         = return (CmmProc info lab live (ListGraph (block':blocks)) : statics)
 
     where BasicBlock bID insns = entry
           block' = BasicBlock bID (X86.FETCHPC picReg : insns)
 
-initializePicBase_x86 _ _ _ _
+initializePicBase_x86 _ _ _
         = panic "initializePicBase_x86: not needed"
