@@ -9,11 +9,10 @@ module GHC.Iface.Ext.Utils where
 import GHC.Prelude
 
 import GHC.Core.Map.Type
-import GHC.Driver.Session    ( DynFlags )
+import GHC.Driver.DynFlags    ( DynFlags )
 import GHC.Driver.Ppr
 import GHC.Data.FastString   ( FastString, mkFastString )
 import GHC.Iface.Type
-import GHC.Core.Multiplicity
 import GHC.Types.Name hiding (varName)
 import GHC.Types.Name.Set
 import GHC.Utils.Outputable hiding ( (<>) )
@@ -22,10 +21,12 @@ import GHC.Types.SrcLoc
 import GHC.CoreToIface
 import GHC.Core.TyCon
 import GHC.Core.TyCo.Rep
+import GHC.Core.TyCo.Compare( nonDetCmpType )
 import GHC.Core.Type
 import GHC.Types.Var
 import GHC.Types.Var.Env
 import GHC.Parser.Annotation
+import qualified GHC.Data.Strict as Strict
 
 import GHC.Iface.Ext.Types
 
@@ -39,7 +40,8 @@ import Data.Monoid
 import Data.List                  (find)
 import Data.Traversable           ( for )
 import Data.Coerce
-import Control.Monad.Trans.State.Strict hiding (get)
+import GHC.Utils.Monad.State.Strict hiding (get)
+import GHC.Utils.Panic.Plain( assert )
 import Control.Monad.Trans.Reader
 import qualified Data.Tree as Tree
 
@@ -60,7 +62,7 @@ renderHieType dflags ht = showSDoc dflags (ppr $ hieTypeToIface ht)
 
 resolveVisibility :: Type -> [Type] -> [(Bool,Type)]
 resolveVisibility kind ty_args
-  = go (mkEmptyTCvSubst in_scope) kind ty_args
+  = go (mkEmptySubst in_scope) kind ty_args
   where
     in_scope = mkInScopeSet (tyCoVarsOfTypes ty_args)
 
@@ -69,8 +71,8 @@ resolveVisibility kind ty_args
       | Just ty' <- coreView ty
       = go env ty' ts
     go env (ForAllTy (Bndr tv vis) res) (t:ts)
-      | isVisibleArgFlag vis = (True , t) : ts'
-      | otherwise            = (False, t) : ts'
+      | isVisibleForAllTyFlag vis = (True , t) : ts'
+      | otherwise                 = (False, t) : ts'
       where
         ts' = go (extendTvSubst env tv t) res ts
 
@@ -105,7 +107,13 @@ data EvidenceInfo a
   , evidenceSpan :: RealSrcSpan
   , evidenceType :: a
   , evidenceDetails :: Maybe (EvVarSource, Scope, Maybe Span)
-  } deriving (Eq,Ord,Functor)
+  } deriving (Eq, Functor)
+
+instance Ord a => Ord (EvidenceInfo a) where
+  compare (EvidenceInfo name span typ dets) (EvidenceInfo name' span' typ' dets') =
+    case stableNameCmp name name' of
+      EQ -> compare (span, typ, dets) (span', typ', dets')
+      r -> r
 
 instance (Outputable a) => Outputable (EvidenceInfo a) where
   ppr (EvidenceInfo name span typ dets) =
@@ -154,15 +162,15 @@ getEvidenceTree refmap var = go emptyNameSet var
 hieTypeToIface :: HieTypeFix -> IfaceType
 hieTypeToIface = foldType go
   where
-    go (HTyVarTy n) = IfaceTyVar $ occNameFS $ getOccName n
+    go (HTyVarTy n) = IfaceTyVar $ (mkIfLclName (occNameFS $ getOccName n))
     go (HAppTy a b) = IfaceAppTy a (hieToIfaceArgs b)
     go (HLitTy l) = IfaceLitTy l
-    go (HForAllTy ((n,k),af) t) = let b = (occNameFS $ getOccName n, k)
+    go (HForAllTy ((n,k),af) t) = let b = (mkIfLclName (occNameFS $ getOccName n), k)
                                   in IfaceForAllTy (Bndr (IfaceTvBndr b) af) t
-    go (HFunTy w a b)   = IfaceFunTy VisArg   w       a    b
-    go (HQualTy pred b) = IfaceFunTy InvisArg many_ty pred b
+    go (HFunTy w a b)   = IfaceFunTy visArgTypeLike   w       a    b
+    go (HQualTy pred b) = IfaceFunTy invisArgTypeLike many_ty pred b
     go (HCastTy a) = a
-    go HCoercionTy = IfaceTyVar "<coercion type>"
+    go HCoercionTy = IfaceTyVar (mkIfLclName "<coercion type>")
     go (HTyConApp a xs) = IfaceTyConApp a (hieToIfaceArgs xs)
 
     -- This isn't fully faithful - we can't produce the 'Inferred' case
@@ -186,7 +194,7 @@ initialHTS = HTS emptyTypeMap IM.empty 0
 freshTypeIndex :: State HieTypeState TypeIndex
 freshTypeIndex = do
   index <- gets freshIndex
-  modify' $ \hts -> hts { freshIndex = index+1 }
+  modify $ \hts -> hts { freshIndex = index+1 }
   return index
 
 compressTypes
@@ -216,7 +224,7 @@ getTypeIndex t
   where
     extendHTS t ht = do
       i <- freshTypeIndex
-      modify' $ \(HTS tm tt fi) ->
+      modify $ \(HTS tm tt fi) ->
         HTS (extendTypeMap tm t i) (IM.insert i ht tt) fi
       return i
 
@@ -239,9 +247,9 @@ getTypeIndex t
       ai <- getTypeIndex a
       bi <- getTypeIndex b
       wi <- getTypeIndex w
-      return $ case af of
-                 InvisArg -> case w of Many -> HQualTy ai bi; _ -> error "Unexpected non-unrestricted predicate"
-                 VisArg   -> HFunTy wi ai bi
+      return $ if isInvisibleFunArg af
+               then assert (isManyTy w) $ HQualTy ai bi
+               else                       HFunTy wi ai bi
     go (LitTy a) = return $ HLitTy $ toIfaceTyLit a
     go (CastTy t _) = do
       i <- getTypeIndex t
@@ -295,8 +303,9 @@ getNameBindingInClass
   -> Maybe Span
 getNameBindingInClass n sp asts = do
   ast <- M.lookup (HiePath (srcSpanFile sp)) asts
+  clsNode <- selectLargestContainedBy sp ast
   getFirst $ foldMap First $ do
-    child <- flattenAst ast
+    child <- flattenAst clsNode
     dets <- maybeToList
       $ M.lookup (Right n) $ sourcedNodeIdents $ sourcedNodeInfo child
     let binding = foldMap (First . getBindSiteFromContext) (identInfo dets)
@@ -524,21 +533,14 @@ locOnly (RealSrcSpan span _) = do
   pure [Node e span []]
 locOnly _ = pure []
 
-mkScopeA :: SrcSpanAnn' ann -> Scope
-mkScopeA l = mkScope (locA l)
+locOnlyE :: Monad m => EpaLocation -> ReaderT NodeOrigin m [HieAST a]
+locOnlyE (EpaSpan s) = locOnly s
+locOnlyE _ = pure []
 
-mkScope :: SrcSpan -> Scope
-mkScope (RealSrcSpan sp _) = LocalScope sp
-mkScope _ = NoScope
-
-mkLScope :: Located a -> Scope
-mkLScope = mkScope . getLoc
-
-mkLScopeA :: GenLocated (SrcSpanAnn' a) e -> Scope
-mkLScopeA = mkScope . locA . getLoc
-
-mkLScopeN :: LocatedN a -> Scope
-mkLScopeN = mkScope . getLocA
+mkScope :: (HasLoc a) => a -> Scope
+mkScope a = case getHasLoc a of
+              (RealSrcSpan sp _) -> LocalScope sp
+              _ -> NoScope
 
 combineScopes :: Scope -> Scope -> Scope
 combineScopes ModuleScope _ = ModuleScope
@@ -546,7 +548,7 @@ combineScopes _ ModuleScope = ModuleScope
 combineScopes NoScope x = x
 combineScopes x NoScope = x
 combineScopes (LocalScope a) (LocalScope b) =
-  mkScope $ combineSrcSpans (RealSrcSpan a Nothing) (RealSrcSpan b Nothing)
+  mkScope $ combineSrcSpans (RealSrcSpan a Strict.Nothing) (RealSrcSpan b Strict.Nothing)
 
 mkSourcedNodeInfo :: NodeOrigin -> NodeInfo a -> SourcedNodeInfo a
 mkSourcedNodeInfo org ni = SourcedNodeInfo $ M.singleton org ni
@@ -554,8 +556,8 @@ mkSourcedNodeInfo org ni = SourcedNodeInfo $ M.singleton org ni
 {-# INLINEABLE makeNodeA #-}
 makeNodeA
   :: (Monad m, Data a)
-  => a                       -- ^ helps fill in 'nodeAnnotations' (with 'Data')
-  -> SrcSpanAnn' ann         -- ^ return an empty list if this is unhelpful
+  => a                 -- ^ helps fill in 'nodeAnnotations' (with 'Data')
+  -> EpAnn ann         -- ^ return an empty list if this is unhelpful
   -> ReaderT NodeOrigin m [HieAST b]
 makeNodeA x spn = makeNode x (locA spn)
 
