@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP #-}
+
 
 -----------------------------------------------------------------------------
 --
@@ -15,13 +15,11 @@ module GHC.StgToCmm.DataCon (
         cgTopRhsCon, buildDynCon, bindConArgs
     ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
 import GHC.Platform
-import GHC.Platform.Profile
 
+import GHC.Stg.Utils (allowTopLevelConApp)
 import GHC.Stg.Syntax
 import GHC.Core  ( AltCon(..) )
 
@@ -40,7 +38,6 @@ import GHC.Runtime.Heap.Layout
 import GHC.Types.CostCentre
 import GHC.Unit
 import GHC.Core.DataCon
-import GHC.Driver.Session
 import GHC.Data.FastString
 import GHC.Types.Id
 import GHC.Types.Id.Info( CafInfo( NoCafRefs ) )
@@ -52,21 +49,23 @@ import GHC.Utils.Panic
 import GHC.Utils.Misc
 import GHC.Utils.Monad (mapMaybeM)
 
-import Control.Monad
 import Data.Char
+import GHC.StgToCmm.Config (stgToCmmPlatform)
+import GHC.StgToCmm.TagCheck (checkConArgsStatic, checkConArgsDyn)
+import GHC.Utils.Outputable
 
 ---------------------------------------------------------------
 --      Top-level constructors
 ---------------------------------------------------------------
 
-cgTopRhsCon :: DynFlags
+cgTopRhsCon :: StgToCmmConfig
             -> Id               -- Name of thing bound to this RHS
             -> DataCon          -- Id
             -> ConstructorNumber
             -> [NonVoid StgArg] -- Args
             -> (CgIdInfo, FCode ())
-cgTopRhsCon dflags id con mn args
-  | Just static_info <- precomputedStaticConInfo_maybe dflags id con args
+cgTopRhsCon cfg id con mn args
+  | Just static_info <- precomputedStaticConInfo_maybe cfg id con args
   , let static_code | isInternalName name = pure ()
                     | otherwise           = gen_code
   = -- There is a pre-allocated static closure available; use it
@@ -82,7 +81,7 @@ cgTopRhsCon dflags id con mn args
   = (id_Info, gen_code)
 
   where
-   platform      = targetPlatform dflags
+   platform      = stgToCmmPlatform cfg
    id_Info       = litIdInfo platform id (mkConLFInfo con) (CmmLabel closure_label)
    name          = idName id
    caffy         = idCafInfo id -- any stgArgHasCafRefs args
@@ -91,11 +90,9 @@ cgTopRhsCon dflags id con mn args
    gen_code =
      do { profile <- getProfile
         ; this_mod <- getModuleName
-        ; when (platformOS platform == OSMinGW32) $
-              -- Windows DLLs have a problem with static cross-DLL refs.
-              MASSERT( not (isDllConApp dflags this_mod con (map fromNonVoid args)) )
-        ; ASSERT( args `lengthIs` countConRepArgs con ) return ()
-
+        ; massert (allowTopLevelConApp platform (stgToCmmExtDynRefs cfg) this_mod con (map fromNonVoid args))
+        ; massert (args `lengthIs` countConRepArgs con )
+        ; checkConArgsStatic (text "TagCheck failed - Top level con") con (map fromNonVoid args)
         -- LAY IT OUT
         ; let
             (tot_wds, --  #ptr_wds + #nonptr_wds
@@ -167,18 +164,20 @@ buildDynCon :: Id                 -- Name of the thing to which this constr will
             -> FCode (CgIdInfo, FCode CmmAGraph)
                -- Return details about how to find it and initialization code
 buildDynCon binder mn actually_bound cc con args
-    = do dflags <- getDynFlags
-         buildDynCon' dflags binder mn actually_bound cc con args
+    = do cfg <- getStgToCmmConfig
+         --   pprTrace "noCodeLocal:" (ppr (binder,con,args,cgInfo)) True
+         case precomputedStaticConInfo_maybe cfg binder con args of
+           Just cgInfo -> return (cgInfo, return mkNop)
+           Nothing     -> buildDynCon' binder mn actually_bound cc con args
 
 
-buildDynCon' :: DynFlags
-             -> Id -> ConstructorNumber
+buildDynCon' :: Id
+             -> ConstructorNumber
              -> Bool
              -> CostCentreStack
              -> DataCon
              -> [NonVoid StgArg]
              -> FCode (CgIdInfo, FCode CmmAGraph)
-
 {- We used to pass a boolean indicating whether all the
 args were of size zero, so we could use a static
 constructor; but I concluded that it just isn't worth it.
@@ -189,14 +188,8 @@ The reason for having a separate argument, rather than looking at
 the addr modes of the args is that we may be in a "knot", and
 premature looking at the args will cause the compiler to black-hole!
 -}
-
-buildDynCon' dflags binder _ _ _cc con args
-  | Just cgInfo <- precomputedStaticConInfo_maybe dflags binder con args
-  -- , pprTrace "noCodeLocal:" (ppr (binder,con,args,cgInfo)) True
-  = return (cgInfo, return mkNop)
-
 -------- buildDynCon': the general case -----------
-buildDynCon' _ binder mn actually_bound ccs con args
+buildDynCon' binder mn actually_bound ccs con args
   = do  { (id_info, reg) <- rhsIdInfo binder lf_info
         ; return (id_info, gen_code reg)
         }
@@ -205,8 +198,9 @@ buildDynCon' _ binder mn actually_bound ccs con args
 
   gen_code reg
     = do  { modu <- getModuleName
-          ; profile <- getProfile
-          ; let platform = profilePlatform profile
+          ; cfg  <- getStgToCmmConfig
+          ; let platform = stgToCmmPlatform cfg
+                profile  = stgToCmmProfile  cfg
                 (tot_wds, ptr_wds, args_w_offsets)
                    = mkVirtConstrOffsets profile (addArgReps args)
                 nonptr_wds = tot_wds - ptr_wds
@@ -215,23 +209,27 @@ buildDynCon' _ binder mn actually_bound ccs con args
           ; let ticky_name | actually_bound = Just binder
                            | otherwise = Nothing
 
+          ; checkConArgsDyn (hang (text "TagCheck failed on constructor application.") 4 $
+                                   text "On binder:" <> ppr binder $$ text "Constructor:" <> ppr con) con (map fromNonVoid args)
           ; hp_plus_n <- allocDynClosure ticky_name info_tbl lf_info
-                                          use_cc blame_cc args_w_offsets
+                                          (use_cc platform) (blame_cc platform)
+                                          args_w_offsets
           ; return (mkRhsInit platform reg lf_info hp_plus_n) }
     where
-      use_cc      -- cost-centre to stick in the object
-        | isCurrentCCS ccs = cccsExpr
+      use_cc platform     -- cost-centre to stick in the object
+        | isCurrentCCS ccs = cccsExpr platform
         | otherwise        = panic "buildDynCon: non-current CCS not implemented"
 
       blame_cc = use_cc -- cost-centre on which to blame the alloc (same)
 
+
 {- Note [Precomputed static closures]
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
 For Char/Int closures there are some value closures
 built into the RTS. This is the case for all values in
 the range mINT_INTLIKE .. mAX_INTLIKE (or CHARLIKE).
-See Note [CHARLIKE and INTLIKE closures.] in the RTS code.
+See Note [CHARLIKE and INTLIKE closures] in the RTS code.
 
 Similarly zero-arity constructors have a closure
 in their defining Module we can use.
@@ -318,36 +316,36 @@ We don't support this optimization when compiling into Windows DLLs yet
 because they don't support cross package data references well.
 -}
 
--- (precomputedStaticConInfo_maybe dflags id con args)
+-- (precomputedStaticConInfo_maybe cfg id con args)
 --     returns (Just cg_id_info)
 -- if there is a precomputed static closure for (con args).
 -- In that case, cg_id_info addresses it.
 -- See Note [Precomputed static closures]
-precomputedStaticConInfo_maybe :: DynFlags -> Id -> DataCon -> [NonVoid StgArg] -> Maybe CgIdInfo
-precomputedStaticConInfo_maybe dflags binder con []
+precomputedStaticConInfo_maybe :: StgToCmmConfig -> Id -> DataCon -> [NonVoid StgArg] -> Maybe CgIdInfo
+precomputedStaticConInfo_maybe cfg binder con []
 -- Nullary constructors
   | isNullaryRepDataCon con
-  = Just $ litIdInfo (targetPlatform dflags) binder (mkConLFInfo con)
+  = Just $ litIdInfo (stgToCmmPlatform cfg) binder (mkConLFInfo con)
                 (CmmLabel (mkClosureLabel (dataConName con) NoCafRefs))
-precomputedStaticConInfo_maybe dflags binder con [arg]
+precomputedStaticConInfo_maybe cfg binder con [arg]
   -- Int/Char values with existing closures in the RTS
   | intClosure || charClosure
-  , platformOS platform /= OSMinGW32 || not (positionIndependent dflags)
+  , platformOS platform /= OSMinGW32 || not (stgToCmmPIE cfg || stgToCmmPIC cfg)
   , Just val <- getClosurePayload arg
   , inRange val
-  = let intlike_lbl   = mkCmmClosureLabel rtsUnitId (fsLit label)
+  = let intlike_lbl   = mkCmmClosureLabel rtsUnitId label
         val_int = fromIntegral val :: Int
-        offsetW = (val_int - (fromIntegral min_static_range)) * (fixedHdrSizeW profile + 1)
+        offsetW = (val_int - fromIntegral min_static_range) * (fixedHdrSizeW profile + 1)
                 -- INTLIKE/CHARLIKE closures consist of a header and one word payload
         static_amode = cmmLabelOffW platform intlike_lbl offsetW
     in Just $ litIdInfo platform binder (mkConLFInfo con) static_amode
   where
-    profile = targetProfile dflags
-    platform = profilePlatform profile
-    intClosure = maybeIntLikeCon con
+    profile     = stgToCmmProfile  cfg
+    platform    = stgToCmmPlatform cfg
+    intClosure  = maybeIntLikeCon  con
     charClosure = maybeCharLikeCon con
     getClosurePayload (NonVoid (StgLitArg (LitNumber LitNumInt val))) = Just val
-    getClosurePayload (NonVoid (StgLitArg (LitChar val))) = Just $ (fromIntegral . ord $ val)
+    getClosurePayload (NonVoid (StgLitArg (LitChar val))) = Just (fromIntegral . ord $ val)
     getClosurePayload _ = Nothing
     -- Avoid over/underflow by comparisons at type Integer!
     inRange :: Integer -> Bool
@@ -366,8 +364,8 @@ precomputedStaticConInfo_maybe dflags binder con [arg]
       | charClosure = fromIntegral (pc_MAX_CHARLIKE constants)
       | otherwise = panic "precomputedStaticConInfo_maybe: Unknown closure type"
     label
-      | intClosure = "stg_INTLIKE"
-      | charClosure =  "stg_CHARLIKE"
+      | intClosure = fsLit "stg_INTLIKE"
+      | charClosure = fsLit "stg_CHARLIKE"
       | otherwise = panic "precomputedStaticConInfo_maybe: Unknown closure type"
 
 precomputedStaticConInfo_maybe _ _ _ _ = Nothing
@@ -382,7 +380,7 @@ bindConArgs :: AltCon -> LocalReg -> [NonVoid Id] -> FCode [LocalReg]
 -- binders args, assuming that we have just returned from a 'case' which
 -- found a con
 bindConArgs (DataAlt con) base args
-  = ASSERT(not (isUnboxedTupleDataCon con))
+  = assert (not (isUnboxedTupleDataCon con)) $
     do profile <- getProfile
        platform <- getPlatform
        let (_, _, args_w_offsets) = mkVirtConstrOffsets profile (addIdReps args)
@@ -402,4 +400,4 @@ bindConArgs (DataAlt con) base args
        mapMaybeM bind_arg args_w_offsets
 
 bindConArgs _other_con _base args
-  = ASSERT( null args ) return []
+  = assert (null args ) return []
