@@ -1,7 +1,8 @@
-{-# LANGUAGE LambdaCase #-}
-{-# LANGUAGE TupleSections #-}
-{-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE LambdaCase    #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE TupleSections #-}
+{-# LANGUAGE ViewPatterns  #-}
 
 -----------------------------------------------------------------------------
 -- |
@@ -30,26 +31,29 @@ where
 
 import GHC.Prelude
 
-import GHC.JS.Unsat.Syntax
+import GHC.JS.JStg.Syntax
+import GHC.JS.JStg.Monad
 import GHC.JS.Transform
 import GHC.JS.Make
+import GHC.JS.Ident
 
 import GHC.StgToJS.Apply
 import GHC.StgToJS.Arg
 import GHC.StgToJS.Closure
+import GHC.StgToJS.DataCon
 import GHC.StgToJS.ExprCtx
 import GHC.StgToJS.FFI
 import GHC.StgToJS.Heap
-import GHC.StgToJS.Monad
-import GHC.StgToJS.DataCon
-import GHC.StgToJS.Types
+import GHC.StgToJS.Ids
 import GHC.StgToJS.Literal
+import GHC.StgToJS.Monad
 import GHC.StgToJS.Prim
 import GHC.StgToJS.Profiling
 import GHC.StgToJS.Regs
-import GHC.StgToJS.Utils
 import GHC.StgToJS.Stack
-import GHC.StgToJS.Ids
+import GHC.StgToJS.Symbols
+import GHC.StgToJS.Types
+import GHC.StgToJS.Utils
 
 import GHC.Types.CostCentre
 import GHC.Types.Tickish
@@ -57,30 +61,34 @@ import GHC.Types.Var.Set
 import GHC.Types.Id
 import GHC.Types.Unique.FM
 import GHC.Types.RepType
+import GHC.Types.Literal
 
 import GHC.Stg.Syntax
 import GHC.Stg.Utils
 
 import GHC.Builtin.PrimOps
+import GHC.Builtin.Names
 
-import GHC.Core
+import GHC.Core hiding (Var)
 import GHC.Core.TyCon
 import GHC.Core.DataCon
 import GHC.Core.Opt.Arity (isOneShotBndr)
 import GHC.Core.Type hiding (typeSize)
 
 import GHC.Utils.Misc
+import GHC.Utils.Encoding
 import GHC.Utils.Monad
 import GHC.Utils.Panic
 import GHC.Utils.Outputable (ppr, renderWithContext, defaultSDocContext)
 import qualified Control.Monad.Trans.State.Strict as State
 import GHC.Data.FastString
+import qualified GHC.Types.Unique.Map as UM
+
 import qualified GHC.Data.List.SetOps as ListSetOps
 
 import Data.Monoid
 import Data.Maybe
 import Data.Function
-import Data.Either
 import qualified Data.List as L
 import qualified Data.Set as S
 import qualified Data.Map as M
@@ -88,17 +96,17 @@ import Control.Monad
 import Control.Arrow ((&&&))
 
 -- | Evaluate an expression in the given expression context (continuation)
-genExpr :: HasDebugCallStack => ExprCtx -> CgStgExpr -> G (JStat, ExprResult)
+genExpr :: HasDebugCallStack => ExprCtx -> CgStgExpr -> G (JStgStat, ExprResult)
 genExpr ctx stg = case stg of
   StgApp f args -> genApp ctx f args
   StgLit l      -> do
     ls <- genLit l
     let r = assignToExprCtx ctx ls
-    pure (r,ExprInline Nothing)
+    pure (r,ExprInline)
   StgConApp con _n args _ -> do
     as <- concatMapM genArg args
     c <- genCon ctx con as
-    return (c, ExprInline (Just as))
+    return (c, ExprInline)
   StgOpApp (StgFCallOp f _) args t
     -> genForeignCall ctx f t (concatMap typex_expr $ ctxTarget ctx) args
   StgOpApp (StgPrimOp op) args t
@@ -128,23 +136,23 @@ genExpr ctx stg = case stg of
 genBind :: HasDebugCallStack
         => ExprCtx
         -> CgStgBinding
-        -> G (JStat, ExprCtx)
+        -> G (JStgStat, ExprCtx)
 genBind ctx bndr =
   case bndr of
     StgNonRec b r -> do
        j <- assign b r >>= \case
          Just ja -> return ja
          Nothing -> allocCls Nothing [(b,r)]
-       return (j, addEvalRhs ctx [(b,r)])
+       return (j, ctx)
     StgRec bs     -> do
        jas <- mapM (uncurry assign) bs -- fixme these might depend on parts initialized by allocCls
        let m = if null jas then Nothing else Just (mconcat $ catMaybes jas)
        j <- allocCls m . map snd . filter (isNothing . fst) $ zip jas bs
-       return (j, addEvalRhs ctx bs)
+       return (j, ctx)
    where
      ctx' = ctxClearLneFrame ctx
 
-     assign :: Id -> CgStgRhs -> G (Maybe JStat)
+     assign :: Id -> CgStgRhs -> G (Maybe JStgStat)
      assign b (StgRhsClosure _ _ccs {-[the_fv]-} _upd [] expr _typ)
        | let strip = snd . stripStgTicksTop (not . tickishIsCode)
        , StgCase (StgApp scrutinee []) _ (AlgAlt _) [GenStgAlt (DataAlt _) params sel_expr] <- strip expr
@@ -163,10 +171,10 @@ genBind ctx bndr =
            the_fvjs <- varsForId the_fv
            case (tgts, the_fvjs) of
              ([tgt], [the_fvj]) -> return $ Just
-               (tgt ||= ApplExpr (var ("h$c_sel_" <> mkFastString sel_tag)) [the_fvj])
+               (tgt ||= ApplExpr (global (hdCSelStr <> mkFastString sel_tag)) [the_fvj])
              _ -> panic "genBind.assign: invalid size"
      assign b (StgRhsClosure _ext _ccs _upd [] expr _typ)
-       | snd (isInlineExpr (ctxEvaluatedIds ctx) expr) = do
+       | isInlineExpr expr = do
            d   <- declVarsForId b
            tgt <- varsForId b
            let ctx' = ctx { ctxTarget = assocIdExprs b tgt }
@@ -175,16 +183,10 @@ genBind ctx bndr =
      assign _b StgRhsCon{} = return Nothing
      assign  b r           = genEntry ctx' b r >> return Nothing
 
-     addEvalRhs c [] = c
-     addEvalRhs c ((b,r):xs)
-       | StgRhsCon{} <- r                         = addEvalRhs (ctxAssertEvaluated b c) xs
-       | (StgRhsClosure _ _ ReEntrant _ _ _) <- r = addEvalRhs (ctxAssertEvaluated b c) xs
-       | otherwise                                = addEvalRhs c xs
-
 genBindLne :: HasDebugCallStack
            => ExprCtx
            -> CgStgBinding
-           -> G (JStat, ExprCtx)
+           -> G (JStgStat, ExprCtx)
 genBindLne ctx bndr = do
   -- compute live variables and the offsets where they will be stored in the
   -- stack
@@ -228,72 +230,74 @@ genEntryLne ctx i rhs@(StgRhsClosure _ext _cc update args body typ) =
         maybe (panic "genEntryLne: updatable binder not found in let-no-escape frame")
               ((payloadSize-) . fst)
               (L.find ((==i) . fst . snd) (zip [0..] vars))
-      bh | isUpdatable update =
-             jVar (\x -> mconcat
-              [ x |= ApplExpr (var "h$bh_lne") [Sub sp (toJExpr myOffset), toJExpr (payloadSize+1)]
-              , IfStat x (ReturnStat x) mempty
-              ])
-         | otherwise = mempty
-  lvs  <- popLneFrame True payloadSize ctx
-  body <- genBody ctx R1 args body typ
-  ei@(TxtI eii) <- identForEntryId i
+      mk_bh :: G JStgStat
+      mk_bh | isUpdatable update =
+              do x <- freshIdent
+                 return $ mconcat
+                   [ x ||= ApplExpr hdBlackHoleLNE [Sub sp (toJExpr myOffset), toJExpr (payloadSize+1)]
+                   , IfStat (Var x) (ReturnStat (Var x)) mempty
+                   ]
+            | otherwise = pure mempty
+  blk_hl <- mk_bh
+  locals <- popLneFrame True payloadSize ctx
+  body   <- genBody ctx R1 args body typ
+  ei@(identFS -> eii) <- identForEntryId i
   sr   <- genStaticRefsRhs rhs
-  let f = (bh <> lvs <> body)
-  emitClosureInfo $
-    ClosureInfo ei
-                (CIRegs 0 $ concatMap idVt args)
-                (eii <> ", " <> mkFastString (renderWithContext defaultSDocContext (ppr i)))
-                (fixedLayout . reverse $
-                    map (stackSlotType . fst) (ctxLneFrameVars ctx))
-                CIStackFrame
-                sr
-  emitToplevel (jFunction ei [] f)
+  let f = blk_hl <> locals <> body
+  emitClosureInfo $ ClosureInfo
+    { ciVar = ei
+    , ciRegs = CIRegs 0 $ concatMap idJSRep args
+    , ciName = eii <> ", " <> mkFastString (renderWithContext defaultSDocContext (ppr i))
+    , ciLayout = fixedLayout . reverse $ map (stackSlotType . fst) (ctxLneFrameVars ctx)
+    , ciType = CIStackFrame
+    , ciStatic = sr
+    }
+  emitToplevel (FuncStat ei [] f)
 genEntryLne ctx i (StgRhsCon cc con _mu _ticks args _typ) = resetSlots $ do
   let payloadSize = ctxLneFrameSize ctx
-  ei@(TxtI _eii) <- identForEntryId i
+  ei <- identForEntryId i
   -- di <- varForDataConWorker con
   ii <- freshIdent
   p  <- popLneFrame True payloadSize ctx
   args' <- concatMapM genArg args
   ac    <- allocCon ii con cc args'
-  emitToplevel (jFunction ei [] (mconcat [decl ii, p, ac, r1 |= toJExpr ii, returnStack]))
+  emitToplevel (FuncStat ei [] (mconcat [decl ii, p, ac, r1 |= toJExpr ii, returnStack]))
 
 -- | Generate the entry function for a local closure
 genEntry :: HasDebugCallStack => ExprCtx -> Id -> CgStgRhs -> G ()
 genEntry _ _i StgRhsCon {} = return ()
-genEntry ctx i rhs@(StgRhsClosure _ext cc {-_bi live-} upd_flag args body typ) = resetSlots $ do
-  let live = stgLneLiveExpr rhs -- error "fixme" -- probably find live vars in body
+genEntry ctx i rhs@(StgRhsClosure _ext cc upd_flag args body typ) = resetSlots $ do
+  let live = stgLneLiveExpr rhs
   ll    <- loadLiveFun live
   llv   <- verifyRuntimeReps live
   upd   <- genUpdFrame upd_flag i
+  let entryCtx = ctxSetTarget [] (ctxClearLneFrame ctx)
   body  <- genBody entryCtx R2 args body typ
-  ei@(TxtI eii) <- identForEntryId i
   et    <- genEntryType args
   setcc <- ifProfiling $
              if et == CIThunk
                then enterCostCentreThunk
                else enterCostCentreFun cc
   sr <- genStaticRefsRhs rhs
-  emitClosureInfo $ ClosureInfo ei
-                                (CIRegs 0 $ PtrV : concatMap idVt args)
-                                (eii <> ", " <> mkFastString (renderWithContext defaultSDocContext (ppr i)))
-                                (fixedLayout $ map (uTypeVt . idType) live)
-                                et
-                                sr
-  emitToplevel (jFunction ei [] (mconcat [ll, llv, upd, setcc, body]))
-  where
-    entryCtx = ctxSetTarget [] (ctxClearLneFrame ctx)
+
+  ei <- identForEntryId i
+  emitClosureInfo $ ClosureInfo
+    { ciVar = ei
+    , ciRegs = CIRegs 0 $ PtrV : concatMap idJSRep args
+    , ciName = identFS ei <> ", " <> mkFastString (renderWithContext defaultSDocContext (ppr i))
+    , ciLayout = fixedLayout $ map (unaryTypeJSRep . idType) live
+    , ciType = et
+    , ciStatic = sr
+    }
+  emitToplevel (FuncStat ei [] (mconcat [ll, llv, upd, setcc, body]))
 
 -- | Generate the entry function types for identifiers. Note that this only
--- returns either 'CIThunk' or 'CIFun'. Everything else (PAP Blackhole etc.) is
--- filtered as not a RuntimeRepKinded type.
+-- returns either 'CIThunk' or 'CIFun'.
 genEntryType :: HasDebugCallStack => [Id] -> G CIType
 genEntryType []   = return CIThunk
-genEntryType args0 = do
+genEntryType args = do
   args' <- mapM genIdArg args
   return $ CIFun (length args) (length $ concat args')
-  where
-    args = filter (not . isRuntimeRepKindedTy . idType) args0
 
 -- | Generate the body of an object
 genBody :: HasDebugCallStack
@@ -302,7 +306,7 @@ genBody :: HasDebugCallStack
          -> [Id]
          -> CgStgExpr
          -> Type
-         -> G JStat
+         -> G JStgStat
 genBody ctx startReg args e typ = do
   -- load arguments into local variables
   la <- do
@@ -349,7 +353,7 @@ genBody ctx startReg args e typ = do
 --
 -- Se we're left to use the applied arguments to peel the type (unwrapped) one
 -- arg at a time. But passed args are args after unarisation so we need to
--- unarise every argument type that we peel (using typePrimRepArgs) to get the
+-- unarise every argument type that we peel (using typePrimRep) to get the
 -- number of passed args consumed by each type arg.
 --
 -- In case of failure to determine the type, we default to LiftedRep as it's
@@ -364,7 +368,7 @@ resultSize ty = result
 
 -- | Ensure that the set of identifiers has valid 'RuntimeRep's. This function
 -- returns a no-op when 'csRuntimeAssert' in 'StgToJSConfig' is False.
-verifyRuntimeReps :: HasDebugCallStack => [Id] -> G JStat
+verifyRuntimeReps :: HasDebugCallStack => [Id] -> G JStgStat
 verifyRuntimeReps xs = do
   runtime_assert <- csRuntimeAssert <$> getSettings
   if not runtime_assert
@@ -373,7 +377,7 @@ verifyRuntimeReps xs = do
   where
     verifyRuntimeRep i = do
       i' <- varsForId i
-      pure $ go i' (idVt i)
+      pure $ go i' (idJSRep i)
     go js         (VoidV:vs) = go js vs
     go (j1:j2:js) (LongV:vs) = v "h$verify_rep_long" [j1,j2] <> go js vs
     go (j1:j2:js) (AddrV:vs) = v "h$verify_rep_addr" [j1,j2] <> go js vs
@@ -385,13 +389,13 @@ verifyRuntimeReps xs = do
     ver j DoubleV = v "h$verify_rep_double"  [j]
     ver j ArrV    = v "h$verify_rep_arr"     [j]
     ver _ _       = mempty
-    v f as = ApplStat (var f) as
+    v f as = ApplStat (global f) as
 
 -- | Given a set of 'Id's, bind each 'Id' to the appropriate data fields in N
 -- registers. This assumes these data fields have already been populated in the
 -- registers. For the empty, singleton, and binary case use register 1, for any
 -- more use as many registers as necessary.
-loadLiveFun :: [Id] -> G JStat
+loadLiveFun :: [Id] -> G JStgStat
 loadLiveFun l = do
    l' <- concat <$> mapM identsForId l
    case l' of
@@ -413,11 +417,11 @@ loadLiveFun l = do
                , l''
                ]
   where
-        loadLiveVar d n v = let ident = TxtI (dataFieldName n)
+        loadLiveVar d n v = let ident = name (dataFieldName n)
                             in  v ||= SelExpr d ident
 
 -- | Pop a let-no-escape frame off the stack
-popLneFrame :: Bool -> Int -> ExprCtx -> G JStat
+popLneFrame :: Bool -> Int -> ExprCtx -> G JStgStat
 popLneFrame inEntry size ctx = do
   -- calculate the new stack size
   let ctx' = ctxLneShrinkStack ctx size
@@ -433,7 +437,7 @@ popLneFrame inEntry size ctx = do
   popSkipI skip is
 
 -- | Generate an updated given an 'Id'
-genUpdFrame :: UpdateFlag -> Id -> G JStat
+genUpdFrame :: UpdateFlag -> Id -> G JStgStat
 genUpdFrame u i
   | isReEntrant u   = pure mempty
   | isOneShotBndr i = maybeBh
@@ -453,11 +457,11 @@ genUpdFrame u i
 -- Useful for making sure that the object is not accidentally entered multiple
 -- times
 --
-bhSingleEntry :: StgToJSConfig -> JStat
+bhSingleEntry :: StgToJSConfig -> JStgStat
 bhSingleEntry _settings = mconcat
-  [ r1 .^ closureEntry_  |= var "h$blackholeTrap"
-  , r1 .^ closureField1_ |= undefined_
-  , r1 .^ closureField2_ |= undefined_
+  [ closureInfo   r1 |= hdBlackHoleTrap
+  , closureField1 r1 |= undefined_
+  , closureField2 r1 |= undefined_
   ]
 
 genStaticRefsRhs :: CgStgRhs -> G CIStatic
@@ -477,7 +481,7 @@ genStaticRefs lv
     sv = liveStatic lv
 
     getStaticRef :: Id -> G (Maybe FastString)
-    getStaticRef = fmap (fmap itxt . listToMaybe) . identsForId
+    getStaticRef = fmap (fmap identFS . listToMaybe) . identsForId
 
 -- | Reorder the things we need to push to reuse existing stack values as much
 -- as possible True if already on the stack at that location
@@ -491,17 +495,18 @@ optimizeFree
                        -- -- Bool: True when the slot already contains a value
 optimizeFree offset ids = do
   -- this line goes wrong                               vvvvvvv
-  let -- ids' = concat $ map (\i -> map (i,) [1..varSize . uTypeVt . idType $ i]) ids
+  let -- ids' = concat $ map (\i -> map (i,) [1..varSize . unaryTypeJSRep . idType $ i]) ids
       idSize :: Id -> Int
-      idSize i = sum $ map varSize (typeVt . idType $ i)
+      idSize i = typeSize $ idType i
       ids' = concatMap (\i -> map (i,) [1..idSize i]) ids
-      -- 1..varSize] . uTypeVt . idType $ i]) (typeVt ids)
+      -- 1..varSize] . unaryTypeJSRep . idType $ i]) (typeJSRep ids)
       l    = length ids'
   slots <- drop offset . take l . (++repeat SlotUnknown) <$> getSlots
   let slm                = M.fromList (zip slots [0..])
-      (remaining, fixed) = partitionEithers $
-         map (\inp@(i,n) -> maybe (Left inp) (\j -> Right (i,n,j,True))
-            (M.lookup (SlotId i n) slm)) ids'
+      (remaining, fixed) = partitionWith (\inp@(i,n) -> maybe (Left inp)
+                                                              (\j -> Right (i,n,j,True))
+                                                              (M.lookup (SlotId i n) slm))
+                                         ids'
       takenSlots         = S.fromList (fmap (\(_,_,x,_) -> x) fixed)
       freeSlots          = filter (`S.notMember` takenSlots) [0..l-1]
       remaining'         = zipWith (\(i,n) j -> (i,n,j,False)) remaining freeSlots
@@ -509,15 +514,15 @@ optimizeFree offset ids = do
   return $ map (\(i,n,_,b) -> (i,n,b)) allSlots
 
 -- | Allocate local closures
-allocCls :: Maybe JStat -> [(Id, CgStgRhs)] -> G JStat
+allocCls :: Maybe JStgStat -> [(Id, CgStgRhs)] -> G JStgStat
 allocCls dynMiddle xs = do
-   (stat, dyn) <- partitionEithers <$> mapM toCl xs
-   ac <- allocDynAll True dynMiddle dyn
+   (stat, dyn) <- partitionWithM toCl xs
+   ac <- allocDynAll False dynMiddle dyn
    pure (mconcat stat <> ac)
   where
     -- left = static, right = dynamic
     toCl :: (Id, CgStgRhs)
-         -> G (Either JStat (Ident,JExpr,[JExpr],CostCentreStack))
+         -> G (Either JStgStat (Ident,JStgExpr,[JStgExpr],CostCentreStack))
     -- statics
     {- making zero-arg constructors static is problematic, see #646
        proper candidates for this optimization should have been floated
@@ -554,20 +559,47 @@ genCase :: HasDebugCallStack
         -> AltType
         -> [CgStgAlt]
         -> LiveVars
-        -> G (JStat, ExprResult)
+        -> G (JStgStat, ExprResult)
 genCase ctx bnd e at alts l
-  | snd (isInlineExpr (ctxEvaluatedIds ctx) e) = do
+  -- For:      unpackCStringAppend# "some string"# str
+  -- Generate: h$appendToHsStringA(str, "some string")
+  --
+  -- The latter has a faster decoding loop.
+  --
+  -- Since #23270 and 7e0c8b3bab30, literals strings aren't STG atoms and we
+  -- need to match the following instead:
+  --
+  --    case "some string"# of b {
+  --      DEFAULT -> unpackCStringAppend# b str
+  --    }
+  --
+  -- Wrinkle: it doesn't kick in when literals are floated out to the top level.
+  --
+  | StgLit (LitString bs) <- e
+  , [GenStgAlt DEFAULT _ rhs] <- alts
+  , StgApp i args <- rhs
+  , getUnique i == unpackCStringAppendIdKey
+  , [StgVarArg b',x] <- args
+  , bnd == b'
+  , d <- utf8DecodeByteString bs
+  , [top] <- concatMap typex_expr (ctxTarget ctx)
+  = do
+      prof <- csProf <$> getSettings
+      let profArg = if prof then [jCafCCS] else []
+      a <- genArg x
+      return ( top |= app "h$appendToHsStringA" (toJExpr d : a ++ profArg)
+             , ExprInline
+             )
+
+  | isInlineExpr e = do
       bndi <- identsForId bnd
       let ctx' = ctxSetTop bnd
                   $ ctxSetTarget (assocIdExprs bnd (map toJExpr bndi))
                   $ ctx
       (ej, r) <- genExpr ctx' e
-      let d = case r of
-                ExprInline d0 -> d0
-                ExprCont -> pprPanic "genCase: expression was not inline"
-                                     (pprStgExpr panicStgPprOpts e)
+      massert (r == ExprInline)
 
-      (aj, ar) <- genAlts (ctxAssertEvaluated bnd ctx) bnd at d alts
+      (aj, ar) <- genAlts ctx bnd at alts
       (saveCCS,restoreCCS) <- ifProfilingM $ do
         ccsVar <- freshIdent
         pure ( ccsVar ||= toJExpr jCurrentCCS
@@ -583,7 +615,7 @@ genCase ctx bnd e at alts l
         , ar
          )
   | otherwise = do
-      rj       <- genRet (ctxAssertEvaluated bnd ctx) bnd at alts l
+      rj       <- genRet ctx bnd at alts l
       let ctx' = ctxSetTop bnd
                   $ ctxSetTarget (assocIdExprs bnd (map toJExpr [R1 ..]))
                   $ ctx
@@ -596,7 +628,7 @@ genRet :: HasDebugCallStack
        -> AltType
        -> [CgStgAlt]
        -> LiveVars
-       -> G JStat
+       -> G JStgStat
 genRet ctx e at as l = freshIdent >>= f
   where
     allRefs :: [Id]
@@ -608,8 +640,8 @@ genRet ctx e at as l = freshIdent >>= f
     isLne i    = ctxIsLneBinding ctx i || ctxIsLneLiveVar ctx' i
     nonLne     = filter (not . isLne) (dVarSetElems l)
 
-    f :: Ident -> G JStat
-    f r@(TxtI ri)    =  do
+    f :: Ident -> G JStgStat
+    f r@(identFS -> ri)    =  do
       pushLne  <- pushLneFrame lneLive ctx
       saveCCS  <- ifProfilingM $ push [jCurrentCCS]
       free     <- optimizeFree 0 nonLne
@@ -617,27 +649,28 @@ genRet ctx e at as l = freshIdent >>= f
       fun'     <- fun free
       sr       <- genStaticRefs l -- srt
       prof     <- profiling
-      emitClosureInfo $
-        ClosureInfo r
-                    (CIRegs 0 altRegs)
-                    ri
-                    (fixedLayout . reverse $
+      emitClosureInfo $ ClosureInfo
+        { ciVar = r
+        , ciRegs = CIRegs 0 altRegs
+        , ciName = ri
+        , ciLayout = fixedLayout . reverse $
                        map (stackSlotType . fst3) free
-                       ++ if prof then [ObjV] else map stackSlotType lneVars)
-                    CIStackFrame
-                    sr
-      emitToplevel $ jFunction r [] fun'
+                       ++ if prof then [ObjV] else map stackSlotType lneVars
+        , ciType = CIStackFrame
+        , ciStatic = sr
+        }
+      emitToplevel $ FuncStat r [] fun'
       return (pushLne <> saveCCS <> pushRet)
     fst3 ~(x,_,_)  = x
 
-    altRegs :: HasDebugCallStack => [VarType]
+    altRegs :: HasDebugCallStack => [JSRep]
     altRegs = case at of
-      PrimAlt ptc    -> [primRepVt ptc]
-      MultiValAlt _n -> idVt e
+      PrimAlt ptc    -> [primRepToJSRep ptc]
+      MultiValAlt _n -> idJSRep e
       _              -> [PtrV]
 
     -- special case for popping CCS but preserving stack size
-    pop_handle_CCS :: [(JExpr, StackSlot)] -> G JStat
+    pop_handle_CCS :: [(JStgExpr, StackSlot)] -> G JStgStat
     pop_handle_CCS [] = return mempty
     pop_handle_CCS xs = do
       -- grab the slots from 'xs' and push
@@ -656,7 +689,7 @@ genRet ctx e at as l = freshIdent >>= f
       restoreCCS    <- ifProfilingM . pop_handle_CCS $ pure (jCurrentCCS, SlotUnknown)
       rlne          <- popLneFrame False lneLive ctx'
       rlnev         <- verifyRuntimeReps lneVars
-      (alts, _altr) <- genAlts ctx' e at Nothing as
+      (alts, _altr) <- genAlts ctx' e at as
       return $ decs <> load <> loadv <> ras <> rasv <> restoreCCS <> rlne <> rlnev <> alts <>
                returnStack
 
@@ -667,10 +700,9 @@ genAlts :: HasDebugCallStack
         => ExprCtx        -- ^ lhs to assign expression result to
         -> Id             -- ^ id being matched
         -> AltType        -- ^ type
-        -> Maybe [JExpr]  -- ^ if known, fields in datacon from earlier expression
         -> [CgStgAlt]     -- ^ the alternatives
-        -> G (JStat, ExprResult)
-genAlts ctx e at me alts = do
+        -> G (JStgStat, ExprResult)
+genAlts ctx e at alts = do
   (st, er) <- case at of
 
     PolyAlt -> case alts of
@@ -690,7 +722,7 @@ genAlts ctx e at me alts = do
       -> do
         ie <- varsForId e
         (r, bss) <- normalizeBranches ctx <$>
-           mapM (isolateSlots . mkPrimIfBranch ctx [primRepVt tc]) alts
+           mapM (isolateSlots . mkPrimIfBranch ctx [primRepToJSRep tc]) alts
         setSlots []
         return (mkSw ie bss, r)
 
@@ -707,15 +739,6 @@ genAlts ctx e at me alts = do
       , isUnboxedTupleTyCon tc
       -> panic "genAlts: unexpected unboxed tuple"
 
-    AlgAlt _tc
-      | Just es <- me
-      , [GenStgAlt (DataAlt dc) bs expr] <- alts
-      , not (isUnboxableCon dc)
-      -> do
-        bsi <- mapM identsForId bs
-        (ej, er) <- genExpr ctx expr
-        return (declAssignAll (concat bsi) es <> ej, er)
-
     AlgAlt _tc
       | [alt] <- alts
       -> do
@@ -754,7 +777,7 @@ genAlts ctx e at me alts = do
 -- | If 'StgToJSConfig.csRuntimeAssert' is set, then generate an assertion that
 -- asserts the pattern match is valid, e.g., the match is attempted on a
 -- Boolean, a Data Constructor, or some number.
-verifyMatchRep :: HasDebugCallStack => Id -> AltType -> G JStat
+verifyMatchRep :: HasDebugCallStack => Id -> AltType -> G JStgStat
 verifyMatchRep x alt = do
   runtime_assert <- csRuntimeAssert <$> getSettings
   if not runtime_assert
@@ -762,14 +785,14 @@ verifyMatchRep x alt = do
     else case alt of
       AlgAlt tc -> do
         ix <- varsForId x
-        pure $ ApplStat (var "h$verify_match_alg") (ValExpr(JStr(mkFastString (renderWithContext defaultSDocContext (ppr tc)))):ix)
+        pure $ ApplStat (global "h$verify_match_alg") (ValExpr (JStr (mkFastString (renderWithContext defaultSDocContext (ppr tc)))):ix)
       _ -> pure mempty
 
 -- | A 'Branch' represents a possible branching path of an Stg case statement,
 -- i.e., a possible code path from an 'StgAlt'
 data Branch a = Branch
   { branch_expr   :: a
-  , branch_stat   :: JStat
+  , branch_stat   :: JStgStat
   , branch_result :: ExprResult
   }
   deriving (Eq,Functor)
@@ -785,29 +808,29 @@ normalizeBranches ctx brs
     | branchResult (fmap branch_result brs) == ExprCont =
         (ExprCont, map mkCont brs)
     | otherwise =
-        (ExprInline Nothing, brs)
+        (ExprInline, brs)
   where
     mkCont b = case branch_result b of
-      ExprInline{} -> b { branch_stat   = branch_stat b <> assignAll jsRegsFromR1
-                                                                     (concatMap typex_expr $ ctxTarget ctx)
-                        , branch_result = ExprCont
-                        }
+      ExprInline -> b { branch_stat   = branch_stat b <> assignAll jsRegsFromR1
+                                                                   (concatMap typex_expr $ ctxTarget ctx)
+                      , branch_result = ExprCont
+                      }
       _ -> b
 
 -- | Load an unboxed tuple. "Loading" means getting all 'Idents' from the input
 -- ID's, declaring them as variables in JS land and binding them, in order, to
 -- 'es'.
-loadUbxTup :: [JExpr] -> [Id] -> Int -> G JStat
+loadUbxTup :: [JStgExpr] -> [Id] -> Int -> G JStgStat
 loadUbxTup es bs _n = do
   bs' <- concatMapM identsForId bs
   return $ declAssignAll bs' es
 
-mkSw :: [JExpr] -> [Branch (Maybe [JExpr])] -> JStat
+mkSw :: [JStgExpr] -> [Branch (Maybe [JStgExpr])] -> JStgStat
 mkSw [e] cases = mkSwitch e (fmap (fmap (fmap head)) cases)
 mkSw es cases  = mkIfElse es cases
 
 -- | Switch for pattern matching on constructors or prims
-mkSwitch :: JExpr -> [Branch (Maybe JExpr)] -> JStat
+mkSwitch :: JStgExpr -> [Branch (Maybe JStgExpr)] -> JStgStat
 mkSwitch e cases
   | [Branch (Just c1) s1 _] <- n
   , [Branch _ s2 _] <- d
@@ -832,7 +855,7 @@ mkSwitch e cases
 -- | if/else for pattern matching on things that js cannot switch on
 -- the list of branches is expected to have the default alternative
 -- first, if it exists
-mkIfElse :: [JExpr] -> [Branch (Maybe [JExpr])] -> JStat
+mkIfElse :: [JStgExpr] -> [Branch (Maybe [JStgExpr])] -> JStgStat
 mkIfElse e s = go (L.reverse s)
     where
       go = \case
@@ -841,11 +864,11 @@ mkIfElse e s = go (L.reverse s)
         [] -> panic "mkIfElse: empty expression list"
         _  -> panic "mkIfElse: multiple DEFAULT cases"
 
--- | Wrapper to contruct sequences of (===), e.g.,
+-- | Wrapper to construct sequences of (===), e.g.,
 --
 -- > mkEq [l0,l1,l2] [r0,r1,r2] = (l0 === r0) && (l1 === r1) && (l2 === r2)
 --
-mkEq :: [JExpr] -> [JExpr] -> JExpr
+mkEq :: [JStgExpr] -> [JStgExpr] -> JStgExpr
 mkEq es1 es2
   | length es1 == length es2 = foldl1 (InfixExpr LAndOp) (zipWith (InfixExpr StrictEqOp) es1 es2)
   | otherwise                = panic "mkEq: incompatible expressions"
@@ -853,7 +876,7 @@ mkEq es1 es2
 mkAlgBranch :: ExprCtx   -- ^ toplevel id for the result
             -> Id        -- ^ datacon to match
             -> CgStgAlt  -- ^ match alternative with binders
-            -> G (Branch (Maybe JExpr))
+            -> G (Branch (Maybe JStgExpr))
 mkAlgBranch top d alt
   | DataAlt dc <- alt_con alt
   , isUnboxableCon dc
@@ -877,52 +900,54 @@ mkAlgBranch top d alt
 
 -- | Generate a primitive If-expression
 mkPrimIfBranch :: ExprCtx
-               -> [VarType]
+               -> [JSRep]
                -> CgStgAlt
-               -> G (Branch (Maybe [JExpr]))
+               -> G (Branch (Maybe [JStgExpr]))
 mkPrimIfBranch top _vt alt =
   (\ic (ej,er) -> Branch ic ej er) <$> ifCond (alt_con alt) <*> genExpr top (alt_rhs alt)
 
 -- fixme are bool things always checked correctly here?
-ifCond :: AltCon -> G (Maybe [JExpr])
+ifCond :: AltCon -> G (Maybe [JStgExpr])
 ifCond = \case
   DataAlt da -> return $ Just [toJExpr (dataConTag da)]
   LitAlt l   -> Just <$> genLit l
   DEFAULT    -> return Nothing
 
-caseCond :: AltCon -> G (Maybe JExpr)
+caseCond :: AltCon -> G (Maybe JStgExpr)
 caseCond = \case
+-- fixme use single tmp var for all branches
   DEFAULT    -> return Nothing
   DataAlt da -> return $ Just (toJExpr $ dataConTag da)
   LitAlt l   -> genLit l >>= \case
     [e] -> pure (Just e)
-    es  -> pprPanic "caseCond: expected single-variable literal" (ppr $ satJExpr Nothing <$> es)
+    es  -> pprPanic "caseCond: expected single-variable literal" (ppr $ jStgExprToJS <$> es)
 
--- fixme use single tmp var for all branches
 -- | Load parameters from constructor
-loadParams :: JExpr -> [Id] -> G JStat
+loadParams :: JStgExpr -> [Id] -> G JStgStat
 loadParams from args = do
   as <- concat <$> zipWithM (\a u -> map (,u) <$> identsForId a) args use
-  return $ case as of
-    []                 -> mempty
-    [(x,u)]            -> loadIfUsed (from .^ closureField1_) x  u
-    [(x1,u1),(x2,u2)]  -> mconcat
+  case as of
+    []                 -> return mempty
+    [(x,u)]            -> return $ loadIfUsed (from .^ closureField1_) x  u
+    [(x1,u1),(x2,u2)]  -> return $ mconcat
                             [ loadIfUsed (from .^ closureField1_) x1 u1
                             , loadIfUsed (from .^ closureField2_) x2 u2
                             ]
-    ((x,u):xs)         -> mconcat
-                            [ loadIfUsed (from .^ closureField1_) x u
-                            , jVar (\d -> mconcat [ d |= from .^ closureField2_
-                                                  , loadConVarsIfUsed d xs
-                                                  ])
-                            ]
+    ((x,u):xs)         -> do d <- freshIdent
+                             return $ mconcat
+                               [ loadIfUsed (from .^ closureField1_) x u
+                               , mconcat [ d ||= from .^ closureField2_
+                                         , loadConVarsIfUsed (Var d) xs
+                                         ]
+                               ]
   where
     use = repeat True -- fixme clean up
+
     loadIfUsed fr tgt True = tgt ||= fr
     loadIfUsed  _ _   _    = mempty
 
     loadConVarsIfUsed fr cs = mconcat $ zipWith f cs [(1::Int)..]
-      where f (x,u) n = loadIfUsed (SelExpr fr (TxtI (dataFieldName n))) x u
+      where f (x,u) n = loadIfUsed (SelExpr fr (name (dataFieldName n))) x u
 
 -- | Determine if a branch will end in a continuation or not. If not the inline
 -- branch must be normalized. See 'normalizeBranches'
@@ -934,98 +959,113 @@ branchResult = \case
   (ExprCont:_)         -> ExprCont
   (_:es)
     | elem ExprCont es -> ExprCont
-    | otherwise        -> ExprInline Nothing
+    | otherwise        -> ExprInline
 
 -- | Push return arguments onto the stack. The 'Bool' tracks whether the value
 -- is already on the stack or not, used in 'StgToJS.Stack.pushOptimized'.
-pushRetArgs :: HasDebugCallStack => [(Id,Int,Bool)] -> JExpr -> G JStat
+pushRetArgs :: HasDebugCallStack => [(Id,Int,Bool)] -> JStgExpr -> G JStgStat
 pushRetArgs free fun = do
   rs <- mapM (\(i,n,b) -> (\es->(es!!(n-1),b)) <$> genIdArg i) free
   pushOptimized (rs++[(fun,False)])
 
 -- | Load the return arguments then pop the stack frame
-loadRetArgs :: HasDebugCallStack => [(Id,Int,Bool)] -> G JStat
+loadRetArgs :: HasDebugCallStack => [(Id,Int,Bool)] -> G JStgStat
 loadRetArgs free = do
   ids <- mapM (\(i,n,_b) -> (!! (n-1)) <$> genIdStackArgI i) free
   popSkipI 1 ids
 
+-- All identifiers referenced by the expression (does not traverse into nested functions)
+allVars :: JStgExpr -> [Ident]
+allVars (ValExpr v) = case v of
+  (JVar i) -> [i]
+  (JList xs) -> concatMap allVars xs
+  (JHash xs) -> concatMap (allVars . snd) (UM.nonDetUniqMapToList xs)
+  (JInt {})  -> []
+  (JDouble {}) -> []
+  (JStr {}) -> []
+  (JRegEx {}) -> []
+  (JBool {}) -> []
+  (JFunc is _s) -> is
+allVars (InfixExpr _op lh rh) = allVars lh ++ allVars rh
+allVars (ApplExpr f xs) = allVars f ++ concatMap allVars xs
+allVars (IfExpr c t e) = allVars c ++ allVars t ++ allVars e
+allVars (UOpExpr _op x) = allVars x
+allVars (SelExpr e _) = allVars e
+allVars (IdxExpr e i) = allVars e ++ allVars i
+
 -- | allocate multiple, possibly mutually recursive, closures
-allocDynAll :: Bool -> Maybe JStat -> [(Ident,JExpr,[JExpr],CostCentreStack)] -> G JStat
-{-
-XXX remove use of template and enable in-place init again
+allocDynAll :: Bool -> Maybe JStgStat -> [(Ident,JStgExpr,[JStgExpr],CostCentreStack)] -> G JStgStat
 allocDynAll haveDecl middle [(to,entry,free,cc)]
-  | isNothing middle && to `notElem` (free ^.. template) = do
+  | isNothing middle && to `notElem` concatMap allVars free = do
       ccs <- ccsVarJ cc
-      return $ allocDynamic s haveDecl to entry free ccs -}
+      s <- getSettings
+      return $ allocDynamic s (not haveDecl) to entry free ccs
 allocDynAll haveDecl middle cls = do
   settings <- getSettings
   let
+    middle' :: JStgStat
     middle' = fromMaybe mempty middle
 
     decl_maybe i e
       | haveDecl  = toJExpr i |= e
       | otherwise = i ||= e
 
-    makeObjs :: G JStat
+    makeObjs :: G JStgStat
     makeObjs =
       fmap mconcat $ forM cls $ \(i,f,_,cc) -> do
       ccs <- maybeToList <$> costCentreStackLbl cc
       pure $ mconcat
         [ decl_maybe i $ if csInlineAlloc settings
-            then ValExpr (jhFromList $ [ (closureEntry_ , f)
+            then ValExpr (jhFromList $ [ (closureInfo_ , f)
                                        , (closureField1_, null_)
                                        , (closureField2_, null_)
                                        , (closureMeta_  , zero_)
                                        ]
                              ++ fmap (\cid -> ("cc", ValExpr (JVar cid))) ccs)
-            else ApplExpr (var "h$c") (f : fmap (ValExpr . JVar) ccs)
+            else ApplExpr hdC (f : fmap (ValExpr . JVar) ccs)
         ]
 
-    fillObjs = mconcat $ map fillObj cls
-    fillObj (i,_,es,_)
-      | csInlineAlloc settings || length es > 24 =
-          case es of
-            []      -> mempty
-            [ex]    -> toJExpr i .^ closureField1_ |= toJExpr ex
-            [e1,e2] -> mconcat
-                        [ toJExpr i .^ closureField1_ |= toJExpr e1
-                        , toJExpr i .^ closureField2_ |= toJExpr e2
-                        ]
-            (ex:es)  -> mconcat
-                        [ toJExpr i .^ closureField1_ |= toJExpr ex
-                        , toJExpr i .^ closureField2_ |= toJExpr (jhFromList (zip (map dataFieldName [1..]) es))
-                        ]
-      | otherwise = case es of
-            []      -> mempty
-            [ex]    -> toJExpr i .^ closureField1_ |= ex
-            [e1,e2] -> mconcat
-                        [ toJExpr i .^ closureField1_ |= e1
-                        , toJExpr i .^ closureField2_ |= e2
-                        ]
-            (ex:es)  -> mconcat
-                        [ toJExpr i .^ closureField1_ |= ex
-                        , toJExpr i .^ closureField2_ |= fillFun es
-                        ]
+    fillObjs :: [JStgStat]
+    fillObjs = map fillObj cls
+    fillObj (ident,_,es,_) =
+      let i = toJExpr ident
+      in case es of
+          []      -> mempty
+          [ex]    -> closureField1 i |= ex
+          [e1,e2] -> mconcat
+                      [ closureField1 i |= e1
+                      , closureField2 i |= e2
+                      ]
+          (ex:es)
+            | csInlineAlloc settings || length es > 24
+            -> mconcat [ closureField1 i |= ex
+                       , closureField2 i |= ValExpr (jhFromList (zip (map dataFieldName [1..]) es))
+                       ]
 
-    fillFun [] = null_
-    fillFun es = ApplExpr (allocData (length es)) es
+            | otherwise
+            -> mconcat [ closureField1 i |= ex
+                       , closureField2 i |= ApplExpr (allocData (length es)) es
+                       ]
 
-    checkObjs | csAssertRts settings  = mconcat $
-                map (\(i,_,_,_) -> ApplStat (ValExpr (JVar (TxtI "h$checkObj"))) [toJExpr i]) cls
+    checkObjs :: [JStgStat]
+    checkObjs | csAssertRts settings  =
+                map (\(i,_,_,_) -> ApplStat hdCheckObj [Var i]) cls
               | otherwise = mempty
 
   objs <- makeObjs
-  pure $ mconcat [objs, middle', fillObjs, checkObjs]
+  return $ mconcat [objs, middle', mconcat fillObjs, mconcat checkObjs]
 
 -- | Generate a primop. This function wraps around the real generator
 -- 'GHC.StgToJS.genPrim', handling the 'ExprCtx' and all arguments before
 -- generating the primop.
-genPrimOp :: ExprCtx -> PrimOp -> [StgArg] -> Type -> G (JStat, ExprResult)
+genPrimOp :: ExprCtx -> PrimOp -> [StgArg] -> Type -> G (JStgStat, ExprResult)
 genPrimOp ctx op args t = do
   as <- concatMapM genArg args
   prof <- csProf <$> getSettings
   bound <- csBoundsCheck <$> getSettings
+  let prim_gen = withTag "h$PRM" $ genPrim prof bound t op (concatMap typex_expr $ ctxTarget ctx) as
   -- fixme: should we preserve/check the primreps?
-  return $ case genPrim prof bound t op (concatMap typex_expr $ ctxTarget ctx) as of
-             PrimInline s -> (s, ExprInline Nothing)
+  jsm <- liftIO initJSM
+  return $ case runJSM jsm prim_gen of
+             PrimInline s -> (s, ExprInline)
              PRPrimCall s -> (s, ExprCont)
