@@ -5,7 +5,7 @@
 
 -}
 
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveFunctor #-}
 
 -- | Functional dependencies
 --
@@ -18,38 +18,39 @@ module GHC.Tc.Instance.FunDeps
    , checkInstCoverage
    , checkFunDeps
    , pprFundeps
+   , instFD, closeWrtFunDeps
    )
 where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
-import GHC.Types.Name
 import GHC.Types.Var
 import GHC.Core.Class
 import GHC.Core.Predicate
 import GHC.Core.Type
-import GHC.Tc.Utils.TcType( transSuperClasses )
+import GHC.Core.RoughMap( RoughMatchTc(..) )
 import GHC.Core.Coercion.Axiom( TypeEqn )
 import GHC.Core.Unify
 import GHC.Core.InstEnv
+import GHC.Core.TyCo.FVs
+import GHC.Core.TyCo.Compare( eqTypes, eqType )
+
+import GHC.Tc.Errors.Types ( CoverageProblem(..), FailedCoverageCondition(..) )
+import GHC.Tc.Types.Constraint ( isUnsatisfiableCt_maybe )
+import GHC.Tc.Utils.TcType( transSuperClasses )
+
 import GHC.Types.Var.Set
 import GHC.Types.Var.Env
-import GHC.Core.TyCo.FVs
-import GHC.Core.TyCo.Ppr( pprWithExplicitKindsWhen )
-import GHC.Types.SrcLoc
 
 import GHC.Utils.Outputable
 import GHC.Utils.FV
-import GHC.Utils.Error( Validity(..), allValid )
+import GHC.Utils.Error( Validity'(..), allValid )
 import GHC.Utils.Misc
 import GHC.Utils.Panic
 
-import GHC.Data.Pair             ( Pair(..) )
-import Data.List        ( nubBy )
-import Data.Maybe
-import Data.Foldable    ( fold )
+import GHC.Data.Pair ( Pair(..) )
+import Data.List     ( nubBy )
+import Data.Maybe    ( isJust, isNothing )
 
 {-
 ************************************************************************
@@ -58,7 +59,6 @@ import Data.Foldable    ( fold )
 *                                                                      *
 ************************************************************************
 
-
 Each functional dependency with one variable in the RHS is responsible
 for generating a single equality. For instance:
      class C a b | a -> b
@@ -66,8 +66,6 @@ The constraints ([Wanted] C Int Bool) and [Wanted] C Int alpha
 will generate the following FunDepEqn
      FDEqn { fd_qtvs = []
            , fd_eqs  = [Pair Bool alpha]
-           , fd_pred1 = C Int Bool
-           , fd_pred2 = C Int alpha
            , fd_loc = ... }
 However notice that a functional dependency may have more than one variable
 in the RHS which will create more than one pair of types in fd_eqs. Example:
@@ -77,38 +75,61 @@ in the RHS which will create more than one pair of types in fd_eqs. Example:
 Will generate:
      FDEqn { fd_qtvs = []
            , fd_eqs  = [Pair Bool alpha, Pair alpha beta]
-           , fd_pred1 = C Int Bool
-           , fd_pred2 = C Int alpha
            , fd_loc = ... }
 
 INVARIANT: Corresponding types aren't already equal
 That is, there exists at least one non-identity equality in FDEqs.
 
+Note [Improving against instances]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 Assume:
-       class C a b c | a -> b c
-       instance C Int x x
-And:   [Wanted] C Int Bool alpha
-We will /match/ the LHS of fundep equations, producing a matching substitution
-and create equations for the RHS sides. In our last example we'd have generated:
-      ({x}, [fd1,fd2])
-where
-       fd1 = FDEq 1 Bool x
-       fd2 = FDEq 2 alpha x
-To ``execute'' the equation, make fresh type variable for each tyvar in the set,
-instantiate the two types with these fresh variables, and then unify or generate
-a new constraint. In the above example we would generate a new unification
-variable 'beta' for x and produce the following constraints:
-     [Wanted] (Bool ~ beta)
-     [Wanted] (alpha ~ beta)
+   class C a b | a -> b
+   instance C Int Bool
+   [W] C Int ty
 
-Notice the subtle difference between the above class declaration and:
-       class C a b c | a -> b, a -> c
-where we would generate:
-      ({x},[fd1]),({x},[fd2])
-This means that the template variable would be instantiated to different
-unification variables when producing the FD constraints.
+Then `improveFromInstEnv` should return a FDEqn with
+   FDEqn { fd_qtvs = [], fd_eqs = [Pair Bool ty] }
 
-Finally, the position parameters will help us rewrite the wanted constraint ``on the spot''
+describing an equality (Bool ~ ty).  To do this we /match/ the instance head
+against the [W], using just the LHS of the fundep; if we match, we return
+an equality for the RHS.
+
+Wrinkles:
+
+(1) meta_tvs: sometimes the instance mentions variables in the RHS that
+    are not bound in the LHS.  For example
+
+     class C a b | a -> b
+     instance C Int (Maybe x)
+     [W] C Int ty
+
+    Note that although the `Int` parts match, that does not fix what `x` is.
+    So we just make up a fresh unification variable (a meta_tv), to give the
+    "shape" of the RHS.  So we emit the FDEqun
+       FDEqn { fd_qtvs = [x], fd_eqs = [Pair (Maybe x) ty] }
+
+    Note that the fd_qtvs can be free in the /first/ component of the Pair,
+    but not in the second (which comes from the [W] constraint).
+
+(2) Multi-range fundeps. When these meta_tvs are involved, there is a subtle
+    difference between the fundep (a -> b c) and the two fundeps (a->b, a->c).
+    Consider
+       class D a b c | a -> b c
+       instance D Int x (Maybe x)
+       [W] D Int Bool ty
+
+    Then we'll generate
+       FDEqn { fd_qtvs = [x0], fd_eqs = [ x0 ~ Bool, Maybe x0 ~ ty] }
+    which generates one fresh unification variable x0
+
+    But if the fundeps had been (a->b, a->c) we'd generate two FDEqns
+       FDEqn { fd_qtvs = [x1], fd_eqs = [ x1 ~ Bool ] }
+       FDEqn { fd_qtvs = [x2], fd_eqs = [ Maybe x2 ~ ty ] }
+    with two FDEqns, generating two separate unification variables.
+
+(3) improveFromInstEnv doesn't return any equations that already hold.
+    Reason: then we know if any actual improvement has happened, in
+    which case we need to iterate the solver
 -}
 
 data FunDepEqn loc
@@ -117,9 +138,21 @@ data FunDepEqn loc
                                  -- Non-empty only for FunDepEqns arising from instance decls
 
           , fd_eqs   :: [TypeEqn]  -- Make these pairs of types equal
-          , fd_pred1 :: PredType   -- The FunDepEqn arose from
-          , fd_pred2 :: PredType   --  combining these two constraints
+                                   -- Invariant: In each (Pair ty1 ty2), the fd_qtvs may be
+                                   -- free in ty1 but not in ty2.  See Wrinkle (1) of
+                                   -- Note [Improving against instances]
+
           , fd_loc   :: loc  }
+    deriving Functor
+
+instance Outputable (FunDepEqn a) where
+  ppr = pprEquation
+
+pprEquation :: FunDepEqn a -> SDoc
+pprEquation (FDEqn { fd_qtvs = qtvs, fd_eqs = pairs })
+  = vcat [text "forall" <+> braces (pprWithCommas ppr qtvs),
+          nest 2 (vcat [ ppr t1 <+> text "~" <+> ppr t2
+                       | Pair t1 t2 <- pairs])]
 
 {-
 Given a bunch of predicates that must hold, such as
@@ -180,7 +213,7 @@ improveFromAnother loc pred1 pred2
   | Just (cls1, tys1) <- getClassPredTys_maybe pred1
   , Just (cls2, tys2) <- getClassPredTys_maybe pred2
   , cls1 == cls2
-  = [ FDEqn { fd_qtvs = [], fd_eqs = eqs, fd_pred1 = pred1, fd_pred2 = pred2, fd_loc = loc }
+  = [ FDEqn { fd_qtvs = [], fd_eqs = eqs, fd_loc = loc }
     | let (cls_tvs, cls_fds) = classTvsFds cls1
     , fd <- cls_fds
     , let (ltys1, rs1) = instFD fd cls_tvs tys1
@@ -195,30 +228,15 @@ improveFromAnother _ _ _ = []
 -- Improve a class constraint from instance declarations
 -- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
-instance Outputable (FunDepEqn a) where
-  ppr = pprEquation
-
-pprEquation :: FunDepEqn a -> SDoc
-pprEquation (FDEqn { fd_qtvs = qtvs, fd_eqs = pairs })
-  = vcat [text "forall" <+> braces (pprWithCommas ppr qtvs),
-          nest 2 (vcat [ ppr t1 <+> text "~" <+> ppr t2
-                       | Pair t1 t2 <- pairs])]
-
 improveFromInstEnv :: InstEnvs
-                   -> (PredType -> SrcSpan -> loc)
-                   -> PredType
+                   -> (ClsInst -> loc)
+                   -> Class -> [Type]
                    -> [FunDepEqn loc] -- Needs to be a FunDepEqn because
                                       -- of quantified variables
+-- See Note [Improving against instances]
 -- Post: Equations oriented from the template (matching instance) to the workitem!
-improveFromInstEnv inst_env mk_loc pred
-  | Just (cls, tys) <- ASSERT2( isClassPred pred, ppr pred )
-                       getClassPredTys_maybe pred
-  , let (cls_tvs, cls_fds) = classTvsFds cls
-        instances          = classInstances inst_env cls
-        rough_tcs          = roughMatchTcs tys
-  = [ FDEqn { fd_qtvs = meta_tvs, fd_eqs = eqs
-            , fd_pred1 = p_inst, fd_pred2 = pred
-            , fd_loc = mk_loc p_inst (getSrcSpan (is_dfun ispec)) }
+improveFromInstEnv inst_env mk_loc cls tys
+  = [ FDEqn { fd_qtvs = meta_tvs, fd_eqs = eqs, fd_loc = mk_loc ispec }
     | fd <- cls_fds             -- Iterate through the fundeps first,
                                 -- because there often are none!
     , let trimmed_tcs = trimRoughMatchTcs cls_tvs fd rough_tcs
@@ -229,53 +247,35 @@ improveFromInstEnv inst_env mk_loc pred
     , ispec <- instances
     , (meta_tvs, eqs) <- improveClsFD cls_tvs fd ispec
                                       tys trimmed_tcs -- NB: orientation
-    , let p_inst = mkClassPred cls (is_tys ispec)
     ]
-improveFromInstEnv _ _ _ = []
-
+  where
+    (cls_tvs, cls_fds) = classTvsFds cls
+    instances          = classInstances inst_env cls
+    rough_tcs          = RM_KnownTc (className cls) : roughMatchTcs tys
 
 improveClsFD :: [TyVar] -> FunDep TyVar    -- One functional dependency from the class
              -> ClsInst                    -- An instance template
              -> [Type] -> [RoughMatchTc]   -- Arguments of this (C tys) predicate
              -> [([TyCoVar], [TypeEqn])]   -- Empty or singleton
+-- See Note [Improving against instances]
 
 improveClsFD clas_tvs fd
              (ClsInst { is_tvs = qtvs, is_tys = tys_inst, is_tcs = rough_tcs_inst })
              tys_actual rough_tcs_actual
 
--- Compare instance      {a,b}    C sx sp sy sq
---         with wanted     [W] C tx tp ty tq
---         for fundep (x,y -> p,q)  from class  (C x p y q)
--- If (sx,sy) unifies with (tx,ty), take the subst S
-
--- 'qtvs' are the quantified type variables, the ones which can be instantiated
--- to make the types match.  For example, given
---      class C a b | a->b where ...
---      instance C (Maybe x) (Tree x) where ..
---
--- and a wanted constraint of form (C (Maybe t1) t2),
--- then we will call checkClsFD with
---
---      is_qtvs = {x}, is_tys = [Maybe x,  Tree x]
---                     tys_actual = [Maybe t1, t2]
---
--- We can instantiate x to t1, and then we want to force
---      (Tree x) [t1/x]  ~   t2
-
   | instanceCantMatch rough_tcs_inst rough_tcs_actual
   = []          -- Filter out ones that can't possibly match,
 
   | otherwise
-  = ASSERT2( equalLength tys_inst tys_actual &&
-             equalLength tys_inst clas_tvs
-            , ppr tys_inst <+> ppr tys_actual )
+  = assertPpr (equalLength tys_inst tys_actual &&
+               equalLength tys_inst clas_tvs)
+              (ppr tys_inst <+> ppr tys_actual) $
 
-    case tcMatchTyKis ltys1 ltys2 of
+    case tcMatchTyKisX init_subst ltys1 ltys2 of
         Nothing  -> []
         Just subst | isJust (tcMatchTyKisX subst rtys1 rtys2)
                         -- Don't include any equations that already hold.
-                        -- Reason: then we know if any actual improvement has happened,
-                        --         in which case we need to iterate the solver
+                        -- See Note [Improving against instances] wrinkle (3)
                         -- In making this check we must taking account of the fact that any
                         -- qtvs that aren't already instantiated can be instantiated to anything
                         -- at all
@@ -306,7 +306,7 @@ improveClsFD clas_tvs fd
                         -- executed.  What we're doing instead is recording the partial
                         -- work of the ls1/ls2 unification leaving a smaller unification problem
                   where
-                    rtys1' = map (substTyUnchecked subst) rtys1
+                    rtys1' = map (substTy subst) rtys1
 
                     fdeqs = zipAndComputeFDEqs (\_ _ -> False) rtys1' rtys2
                         -- Don't discard anything!
@@ -314,8 +314,10 @@ improveClsFD clas_tvs fd
                         -- eqType again, since we know for sure that /at least one/
                         -- equation in there is useful)
 
-                    meta_tvs = [ setVarType tv (substTyUnchecked subst (varType tv))
-                               | tv <- qtvs, tv `notElemTCvSubst` subst ]
+                    meta_tvs = [ setVarType tv (substTy subst (varType tv))
+                               | tv <- qtvs
+                               , tv `notElemSubst` subst
+                               , tv `elemVarSet` rtys1_tvs ]
                         -- meta_tvs are the quantified type variables
                         -- that have not been substituted out
                         --
@@ -330,11 +332,14 @@ improveClsFD clas_tvs fd
                         --              type variables' kinds
                         --          (b) we must apply 'subst' to the kinds, in case we have
                         --              matched out a kind variable, but not a type variable
-                        --              whose kind mentions that kind variable!
-                        --          #6015, #6068
+                        --              whose kind mentions that kind variable! #6015, #6068
+                        --          (c) no need to include tyvars not in rtys1
   where
+    init_subst     = mkEmptySubst $ mkInScopeSet $
+                     mkVarSet qtvs `unionVarSet` tyCoVarsOfTypes ltys2
     (ltys1, rtys1) = instFD fd clas_tvs tys_inst
     (ltys2, rtys2) = instFD fd clas_tvs tys_actual
+    rtys1_tvs      = tyCoVarsOfTypes rtys1
 
 {-
 %************************************************************************
@@ -351,7 +356,7 @@ Example
 
 For the coverage condition, we check
    (normal)    fv(t2) `subset` fv(t1)
-   (liberal)   fv(t2) `subset` oclose(fv(t1), theta)
+   (liberal)   fv(t2) `subset` closeWrtFunDeps(fv(t1), theta)
 
 The liberal version  ensures the self-consistency of the instance, but
 it does not guarantee termination. Example:
@@ -364,7 +369,7 @@ it does not guarantee termination. Example:
    instance Mul a b c => Mul a [b] [c] where x .*. v = map (x.*.) v
 
 In the third instance, it's not the case that fv([c]) `subset` fv(a,[b]).
-But it is the case that fv([c]) `subset` oclose( theta, fv(a,[b]) )
+But it is the case that fv([c]) `subset` closeWrtFunDeps( theta, fv(a,[b]) )
 
 But it is a mistake to accept the instance because then this defn:
         f = \ b x y -> if b then x .*. [y] else y
@@ -374,60 +379,56 @@ makes instance inference go into a loop, because it requires the constraint
 
 checkInstCoverage :: Bool   -- Be liberal
                   -> Class -> [PredType] -> [Type]
-                  -> Validity
+                  -> Validity' CoverageProblem
 -- "be_liberal" flag says whether to use "liberal" coverage of
---              See Note [Coverage Condition] below
+--              See Note [Coverage condition] below
 --
 -- Return values
 --    Nothing  => no problems
 --    Just msg => coverage problem described by msg
 
 checkInstCoverage be_liberal clas theta inst_taus
+  | any (isJust . isUnsatisfiableCt_maybe) theta
+  -- As per [GHC proposal #433](https://github.com/ghc-proposals/ghc-proposals/blob/master/proposals/0433-unsatisfiable.rst),
+  -- we skip checking the coverage condition if there is an "Unsatisfiable"
+  -- constraint in the instance context.
+  --
+  -- See Note [Implementation of Unsatisfiable constraints] in GHC.Tc.Errors,
+  -- point (E).
+  = IsValid
+  | otherwise
   = allValid (map fundep_ok fds)
   where
     (tyvars, fds) = classTvsFds clas
     fundep_ok fd
-       | and (isEmptyVarSet <$> undetermined_tvs) = IsValid
-       | otherwise                                = NotValid msg
+       | all isEmptyVarSet undetermined_tvs
+       = IsValid
+       | otherwise
+       = NotValid not_covered_msg
        where
          (ls,rs) = instFD fd tyvars inst_taus
          ls_tvs = tyCoVarsOfTypes ls
-         rs_tvs = splitVisVarsOfTypes rs
+         rs_tvs = visVarsOfTypes rs
 
          undetermined_tvs | be_liberal = liberal_undet_tvs
                           | otherwise  = conserv_undet_tvs
 
-         closed_ls_tvs = oclose theta ls_tvs
+         closed_ls_tvs = closeWrtFunDeps theta ls_tvs
          liberal_undet_tvs = (`minusVarSet` closed_ls_tvs) <$> rs_tvs
          conserv_undet_tvs = (`minusVarSet` ls_tvs)        <$> rs_tvs
 
-         undet_set = fold undetermined_tvs
+         not_covered_msg =
+           CoverageProblem
+            { not_covered_fundep        = fd
+            , not_covered_fundep_inst   = (ls, rs)
+            , not_covered_invis_vis_tvs = undetermined_tvs
+            , not_covered_liberal       =
+                if be_liberal
+                then FailedLICC
+                else FailedICC
+                      { alsoFailedLICC = not $ all isEmptyVarSet liberal_undet_tvs }
+            }
 
-         msg = pprWithExplicitKindsWhen
-                 (isEmptyVarSet $ pSnd undetermined_tvs) $
-               vcat [ -- text "ls_tvs" <+> ppr ls_tvs
-                      -- , text "closed ls_tvs" <+> ppr (closeOverKinds ls_tvs)
-                      -- , text "theta" <+> ppr theta
-                      -- , text "oclose" <+> ppr (oclose theta (closeOverKinds ls_tvs))
-                      -- , text "rs_tvs" <+> ppr rs_tvs
-                      sep [ text "The"
-                            <+> ppWhen be_liberal (text "liberal")
-                            <+> text "coverage condition fails in class"
-                            <+> quotes (ppr clas)
-                          , nest 2 $ text "for functional dependency:"
-                            <+> quotes (pprFunDep fd) ]
-                    , sep [ text "Reason: lhs type"<>plural ls <+> pprQuotedList ls
-                          , nest 2 $
-                            (if isSingleton ls
-                             then text "does not"
-                             else text "do not jointly")
-                            <+> text "determine rhs type"<>plural rs
-                            <+> pprQuotedList rs ]
-                    , text "Un-determined variable" <> pluralVarSet undet_set <> colon
-                            <+> pprVarSet undet_set (pprWithCommas ppr)
-                    , ppWhen (not be_liberal &&
-                              and (isEmptyVarSet <$> liberal_undet_tvs)) $
-                      text "Using UndecidableInstances might help" ]
 
 {- Note [Closing over kinds in coverage]
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -467,17 +468,17 @@ Is the instance OK? Does {l,r,xs} determine v?  Well:
     we get {l,k,xs} -> b
 
   * Note the 'k'!! We must call closeOverKinds on the seed set
-    ls_tvs = {l,r,xs}, BEFORE doing oclose, else the {l,k,xs}->b
+    ls_tvs = {l,r,xs}, BEFORE doing closeWrtFunDeps, else the {l,k,xs}->b
     fundep won't fire.  This was the reason for #10564.
 
-  * So starting from seeds {l,r,xs,k} we do oclose to get
+  * So starting from seeds {l,r,xs,k} we do closeWrtFunDeps to get
     first {l,r,xs,k,b}, via the HMemberM constraint, and then
     {l,r,xs,k,b,v}, via the HasFieldM1 constraint.
 
   * And that fixes v.
 
 However, we must closeOverKinds whenever augmenting the seed set
-in oclose!  Consider #10109:
+in closeWrtFunDeps!  Consider #10109:
 
   data Succ a   -- Succ :: forall k. k -> *
   class Add (a :: k1) (b :: k2) (ab :: k3) | a b -> ab
@@ -493,25 +494,27 @@ the variables free in (Succ {k3} ab).
 Bottom line:
   * closeOverKinds on initial seeds (done automatically
     by tyCoVarsOfTypes in checkInstCoverage)
-  * and closeOverKinds whenever extending those seeds (in oclose)
+  * and closeOverKinds whenever extending those seeds (in closeWrtFunDeps)
 
 Note [The liberal coverage condition]
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-(oclose preds tvs) closes the set of type variables tvs,
+(closeWrtFunDeps preds tvs) closes the set of type variables tvs,
 wrt functional dependencies in preds.  The result is a superset
 of the argument set.  For example, if we have
         class C a b | a->b where ...
 then
-        oclose [C (x,y) z, C (x,p) q] {x,y} = {x,y,z}
+        closeWrtFunDeps [C (x,y) z, C (x,p) q] {x,y} = {x,y,z}
 because if we know x and y then that fixes z.
 
 We also use equality predicates in the predicates; if we have an
 assumption `t1 ~ t2`, then we use the fact that if we know `t1` we
 also know `t2` and the other way.
-  eg    oclose [C (x,y) z, a ~ x] {a,y} = {a,y,z,x}
+  eg    closeWrtFunDeps [C (x,y) z, a ~ x] {a,y} = {a,y,z,x}
 
-oclose is used (only) when checking the coverage condition for
-an instance declaration
+closeWrtFunDeps is used
+ - when checking the coverage condition for an instance declaration
+ - when determining which tyvars are unquantifiable during generalization, in
+   GHC.Tc.Solver.decidePromotedTyVars.
 
 Note [Equality superclasses]
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -522,10 +525,10 @@ Remember from Note [The equality types story] in GHC.Builtin.Types.Prim, that
   * (a ~~ b) is a superclass of (a ~ b)
   * (a ~# b) is a superclass of (a ~~ b)
 
-So when oclose expands superclasses we'll get a (a ~# [b]) superclass.
+So when closeWrtFunDeps expands superclasses we'll get a (a ~# [b]) superclass.
 But that's an EqPred not a ClassPred, and we jolly well do want to
 account for the mutual functional dependencies implied by (t1 ~# t2).
-Hence the EqPred handling in oclose.  See #10778.
+Hence the EqPred handling in closeWrtFunDeps.  See #10778.
 
 Note [Care with type functions]
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -535,7 +538,7 @@ Consider (#12803)
   type family G c d = r | r -> d
 
 Now consider
-  oclose (C (F a b) (G c d)) {a,b}
+  closeWrtFunDeps (C (F a b) (G c d)) {a,b}
 
 Knowing {a,b} fixes (F a b) regardless of the injectivity of F.
 But knowing (G c d) fixes only {d}, because G is only injective
@@ -544,12 +547,17 @@ in its second parameter.
 Hence the tyCoVarsOfTypes/injTyVarsOfTypes dance in tv_fds.
 -}
 
-oclose :: [PredType] -> TyCoVarSet -> TyCoVarSet
+closeWrtFunDeps :: [PredType] -> TyCoVarSet -> TyCoVarSet
 -- See Note [The liberal coverage condition]
-oclose preds fixed_tvs
+closeWrtFunDeps preds fixed_tvs
   | null tv_fds = fixed_tvs -- Fast escape hatch for common case.
-  | otherwise   = fixVarSet extend fixed_tvs
+  | otherwise   = assertPpr (closeOverKinds fixed_tvs == fixed_tvs)
+                    (vcat [ text "closeWrtFunDeps: fixed_tvs is not closed over kinds"
+                          , text "fixed_tvs:" <+> ppr fixed_tvs
+                          , text "closure:" <+> ppr (closeOverKinds fixed_tvs) ])
+                $ fixVarSet extend fixed_tvs
   where
+
     extend fixed_tvs = foldl' add fixed_tvs tv_fds
        where
           add fixed_tvs (ls,rs)
@@ -613,7 +621,7 @@ In checkFunDeps we check that a new ClsInst is consistent with all the
 ClsInsts in the environment.
 
 The bogus aspect is discussed in #10675. Currently it if the two
-types are *contradicatory*, using (isNothing . tcUnifyTys).  But all
+types are *contradictory*, using (isNothing . tcUnifyTys).  But all
 the papers say we should check if the two types are *equal* thus
    not (substTys subst rtys1 `eqTypes` substTys subst rtys2)
 For now I'm leaving the bogus form because that's the way it has
@@ -675,8 +683,9 @@ trimRoughMatchTcs :: [TyVar] -> FunDep TyVar -> [RoughMatchTc] -> [RoughMatchTc]
 -- Hence, we Nothing-ise the tb and tc types right here
 --
 -- Result list is same length as input list, just with more Nothings
-trimRoughMatchTcs clas_tvs (ltvs, _) mb_tcs
-  = zipWith select clas_tvs mb_tcs
+trimRoughMatchTcs _clas_tvs _ [] = panic "trimRoughMatchTcs: nullary [RoughMatchTc]"
+trimRoughMatchTcs clas_tvs (ltvs, _) (cls:mb_tcs)
+  = cls : zipWith select clas_tvs mb_tcs
   where
     select clas_tv mb_tc | clas_tv `elem` ltvs = mb_tc
-                         | otherwise           = OtherTc
+                         | otherwise           = RM_WildCard
