@@ -34,7 +34,7 @@ module GHC.Tc.Solver.Monad (
     -- The pipeline
     StopOrContinue(..), continueWith, stopWith,
     startAgainWith, SolverStage(Stage, runSolverStage), simpleStage,
-    stopWithStage,
+    stopWithStage, nopStage,
 
     -- Tracing etc
     panicTcS, traceTcS, tryEarlyAbortTcS,
@@ -44,13 +44,14 @@ module GHC.Tc.Solver.Monad (
 
     -- Evidence creation and transformation
     MaybeNew(..), freshGoals, isFresh, getEvExpr,
+    CanonicalEvidence(..),
 
     newTcEvBinds, newNoTcEvBinds,
     newWantedEq, emitNewWantedEq,
     newWanted,
     newWantedNC, newWantedEvVarNC,
     newBoundEvVarId,
-    unifyTyVar, reportUnifications, touchabilityAndShapeTest,
+    unifyTyVar, reportUnifications,
     setEvBind, setWantedEq,
     setWantedEvTerm, setEvBindIfWanted,
     newEvVar, newGivenEvVar, emitNewGivens,
@@ -148,15 +149,20 @@ import GHC.Driver.DynFlags
 
 import GHC.Tc.Instance.Class( safeOverlap, instanceReturnsDictCon )
 import GHC.Tc.Instance.FunDeps( FunDepEqn(..) )
-import GHC.Tc.Utils.TcType
+
+
 import GHC.Tc.Solver.Types
 import GHC.Tc.Solver.InertSet
-import GHC.Tc.Types.Evidence
 import GHC.Tc.Errors.Types
+
+import GHC.Tc.Utils.TcType
+import GHC.Tc.Utils.Unify
+
+import GHC.Tc.Types.Evidence
 import GHC.Tc.Types
 import GHC.Tc.Types.Origin
+import GHC.Tc.Types.CtLoc
 import GHC.Tc.Types.Constraint
-import GHC.Tc.Utils.Unify
 
 import GHC.Builtin.Names ( unsatisfiableClassNameKey )
 
@@ -172,6 +178,7 @@ import GHC.Core.TyCon
 import GHC.Types.Name
 import GHC.Types.TyThing
 import GHC.Types.Name.Reader
+import GHC.Types.DefaultEnv ( DefaultEnv )
 import GHC.Types.Var
 import GHC.Types.Var.Set
 import GHC.Types.Unique.Supply
@@ -284,6 +291,9 @@ instance Monad SolverStage where
                            Stop ev d      -> return (Stop ev d)
                            ContinueWith x -> runSolverStage (k x) }
 
+nopStage :: a -> SolverStage a
+nopStage res = Stage (continueWith res)
+
 simpleStage :: TcS a -> SolverStage a
 -- Always does a ContinueWith; no Stop or StartAgain
 simpleStage thing = Stage (do { res <- thing; continueWith res })
@@ -622,7 +632,7 @@ get_sc_pending this_lvl ic@(IC { inert_dicts = dicts, inert_insts = insts })
        | otherwise
        = (cts, qci)
 
-    belongs_to_this_level ev = ctLocLevel (ctEvLoc ev) == this_lvl
+    belongs_to_this_level ev = ctLocLevel (ctEvLoc ev) `sameDepthAs` this_lvl
     -- We only want Givens from this level; see (3a) in
     -- Note [The superclass story] in GHC.Tc.Solver.Dict
 
@@ -682,9 +692,9 @@ getHasGivenEqs tclvl
 
              -- See Note [HasGivenEqs] in GHC.Tc.Types.Constraint, and
              -- Note [Tracking Given equalities] in GHC.Tc.Solver.InertSet
-             has_ge | ge_lvl == tclvl = MaybeGivenEqs
-                    | given_eqs       = LocalGivenEqs
-                    | otherwise       = NoGivenEqs
+             has_ge | ge_lvl `sameDepthAs` tclvl = MaybeGivenEqs
+                    | given_eqs                  = LocalGivenEqs
+                    | otherwise                  = NoGivenEqs
 
        ; traceTcS "getHasGivenEqs" $
          vcat [ text "given_eqs:" <+> ppr given_eqs
@@ -817,7 +827,8 @@ data TcSEnv
 
       tcs_unified     :: IORef Int,
          -- The number of unification variables we have filled
-         -- The important thing is whether it is non-zero
+         -- The important thing is whether it is non-zero, so it
+         -- could equally well be a Bool instead of an Int.
 
       tcs_unif_lvl  :: IORef (Maybe TcLevel),
          -- The Unification Level Flag
@@ -1310,6 +1321,9 @@ unifyTyVar tv ty
        ; TcM.updTcRef (tcs_unified env) (+1) }
 
 reportUnifications :: TcS a -> TcS (Int, a)
+-- Record how many unifications are done by thing_inside
+-- We could return a Bool instead of an Int;
+-- all that matters is whether it is no-zero
 reportUnifications (TcS thing_inside)
   = TcS $ \ env ->
     do { inner_unified <- TcM.newTcRef 0
@@ -1318,7 +1332,7 @@ reportUnifications (TcS thing_inside)
        ; TcM.updTcRef (tcs_unified env) (+ n_unifs)
        ; return (n_unifs, res) }
 
-getDefaultInfo ::  TcS ([Type], (Bool, Bool))
+getDefaultInfo ::  TcS (DefaultEnv, Bool)
 getDefaultInfo = wrapTcS TcM.tcGetDefaultTys
 
 getWorkList :: TcS WorkList
@@ -1693,7 +1707,7 @@ setWantedEq (HoleDest hole) co
 setWantedEq (EvVarDest ev) _ = pprPanic "setWantedEq: EvVarDest" (ppr ev)
 
 -- | Good for both equalities and non-equalities
-setWantedEvTerm :: TcEvDest -> Canonical -> EvTerm -> TcS ()
+setWantedEvTerm :: TcEvDest -> CanonicalEvidence -> EvTerm -> TcS ()
 setWantedEvTerm (HoleDest hole) _canonical tm
   | Just co <- evTermCoercion_maybe tm
   = do { useVars (coVarsOfCo co)
@@ -1701,7 +1715,7 @@ setWantedEvTerm (HoleDest hole) _canonical tm
   | otherwise
   = -- See Note [Yukky eq_sel for a HoleDest]
     do { let co_var = coHoleCoVar hole
-       ; setEvBind (mkWantedEvBind co_var True tm)
+       ; setEvBind (mkWantedEvBind co_var EvCanonical tm)
        ; fillCoercionHole hole (mkCoVarCo co_var) }
 
 setWantedEvTerm (EvVarDest ev_id) canonical tm
@@ -1731,7 +1745,7 @@ fillCoercionHole hole co
   = do { wrapTcS $ TcM.fillCoercionHole hole co
        ; kickOutAfterFillingCoercionHole hole }
 
-setEvBindIfWanted :: CtEvidence -> Canonical -> EvTerm -> TcS ()
+setEvBindIfWanted :: CtEvidence -> CanonicalEvidence -> EvTerm -> TcS ()
 setEvBindIfWanted ev canonical tm
   = case ev of
       CtWanted { ctev_dest = dest } -> setWantedEvTerm dest canonical tm
@@ -1763,11 +1777,13 @@ newBoundEvVarId pred rhs
        ; setEvBind (mkGivenEvBind new_ev rhs)
        ; return new_ev }
 
-emitNewGivens :: CtLoc -> [(Role,TcType,TcType,TcCoercion)] -> TcS ()
+emitNewGivens :: CtLoc -> [(Role,TcCoercion)] -> TcS ()
 emitNewGivens loc pts
-  = do { evs <- mapM (newGivenEvVar loc) $
+  = do { traceTcS "emitNewGivens" (ppr pts)
+       ; evs <- mapM (newGivenEvVar loc) $
                 [ (mkPrimEqPredRole role ty1 ty2, evCoercion co)
-                | (role, ty1, ty2, co) <- pts
+                | (role, co) <- pts
+                , let Pair ty1 ty2 = coercionKind co
                 , not (ty1 `tcEqType` ty2) ] -- Kill reflexive Givens at birth
        ; emitWorkNC evs }
 
@@ -1889,8 +1905,7 @@ solverDepthError loc ty
            do { ty   <- TcM.zonkTcType ty
               ; env0 <- TcM.tcInitTidyEnv
               ; return (ty, env0) }
-       ; let tidy_env     = tidyFreeTyCoVars env0 (tyCoVarsOfTypeList ty)
-             tidy_ty      = tidyType tidy_env ty
+       ; let (tidy_env, tidy_ty)  = tidyOpenTypeX env0 ty
              msg = TcRnSolverDepthError tidy_ty depth
        ; TcM.failWithTcM (tidy_env, msg) }
   where
@@ -2087,8 +2102,7 @@ checkTouchableTyVarEq
 --   with extra wanteds 'cts'
 -- If it returns (PuFail reason) we can't unify, and the reason explains why.
 checkTouchableTyVarEq ev lhs_tv rhs
-  | simpleUnifyCheck True lhs_tv rhs
-    -- True <=> type families are ok on the RHS
+  | simpleUnifyCheck UC_Solver lhs_tv rhs   -- An (optional) short-cut
   = do { traceTcS "checkTouchableTyVarEq: simple-check wins" (ppr lhs_tv $$ ppr rhs)
        ; return (pure (mkReflRedn Nominal rhs)) }
 
@@ -2121,12 +2135,13 @@ checkTouchableTyVarEq ev lhs_tv rhs
 
     flags = TEF { tef_foralls  = False -- isRuntimeUnkSkol lhs_tv
                 , tef_fam_app  = mkTEFA_Break ev NomEq break_wanted
-                , tef_unifying = Unifying lhs_tv_info lhs_tv_lvl LC_Promote
+                , tef_unifying = Unifying lhs_tv_info lhs_tv_lvl (LC_Promote False)
                 , tef_lhs      = TyVarLHS lhs_tv
                 , tef_occurs   = cteInsolubleOccurs }
 
     arg_flags = famAppArgFlags flags
 
+    break_wanted :: FamAppBreaker Ct
     break_wanted fam_app
       -- Occurs check or skolem escape; so flatten
       = do { let fam_app_kind = typeKind fam_app
