@@ -1,4 +1,5 @@
 {-# LANGUAGE AllowAmbiguousTypes     #-} -- for unXRec, etc.
+{-# LANGUAGE CPP                     #-}
 {-# LANGUAGE ConstraintKinds         #-}
 {-# LANGUAGE DataKinds               #-}
 {-# LANGUAGE DeriveDataTypeable      #-}
@@ -20,14 +21,18 @@ module Language.Haskell.Syntax.Extension where
 -- This module captures the type families to precisely identify the extension
 -- points for GHC.Hs syntax
 
-import GHC.Prelude
+#if MIN_VERSION_GLASGOW_HASKELL(9,3,0,0)
+import Data.Type.Equality (type (~))
+#endif
 
 import Data.Data hiding ( Fixity )
 import Data.Kind (Type)
-import GHC.Utils.Outputable
+
+import Data.Eq
+import Data.Ord
 
 {-
-Note [Trees that grow]
+Note [Trees That Grow]
 ~~~~~~~~~~~~~~~~~~~~~~
 
 See https://gitlab.haskell.org/ghc/ghc/wikis/implementing-trees-that-grow
@@ -61,9 +66,9 @@ See also Note [IsPass] and Note [NoGhcTc] in GHC.Hs.Extension.
 -}
 
 -- | A placeholder type for TTG extension points that are not currently
--- unused to represent any particular value.
+-- used to represent any particular value.
 --
--- This should not be confused with 'NoExtCon', which are found in unused
+-- This should not be confused with 'DataConCantHappen', which are found in unused
 -- extension /constructors/ and therefore should never be inhabited. In
 -- contrast, 'NoExtField' is used in extension /points/ (e.g., as the field of
 -- some constructor), so it must have an inhabitant to construct AST passes
@@ -71,31 +76,44 @@ See also Note [IsPass] and Note [NoGhcTc] in GHC.Hs.Extension.
 data NoExtField = NoExtField
   deriving (Data,Eq,Ord)
 
-instance Outputable NoExtField where
-  ppr _ = text "NoExtField"
-
 -- | Used when constructing a term with an unused extension point.
 noExtField :: NoExtField
 noExtField = NoExtField
 
--- | Used in TTG extension constructors that have yet to be extended with
--- anything. If an extension constructor has 'NoExtCon' as its field, it is
--- not intended to ever be constructed anywhere, and any function that consumes
--- the extension constructor can eliminate it by way of 'noExtCon'.
---
--- This should not be confused with 'NoExtField', which are found in unused
--- extension /points/ (not /constructors/) and therefore can be inhabited.
+{-
+Note [Constructor cannot occur]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+Some data constructors can't occur in certain phases; e.g. the output
+of the type checker never has OverLabel. We signal this by
+* setting the extension field to DataConCantHappen
+* using dataConCantHappen in the cases that can't happen
 
--- See also [NoExtCon and strict fields].
-data NoExtCon
+For example:
+
+   type instance XOverLabel GhcTc = DataConCantHappen
+
+   dsExpr :: HsExpr GhcTc -> blah
+   dsExpr (HsOverLabel x _) = dataConCantHappen x
+
+The function dataConCantHappen is defined thus:
+   dataConCantHappen :: DataConCantHappen -> a
+   dataConCantHappen x = case x of {}
+(i.e. identically to Data.Void.absurd, but more helpfully named).
+Remember DataConCantHappen is a type whose only element is bottom.
+
+This should not be confused with 'NoExtField', which are found in unused
+extension /points/ (not /constructors/) and therefore can be inhabited.
+
+It would be better to omit the pattern match altogether, but we
+can only do that if the extension field was strict (#18764).
+See also [DataConCantHappen and strict fields].
+-}
+data DataConCantHappen
   deriving (Data,Eq,Ord)
 
-instance Outputable NoExtCon where
-  ppr = noExtCon
-
--- | Eliminate a 'NoExtCon'. Much like 'Data.Void.absurd'.
-noExtCon :: NoExtCon -> a
-noExtCon x = case x of {}
+-- | Eliminate a 'DataConCantHappen'. See Note [Constructor cannot occur].
+dataConCantHappen :: DataConCantHappen -> a
+dataConCantHappen x = case x of {}
 
 -- | GHC's L prefixed variants wrap their vanilla variant in this type family,
 -- to add 'SrcLoc' info via 'Located'. Other passes than 'GhcPass' not
@@ -170,14 +188,9 @@ type family XXValBindsLR x x'
 type family XFunBind    x x'
 type family XPatBind    x x'
 type family XVarBind    x x'
-type family XAbsBinds   x x'
 type family XPatSynBind x x'
 type family XXHsBindsLR x x'
 
--- ABExport type families
-type family XABE x
-type family XXABExport x
-
 -- PatSynBind type families
 type family XPSB x x'
 type family XXPatSynBind x x'
@@ -338,6 +351,10 @@ type family XXDefaultDecl      x
 type family XForeignImport     x
 type family XForeignExport     x
 type family XXForeignDecl      x
+type family XCImport           x
+type family XXForeignImport    x
+type family XCExport           x
+type family XXForeignExport    x
 
 -- -------------------------------------
 -- RuleDecls type families
@@ -380,13 +397,18 @@ type family XXRoleAnnotDecl  x
 type family XCInjectivityAnn  x
 type family XXInjectivityAnn  x
 
+-- =====================================================================
+-- Type families for the HsModule extension points
+
+type family XCModule x
+type family XXModule x
+
 -- =====================================================================
 -- Type families for the HsExpr extension points
 
 type family XVar            x
 type family XUnboundVar     x
-type family XConLikeOut     x
-type family XRecFld         x
+type family XRecSel         x
 type family XOverLabel      x
 type family XIPVar          x
 type family XOverLitE       x
@@ -410,38 +432,41 @@ type family XDo             x
 type family XExplicitList   x
 type family XRecordCon      x
 type family XRecordUpd      x
+type family XLHsRecUpdLabels x
+type family XLHsOLRecUpdLabels x
 type family XGetField       x
 type family XProjection     x
 type family XExprWithTySig  x
 type family XArithSeq       x
-type family XBracket        x
-type family XRnBracketOut   x
-type family XTcBracketOut   x
-type family XSpliceE        x
+type family XTypedBracket   x
+type family XUntypedBracket x
+type family XTypedSplice    x
+type family XUntypedSplice  x
 type family XProc           x
 type family XStatic         x
 type family XTick           x
 type family XBinTick        x
 type family XPragE          x
+type family XEmbTy          x
+type family XForAll         x
+type family XQual           x
+type family XFunArr         x
 type family XXExpr          x
 
 -- -------------------------------------
--- FieldLabel type families
-type family XCHsFieldLabel  x
-type family XXHsFieldLabel  x
+-- HsMatchContext type families
+type family XFunRhs
+
+-- -------------------------------------
+-- DotFieldOcc type families
+type family XCDotFieldOcc  x
+type family XXDotFieldOcc  x
 
 -- -------------------------------------
 -- HsPragE type families
 type family XSCC            x
 type family XXPragE         x
 
-
--- -------------------------------------
--- AmbiguousFieldOcc type families
-type family XUnambiguous        x
-type family XAmbiguous          x
-type family XXAmbiguousFieldOcc x
-
 -- -------------------------------------
 -- HsTupArg type families
 type family XPresent  x
@@ -449,23 +474,20 @@ type family XMissing  x
 type family XXTupArg  x
 
 -- -------------------------------------
--- HsSplice type families
-type family XTypedSplice   x
-type family XUntypedSplice x
-type family XQuasiQuote    x
-type family XSpliced       x
-type family XXSplice       x
+-- HsUntypedSplice type families
+type family XUntypedSpliceExpr x
+type family XQuasiQuote        x
+type family XXUntypedSplice    x
 
 -- -------------------------------------
--- HsBracket type families
-type family XExpBr      x
-type family XPatBr      x
-type family XDecBrL     x
-type family XDecBrG     x
-type family XTypBr      x
-type family XVarBr      x
-type family XTExpBr     x
-type family XXBracket   x
+-- HsQuoteBracket type families
+type family XExpBr  x
+type family XPatBr  x
+type family XDecBrL x
+type family XDecBrG x
+type family XTypBr  x
+type family XVarBr  x
+type family XXQuote x
 
 -- -------------------------------------
 -- HsCmdTop type families
@@ -496,7 +518,6 @@ type family XXGRHS  x b
 -- StmtLR type families
 type family XLastStmt        x x' b
 type family XBindStmt        x x' b
-type family XApplicativeStmt x x' b
 type family XBodyStmt        x x' b
 type family XLetStmt         x x' b
 type family XParStmt         x x' b
@@ -506,35 +527,24 @@ type family XXStmtLR         x x' b
 
 -- -------------------------------------
 -- HsCmd type families
-type family XCmdArrApp  x
-type family XCmdArrForm x
-type family XCmdApp     x
-type family XCmdLam     x
-type family XCmdPar     x
-type family XCmdCase    x
-type family XCmdLamCase x
-type family XCmdIf      x
-type family XCmdLet     x
-type family XCmdDo      x
-type family XCmdWrap    x
-type family XXCmd       x
+type family XCmdArrApp   x
+type family XCmdArrForm  x
+type family XCmdApp      x
+type family XCmdLam      x
+type family XCmdPar      x
+type family XCmdCase     x
+type family XCmdLamCase  x
+type family XCmdIf       x
+type family XCmdLet      x
+type family XCmdDo       x
+type family XCmdWrap     x
+type family XXCmd        x
 
 -- -------------------------------------
 -- ParStmtBlock type families
 type family XParStmtBlock  x x'
 type family XXParStmtBlock x x'
 
--- -------------------------------------
--- ApplicativeArg type families
-type family XApplicativeArgOne   x
-type family XApplicativeArgMany  x
-type family XXApplicativeArg     x
-
--- =====================================================================
--- Type families for the HsImpExp extension points
-
--- TODO
-
 -- =====================================================================
 -- Type families for the HsLit extension points
 
@@ -543,11 +553,18 @@ type family XXApplicativeArg     x
 type family XHsChar x
 type family XHsCharPrim x
 type family XHsString x
+type family XHsMultilineString x
 type family XHsStringPrim x
 type family XHsInt x
 type family XHsIntPrim x
 type family XHsWordPrim x
+type family XHsInt8Prim x
+type family XHsInt16Prim x
+type family XHsInt32Prim x
 type family XHsInt64Prim x
+type family XHsWord8Prim x
+type family XHsWord16Prim x
+type family XHsWord32Prim x
 type family XHsWord64Prim x
 type family XHsInteger x
 type family XHsRat x
@@ -563,25 +580,28 @@ type family XXOverLit x
 -- =====================================================================
 -- Type families for the HsPat extension points
 
-type family XWildPat    x
-type family XVarPat     x
-type family XLazyPat    x
-type family XAsPat      x
-type family XParPat     x
-type family XBangPat    x
-type family XListPat    x
-type family XTuplePat   x
-type family XSumPat     x
-type family XConPat     x
-type family XViewPat    x
-type family XSplicePat  x
-type family XLitPat     x
-type family XNPat       x
-type family XNPlusKPat  x
-type family XSigPat     x
-type family XCoPat      x
-type family XXPat       x
-type family XHsRecField x
+type family XWildPat     x
+type family XVarPat      x
+type family XLazyPat     x
+type family XAsPat       x
+type family XParPat      x
+type family XBangPat     x
+type family XListPat     x
+type family XTuplePat    x
+type family XSumPat      x
+type family XOrPat       x
+type family XConPat      x
+type family XViewPat     x
+type family XSplicePat   x
+type family XLitPat      x
+type family XNPat        x
+type family XNPlusKPat   x
+type family XSigPat      x
+type family XEmbTyPat    x
+type family XInvisPat    x
+type family XCoPat       x
+type family XXPat        x
+type family XHsFieldBind x
 
 -- =====================================================================
 -- Type families for the HsTypes type families
@@ -613,6 +633,11 @@ type family XXHsWildCardBndrs  x b
 type family XHsPS x
 type family XXHsPatSigType x
 
+-- -------------------------------------
+-- HsTyPat type families
+type family XHsTP x
+type family XXHsTyPat x
+
 -- -------------------------------------
 -- HsType type families
 type family XForAllTy        x
@@ -639,6 +664,13 @@ type family XTyLit           x
 type family XWildCardTy      x
 type family XXType           x
 
+-- ---------------------------------------------------------------------
+-- HsTyLit type families
+type family XNumTy           x
+type family XStrTy           x
+type family XCharTy          x
+type family XXTyLit          x
+
 -- ---------------------------------------------------------------------
 -- HsForAllTelescope type families
 type family XHsForAllVis        x
@@ -647,8 +679,7 @@ type family XXHsForAllTelescope x
 
 -- ---------------------------------------------------------------------
 -- HsTyVarBndr type families
-type family XUserTyVar   x
-type family XKindedTyVar x
+type family XTyVarBndr   x
 type family XXTyVarBndr  x
 
 -- ---------------------------------------------------------------------
@@ -662,12 +693,13 @@ type family XCFieldOcc x
 type family XXFieldOcc x
 
 -- =====================================================================
--- Type families for the HsImpExp type families
+-- Type families for the HsImpExp extension points
 
 -- -------------------------------------
 -- ImportDecl type families
 type family XCImportDecl       x
 type family XXImportDecl       x
+type family ImportDeclPkgQual  x -- stores the package qualifier in an import statement
 
 -- -------------------------------------
 -- IE type families
@@ -682,6 +714,14 @@ type family XIEDocNamed        x
 type family XXIE               x
 
 -- -------------------------------------
+-- IEWrappedName type families
+type family XIEName p
+type family XIEDefault p
+type family XIEPattern p
+type family XIEType p
+type family XXIEWrappedName p
+
+
 
 -- =====================================================================
 -- Misc
