@@ -41,12 +41,13 @@ nitpick_ignore = [
 ]
 
 rst_prolog = """
-.. |llvm-version| replace:: {llvm_version}
-""".format(llvm_version=ghc_config.llvm_version)
+.. |llvm-version-min| replace:: {llvm_version_min}
+.. |llvm-version-max| replace:: {llvm_version_max}
+""".format(llvm_version_min=ghc_config.llvm_version_min, llvm_version_max=ghc_config.llvm_version_max)
 
 # General information about the project.
 project = u'Glasgow Haskell Compiler'
-copyright = u'2020, GHC Team'
+copyright = u'2023, GHC Team'
 # N.B. version comes from ghc_config
 release = version  # The full version, including alpha/beta/rc tags.
 
@@ -224,7 +225,7 @@ def haddock_role(lib):
       # for the template_haskell.rst example this will be '..'
       current_doc_to_topdir = os.path.relpath(topdir, os.path.dirname(current_doc))
 
-      relative_path = '%s/%s/%s-%s' % (current_doc_to_topdir, libs_base_uri, lib, lib_version)
+      relative_path = '%s/%s/%s' % (current_doc_to_topdir, libs_base_uri, lib_version)
 
       uri = '%s/%s.html%s' % (relative_path, module, anchor)
 
@@ -276,7 +277,6 @@ def setup(app):
     app.add_role('cabal-ref', haddock_role('Cabal'))
     app.add_role('ghc-compact-ref', haddock_role('ghc-compact'))
     app.add_role('ghc-prim-ref', haddock_role('ghc-prim'))
-    app.add_role('parallel-ref', haddock_role('parallel'))
     app.add_role('array-ref', haddock_role('array'))
 
     app.add_object_type('rts-flag', 'rts-flag',
