@@ -17,6 +17,7 @@ import Development.Shake.Classes
 import Development.Shake.Command
 import Development.Shake.FilePath
 import GHC.Generics
+import GHC.Platform.ArchOS (ArchOS(..), Arch(..))
 import qualified Hadrian.Builder as H
 import Hadrian.Builder hiding (Builder)
 import Hadrian.Builder.Ar
@@ -33,6 +34,7 @@ import Base
 import Context
 import Oracles.Flag
 import Oracles.Setting (setting, Setting(..))
+import Oracles.Setting (settingsFileSetting, ToolchainSetting(..))
 import Packages
 
 import GHC.IO.Encoding (getFileSystemEncoding)
@@ -163,7 +165,7 @@ data Builder = Alex
              | Cc CcMode Stage
              | Configure FilePath
              | DeriveConstants
-             | GenApply
+             | GenApply (Maybe Int) -- ^ vector size, or Nothing for non-vectors
              | GenPrimopCode
              | Ghc GhcMode Stage
              | GhcPkg GhcPkgMode Stage
@@ -172,6 +174,7 @@ data Builder = Alex
              | Hp2Ps
              | Hpc
              | HsCpp
+             | JsCpp
              | Hsc2Hs Stage
              | Ld Stage --- ^ linker
              | Make FilePath
@@ -201,7 +204,7 @@ instance NFData   Builder
 builderProvenance :: Builder -> Maybe Context
 builderProvenance = \case
     DeriveConstants  -> context stage0Boot deriveConstants
-    GenApply         -> context stage0Boot genapply
+    GenApply {}      -> context stage0Boot genapply
     GenPrimopCode    -> context stage0Boot genprimopcode
     Ghc _ (Stage0 {})-> Nothing
     Ghc _ stage      -> context (predStage stage) ghc
@@ -234,15 +237,25 @@ instance H.Builder Builder where
           -- changes (#18001).
           _bootGhcVersion <- setting GhcVersion
           pure []
-        Ghc {} -> do
+        Ghc _ st -> do
             root <- buildRoot
             unlitPath  <- builderPath Unlit
+            distro_mingw <- settingsFileSetting ToolchainSetting_DistroMinGW
+            libffi_adjustors <- useLibffiForAdjustors
+            use_system_ffi <- flag UseSystemFfi
 
             return $ [ unlitPath ]
-                  ++ [ root -/- mingwStamp | windowsHost ]
+                  ++ [ root -/- mingwStamp | windowsHost, distro_mingw == "NO" ]
                      -- proxy for the entire mingw toolchain that
                      -- we have in inplace/mingw initially, and then at
                      -- root -/- mingw.
+                  -- ffi.h needed by the compiler when using libffi_adjustors (#24864)
+                  -- It would be nicer to not duplicate this logic between here
+                  -- and needRtsLibffiTargets and libffiHeaderFiles but this doesn't change
+                  -- very often.
+                  ++ [ root -/- buildDir (rtsContext st) -/- "include" -/- header
+                     | header <- ["ffi.h", "ffitarget.h"]
+                     , libffi_adjustors && not use_system_ffi ]
 
         Hsc2Hs stage -> (\p -> [p]) <$> templateHscPath stage
         Make dir  -> return [dir -/- "Makefile"]
@@ -324,7 +337,7 @@ instance H.Builder Builder where
                     let env = AddEnv "CONFIG_SHELL" bash
                     cmd' env [Cwd dir] ["sh", path] buildOptions buildArgs
 
-                GenApply -> captureStdout
+                GenApply {} -> captureStdout
 
                 GenPrimopCode -> do
                     need [input]
@@ -406,6 +419,7 @@ isOptional target = \case
     Alex     -> True
     -- Most ar implemententions no longer need ranlib, but some still do
     Ranlib   -> not $ Toolchain.arNeedsRanlib (tgtAr target)
+    JsCpp    -> not $ (archOS_arch . tgtArchOs) target == ArchJavaScript -- ArchWasm32 too?
     _        -> False
 
 -- | Determine the location of a system 'Builder'.
@@ -421,6 +435,7 @@ systemBuilderPath builder = case builder of
     GhcPkg _ (Stage0 {}) -> fromKey "system-ghc-pkg"
     Happy           -> fromKey "happy"
     HsCpp           -> fromTargetTC "hs-cpp" (Toolchain.hsCppProgram . tgtHsCPreprocessor)
+    JsCpp           -> fromTargetTC "js-cpp" (maybeProg Toolchain.jsCppProgram . tgtJsCPreprocessor)
     Ld _            -> fromTargetTC "ld" (Toolchain.ccLinkProgram . tgtCCompilerLink)
     -- MergeObjects Stage0 is a special case in case of
     -- cross-compiling. We're building stage1, e.g. code which will be
