@@ -29,8 +29,6 @@ sourceDistRules = alternatives $ do
       archiveSourceTree prepareTree
     "GIT_COMMIT_ID" %> \fname ->
         writeFileChanged fname =<< setting ProjectGitCommitId
-    "VERSION" %> \fname ->
-        writeFileChanged fname =<< setting ProjectVersion
 
     -- Rules to download mingw tarballs
     let mingw_tarballs_stamp = "ghc-tarballs/mingw-w64/.mingw-w64.download.stamp"
@@ -57,8 +55,9 @@ archiveSourceTree prepare fname = do
   runBuilderWithCmdOptions
       [Cwd $ sourceDistRoot -/- dirName]
       (Tar Create)
-      ["chJf", ".." -/- tarName,  baseName]
-      ["chJf", ".." -/- tarName] [baseName]
+      -- See https://github.com/haskell/cabal/issues/10442 for why we exclude this file.
+      ["--exclude=cabal.project.symlink.broken","-chJf", ".." -/- tarName,  baseName]
+      ["--exclude=cabal.project.symlink.broken","-chJf", ".." -/- tarName] [baseName]
 
 
 -- | This creates a symlink to the 'source' at 'target'
@@ -76,7 +75,9 @@ copyFileSourceDist source target = do
       error ("source-dist: tried to create non-relative symlink in source dist: " ++ show link_target)
     putProgressInfo =<< renderAction ("Create symlink (" ++ link_target ++ ")") source target
     isDirectory <- liftIO $ IO.doesDirectoryExist source
-    when (not isDirectory) $
+    -- We don't want to call `need` on broken symlinks
+    linkTargetExists <- liftIO $ IO.doesPathExist link_target
+    when (not isDirectory && linkTargetExists) $
       need [source]
     let createLink src tgt
           | isDirectory = liftIO $ IO.createDirectoryLink src tgt
@@ -120,22 +121,55 @@ prepareTree dest = do
     out <- askWithResources [] (target (vanillaContext Stage1 compiler) (Git ListFiles) [] [])
     top <- topDirectory
     let files = ["GIT_COMMIT_ID", "VERSION"] ++ getFiles out
-    need ["GIT_COMMIT_ID", "VERSION"]
+    need ["GIT_COMMIT_ID"]
     forM_ files $ \source -> do
       let target = dest -/- source
       copyFileSourceDist (top -/- source) target
     copyAlexHappyFiles
+    copyBootFiles
   where
 
     getFiles = filter treeFilter . split (=='\NUL')
     treeFilter file = not (null file) && not ("testsuite//" ?== file)
 
+    -- Copy files created by running ./boot
+    copyBootFiles = do
+      top <- topDirectory
+      forM_ bootFiles $ \file -> do
+        let src_file = top -/- file
+            dest_file = top -/- dest -/- file
+        createFileLink src_file dest_file
+      -- And move ./boot so we can't accidentally call it in CI
+      moveFile (dest -/- "boot") (dest -/- "boot.source")
+
+    bootFiles =
+      [ pkgPath rts -/- "configure"
+      , pkgPath rts -/- "ghcautoconf.h.autoconf.in"
+      , pkgPath process -/- "include" -/- "HsProcessConfig.h.in"
+      , pkgPath process -/- "configure"
+      , pkgPath ghcBignum -/- "configure"
+      , pkgPath ghcInternal -/- "configure"
+      , pkgPath ghcInternal -/- "include" -/- "HsBaseConfig.h.in"
+      , pkgPath directory -/- "configure"
+      , pkgPath directory -/- "HsDirectoryConfig.h.in"
+      , pkgPath time -/- "configure"
+      , pkgPath time -/- "lib" -/- "include" -/- "HsTimeConfig.h.in"
+      , pkgPath unix -/- "configure"
+      , pkgPath unix -/- "include" -/- "HsUnixConfig.h.in"
+      , pkgPath terminfo -/- "configure"
+      , "configure"
+      , "aclocal.m4"
+      ]
+
     copyAlexHappyFiles =
       forM_ alexHappyFiles $ \(stg, pkg, inp, out) -> do
-        let ctx = Context stg pkg vanilla
+        let ctx = Context stg pkg vanilla Inplace
             srcInputFile = dest -/- pkgPath pkg -/- inp
             generatedFile = dest -/- pkgPath pkg -/- out
-            Just builder = determineBuilder stg inp
+            builder =
+                case determineBuilder stg inp of
+                  Just builder -> builder
+                  Nothing -> error $ "Failed to determine builder for " ++ inp
 
         -- We first make sure that the generated file is... generated.
         build $ target ctx builder [srcInputFile] [generatedFile]
@@ -148,11 +182,13 @@ prepareTree dest = do
 
     -- (stage, package, input file, output file)
     alexHappyFiles =
-        [ (Stage0, compiler,      "GHC/Cmm/Parser.y",   "GHC/Cmm/Parser.hs")
-        , (Stage0, compiler,      "GHC/Cmm/Lexer.x",    "GHC/Cmm/Lexer.hs")
-        , (Stage0, compiler,      "GHC/Parser.y",       "GHC/Parser.hs")
-        , (Stage0, compiler,      "GHC/Parser/Lexer.x", "GHC/Parser/Lexer.hs")
-        , (Stage0, hpcBin,        "HpcParser.y",        "HpcParser.hs")
-        , (Stage0, genprimopcode, "Parser.y",           "Parser.hs")
-        , (Stage0, genprimopcode, "Lexer.x",            "Lexer.hs")
+        [ (stage0InTree , compiler,      "GHC/Cmm/Parser.y",   "GHC/Cmm/Parser.hs")
+        , (stage0InTree , compiler,      "GHC/Cmm/Lexer.x",    "GHC/Cmm/Lexer.hs")
+        , (stage0InTree , compiler,      "GHC/Parser.y",       "GHC/Parser.hs")
+        , (stage0InTree , compiler,      "GHC/Parser/Lexer.x", "GHC/Parser/Lexer.hs")
+        , (stage0InTree , compiler,      "GHC/Parser/HaddockLex.x", "GHC/Parser/HaddockLex.hs")
+        , (stage0InTree , hpcBin,        "src/Trace/Hpc/Parser.y", "src/Trace/Hpc/Parser.hs")
+        , (stage0InTree , genprimopcode, "Parser.y",           "Parser.hs")
+        , (stage0InTree , genprimopcode, "Lexer.x",            "Lexer.hs")
+        , (stage0InTree , cabalSyntax  , "src/Distribution/Fields/Lexer.x",  "src/Distribution/Fields/Lexer.hs")
         ]
