@@ -1,13 +1,16 @@
 module Settings.Builders.DeriveConstants (
-    deriveConstantsBuilderArgs, deriveConstantsPairs
+    deriveConstantsBuilderArgs
     ) where
 
 import Builder
 import Settings.Builders.Common
+import GHC.Toolchain (tgtCCompiler, ccProgram, tgtUnregisterised)
+import GHC.Toolchain.Program
 
 deriveConstantsPairs :: [(String, String)]
 deriveConstantsPairs =
-  [ ("DerivedConstants.h", "--gen-header")
+  [ ("Constants.hs", "--gen-haskell-type")
+  , ("DerivedConstants.h", "--gen-header")
   ]
 
 deriveConstantsBuilderArgs :: Args
@@ -31,18 +34,18 @@ deriveConstantsBuilderArgs = builder DeriveConstants ? do
         , arg "--nm-program", arg =<< getBuilderPath Nm
         , isSpecified Objdump ? mconcat [ arg "--objdump-program"
                                         , arg =<< getBuilderPath Objdump ]
-        , arg "--target-os", arg =<< getSetting TargetOs ]
+        , arg "--target-os", arg =<< queryTarget queryOS ]
 
 includeCcArgs :: Args
 includeCcArgs = do
     stage <- getStage
-    libPath <- expr $ stageLibPath stage
+    rtsPath <- expr $ rtsBuildPath stage
     mconcat [ cArgs
             , cWarnings
-            , getSettingList $ ConfCcArgs Stage1
-            , flag GhcUnregisterised ? arg "-DUSE_MINIINTERPRETER"
+            , prgFlags . ccProgram . tgtCCompiler <$> expr (targetStage Stage1)
+            , queryTargetTarget tgtUnregisterised ? arg "-DUSE_MINIINTERPRETER"
             , arg "-Irts"
-            , arg "-Iincludes"
-            , arg $ "-I" ++ libPath
+            , arg "-Irts/include"
+            , arg $ "-I" ++ rtsPath </> "include"
             , notM targetSupportsSMP ? arg "-DNOSMP"
             , arg "-fcommon" ]
