@@ -80,40 +80,41 @@ stageBootPackages = return
 stage0Packages :: Action [Package]
 stage0Packages = do
     cross <- flag CrossCompiling
-    return $ [ binary
-             , bytestring
-             , cabalSyntax
+    return $ [ cabalSyntax
              , cabal
              , compiler
-             , containers
-             , directory
-             , process
-             , exceptions
-             , filepath
+             , directory -- depends on filepath, fileIo
+             , filepath -- depends on os-string
+             , fileio
              , ghc
-             , runGhc
              , ghcBoot
-             , ghcBootTh
-             , ghcPlatform
+             , ghcBootThNext
              , ghcHeap
+             , ghcPkg
+             , ghcPlatform
              , ghcToolchain
              , ghci
-             , ghcPkg
+             , haskeline
+             , haddockApi
+             , haddockLibrary
              , haddock
-             , hsc2hs
+             , hp2ps
              , hpc
              , hpcBin
-             , mtl
-             , osString
-             , parsec
-             , semaphoreCompat
-             , time
-             , templateHaskell
-             , text
-             , transformers
+             , hsc2hs
+             , osString -- new library not yet present for boot compilers
+             , process -- depends on filepath
+             , runGhc
+             , semaphoreCompat -- depends on
+             , time -- depends on win32
              , unlit
-             , hp2ps
              , if windowsHost then win32 else unix
+             -- We must use the in-tree `Win32` as the version
+             -- bundled with GHC 9.6 is too old for `semaphore-compat`.
+             -- Once 9.6 is no longer a boot compiler, we can drop win32/unix.
+             -- These depend on `filepath`/`os-string` through an automatic flag
+             -- that confused Hadrian, so we must make those a stage0 package as well.
+             -- Once we drop `Win32`/`unix` it should be possible to drop those too.
              ]
           ++ [ terminfo | not windowsHost, not cross ]
           ++ [ timeout  | windowsHost                ]
@@ -127,6 +128,9 @@ stage1Packages = do
           -- but not win32/unix because it depends on cross-compilation target
           | p == win32        = False
           | p == unix         = False
+          -- These packages are only needed for bootstrapping.
+          -- See Note [Bootstrapping Template Haskell]
+          | p == ghcBootThNext = False
           | otherwise         = True
 
     libraries0 <- filter good_stage0_package <$> stage0Packages
@@ -139,11 +143,14 @@ stage1Packages = do
       [ libraries0 -- Build all Stage0 libraries in Stage1
       , [ array
         , base
+        , binary
+        , bytestring
         , containers
         , deepseq
         , exceptions
         , ghc
         , ghcBignum
+        , ghcBootTh
         , ghcCompact
         , ghcExperimental
         , ghcInternal
@@ -153,10 +160,15 @@ stage1Packages = do
         , hp2ps
         , hsc2hs
         , integerGmp
+        , mtl
+        , parsec
         , pretty
         , rts
         , semaphoreCompat
         , stm
+        , templateHaskell
+        , text
+        , transformers
         , unlit
         , xhtml
         , if winTarget then win32 else unix
@@ -191,7 +203,7 @@ defaultLibraryWays = Set.fromList <$>
     mconcat
     [ pure [vanilla]
     , notStage0 ? pure [profiling]
-    , notStage0 ? platformSupportsSharedLibs ? pure [dynamic]
+    , notStage0 ? platformSupportsSharedLibs ? pure [dynamic, profilingDynamic]
     ]
 
 -- | Default build ways for the RTS.
@@ -205,9 +217,11 @@ defaultRtsWays = Set.fromList <$>
       ]
   , notStage0 ? targetSupportsThreadedRts ? pure [threaded, threadedProfiling, threadedDebugProfiling, threadedDebug]
   , notStage0 ? platformSupportsSharedLibs ? pure
-      [ dynamic, debugDynamic
+      [ dynamic, profilingDynamic, debugDynamic, debugProfilingDynamic
+      ]
+  , notStage0 ? platformSupportsSharedLibs ? targetSupportsThreadedRts ? pure
+      [ threadedDynamic, threadedDebugDynamic, threadedProfilingDynamic, threadedDebugProfilingDynamic
       ]
-  , notStage0 ? platformSupportsSharedLibs ? targetSupportsThreadedRts ? pure [ threadedDynamic, threadedDebugDynamic ]
   ]
 
 -- TODO: Move C source arguments here
@@ -259,6 +273,7 @@ defaultFlavour = Flavour
     , packages           = defaultPackages
     , bignumBackend      = defaultBignumBackend
     , bignumCheck        = False
+    , textWithSIMDUTF    = False
     , libraryWays        = defaultLibraryWays
     , rtsWays            = defaultRtsWays
     , dynamicGhcPrograms = defaultDynamicGhcPrograms
