@@ -1,3 +1,5 @@
+{-# LANGUAGE LambdaCase #-}
+
 -----------------------------------------------------------------------------
 
 -- Module      :  Distribution.Simple.Errors
@@ -20,18 +22,15 @@ import Distribution.Compiler
 import Distribution.InstalledPackageInfo
 import Distribution.ModuleName
 import Distribution.Package
-import Distribution.PackageDescription (FlagName, UnqualComponentName)
+import Distribution.PackageDescription
 import Distribution.Pretty
   ( Pretty (pretty)
   , prettyShow
   )
 import Distribution.Simple.InstallDirs
 import Distribution.Simple.PreProcess.Types (Suffix)
+import Distribution.Simple.SetupHooks.Errors
 import Distribution.System (OS)
-import Distribution.Types.BenchmarkType
-import Distribution.Types.LibraryName
-import Distribution.Types.PkgconfigVersion
-import Distribution.Types.TestType
 import Distribution.Types.VersionRange.Internal ()
 import Distribution.Version
 import Text.PrettyPrint
@@ -48,7 +47,8 @@ data CabalException
   | EnableBenchMark
   | BenchMarkNameDisabled String
   | NoBenchMark String
-  | NoLibraryFound
+  | -- | @NoLibraryFound@ has been downgraded to a warning, and is therefore no longer emitted.
+    NoLibraryFound
   | CompilerNotInstalled CompilerFlavor
   | CantFindIncludeFile String
   | UnsupportedTestSuite String
@@ -115,7 +115,7 @@ data CabalException
   | CheckSemaphoreSupport
   | NoLibraryForPackage
   | SanityCheckHookedBuildInfo UnqualComponentName
-  | ConfigureScriptNotFound
+  | ConfigureScriptNotFound FilePath
   | NoValidComponent
   | ConfigureEitherSingleOrAll
   | ConfigCIDValidForPreComponent
@@ -138,7 +138,7 @@ data CabalException
   | BadVersion String String PkgconfigVersion
   | UnknownCompilerException
   | NoWorkingGcc
-  | NoOSSupport OS
+  | NoOSSupport OS String
   | NoCompilerSupport String
   | InstallDirsNotPrefixRelative (InstallDirs FilePath)
   | ExplainErrors (Maybe (Either [Char] [Char])) [String]
@@ -152,8 +152,7 @@ data CabalException
   | Couldn'tFindTestProgLibV09 FilePath
   | TestCoverageSupportLibV09
   | RawSystemStdout String
-  | FindFileCwd FilePath
-  | FindFileEx FilePath
+  | FindFile FilePath
   | FindModuleFileEx ModuleName [Suffix] [FilePath]
   | MultipleFilesWithExtension String
   | NoDesc
@@ -172,6 +171,7 @@ data CabalException
   | BadVersionDb String Version VersionRange FilePath
   | UnknownVersionDb String VersionRange FilePath
   | MissingCoveredInstalledLibrary UnitId
+  | SetupHooksException SetupHooksException
   deriving (Show, Typeable)
 
 exceptionCode :: CabalException -> Int
@@ -284,8 +284,7 @@ exceptionCode e = case e of
   Couldn'tFindTestProgLibV09{} -> 9012
   TestCoverageSupportLibV09{} -> 1076
   RawSystemStdout{} -> 3098
-  FindFileCwd{} -> 4765
-  FindFileEx{} -> 2115
+  FindFile{} -> 2115
   FindModuleFileEx{} -> 6663
   MultipleFilesWithExtension{} -> 3333
   NoDesc{} -> 7654
@@ -304,6 +303,8 @@ exceptionCode e = case e of
   BadVersionDb{} -> 8038
   UnknownVersionDb{} -> 1008
   MissingCoveredInstalledLibrary{} -> 9341
+  SetupHooksException err ->
+    setupHooksExceptionCode err
 
 versionRequirement :: VersionRange -> String
 versionRequirement range
@@ -313,13 +314,13 @@ versionRequirement range
 exceptionMessage :: CabalException -> String
 exceptionMessage e = case e of
   NoBenchMarkProgram cmd -> "Could not find benchmark program \"" ++ cmd ++ "\". Did you build the package first?"
-  EnableBenchMark -> "No benchmarks enabled. Did you remember to configure with " ++ "\'--enable-benchmarks\'?"
+  EnableBenchMark -> "No benchmarks enabled. Did you remember to \'Setup configure\' with " ++ "\'--enable-benchmarks\'?"
   BenchMarkNameDisabled bmName -> "Package configured with benchmark " ++ bmName ++ " disabled."
   NoBenchMark bmName -> "no such benchmark: " ++ bmName
   NoLibraryFound -> "No executables and no library found. Nothing to do."
   CompilerNotInstalled compilerFlavor -> "installing with " ++ prettyShow compilerFlavor ++ "is not implemented"
   CantFindIncludeFile file -> "can't find include file " ++ file
-  UnsupportedTestSuite testType -> "Unsupported test suite type: " ++ testType
+  UnsupportedTestSuite test_type -> "Unsupported test suite type: " ++ test_type
   UnsupportedBenchMark benchMarkType -> "Unsupported benchmark type: " ++ benchMarkType
   NoIncludeFileFound f -> "can't find include file " ++ f
   NoModuleFound m suffixes ->
@@ -330,8 +331,8 @@ exceptionMessage e = case e of
       ++ ".\n"
       ++ "If the module "
       ++ "is autogenerated it should be added to 'autogen-modules'."
-  RegMultipleInstancePkg -> "HcPkg.register: the compiler does not support,registering multiple instances of packages."
-  SuppressingChecksOnFile -> "HcPkg.register: the compiler does not support ,suppressing checks on files."
+  RegMultipleInstancePkg -> "HcPkg.register: the compiler does not support registering multiple instances of packages."
+  SuppressingChecksOnFile -> "HcPkg.register: the compiler does not support suppressing checks on files."
   NoSupportDirStylePackageDb -> "HcPkg.writeRegistrationFileDirectly: compiler does not support dir style package dbs"
   OnlySupportSpecificPackageDb -> "HcPkg.writeRegistrationFileDirectly: only supports SpecificPackageDB for now"
   FailedToParseOutputDescribe programId pkgId -> "failed to parse output of '" ++ programId ++ " describe " ++ prettyShow pkgId ++ "'"
@@ -352,7 +353,7 @@ exceptionMessage e = case e of
       ++ " but "
       ++ "haddock is using GHC version "
       ++ prettyShow haddockGhcVersion
-  MustHaveSharedLibraries -> "Must have vanilla or shared libraries " ++ "enabled in order to run haddock"
+  MustHaveSharedLibraries -> "Must have vanilla or shared libraries enabled in order to run haddock"
   HaddockPackageFlags inf ->
     "internal error when calculating transitive "
       ++ "package dependencies.\nDebug info: "
@@ -361,7 +362,7 @@ exceptionMessage e = case e of
   FailedToDetermineTarget -> "Failed to determine target."
   NoMultipleTargets -> "The 'repl' command does not support multiple targets at once."
   REPLNotSupported -> "A REPL is not supported with this compiler."
-  NoSupportBuildingTestSuite testType -> "No support for building test suite type " ++ show testType
+  NoSupportBuildingTestSuite test_type -> "No support for building test suite type " ++ show test_type
   NoSupportBuildingBenchMark benchMarkType -> "No support for building benchmark type " ++ show benchMarkType
   BuildingNotSupportedWithCompiler -> "Building is not supported with this compiler."
   ProvideHaskellSuiteTool msg -> show msg
@@ -390,7 +391,7 @@ exceptionMessage e = case e of
   GlobalPackageDBLimitation ->
     "With current ghc versions the global package db is always used "
       ++ "and must be listed first. This ghc limitation may be lifted in "
-      ++ "future, see https://gitlab.haskell.org/ghc/ghc/-/issues/5977"
+      ++ "the future, see https://gitlab.haskell.org/ghc/ghc/-/issues/5977"
   GlobalPackageDBSpecifiedFirst ->
     "If the global package db is specified, it must be "
       ++ "specified first and cannot be specified multiple times"
@@ -486,8 +487,7 @@ exceptionMessage e = case e of
       ++ "suite type "
       ++ prettyShow tt
   NoSupportForPreProcessingBenchmark tt ->
-    "No support for preprocessing benchmark "
-      ++ "type "
+    "No support for preprocessing benchmark type "
       ++ prettyShow tt
   CantFindSourceForPreProcessFile errorStr -> errorStr
   NoSupportPreProcessingTestExtras tt ->
@@ -510,11 +510,11 @@ exceptionMessage e = case e of
   SanityCheckHookedBuildInfo exe1 ->
     "The buildinfo contains info for an executable called '"
       ++ prettyShow exe1
-      ++ "' but the package does not have a "
+      ++ "' but the package does not have an "
       ++ "executable with that name."
-  ConfigureScriptNotFound -> "configure script not found."
+  ConfigureScriptNotFound fp -> "configure script not found at " ++ fp ++ "."
   NoValidComponent -> "No valid component targets found"
-  ConfigureEitherSingleOrAll -> "Can only configure either single component or all of them"
+  ConfigureEitherSingleOrAll -> "Can only configure either a single component or all of them"
   ConfigCIDValidForPreComponent -> "--cid is only supported for per-component configure"
   SanityCheckForEnableComponents ->
     "--enable-tests/--enable-benchmarks are incompatible with"
@@ -524,23 +524,23 @@ exceptionMessage e = case e of
       ++ " are incompatible with each other."
   UnsupportedLanguages pkgId compilerId langs ->
     "The package "
-      ++ prettyShow (pkgId)
+      ++ prettyShow pkgId
       ++ " requires the following languages which are not "
       ++ "supported by "
-      ++ prettyShow (compilerId)
+      ++ prettyShow compilerId
       ++ ": "
       ++ intercalate ", " langs
   UnsupportedLanguageExtension pkgId compilerId exts ->
     "The package "
-      ++ prettyShow (pkgId)
+      ++ prettyShow pkgId
       ++ " requires the following language extensions which are not "
       ++ "supported by "
-      ++ prettyShow (compilerId)
+      ++ prettyShow compilerId
       ++ ": "
       ++ intercalate ", " exts
   CantFindForeignLibraries unsupportedFLibs ->
     "Cannot build some foreign libraries: "
-      ++ intercalate "," unsupportedFLibs
+      ++ intercalate ", " unsupportedFLibs
   ExpectedAbsoluteDirectory fPath -> "expected an absolute directory name for --prefix: " ++ fPath
   FlagsNotSpecified diffFlags ->
     "'--exact-configuration' was given, "
@@ -571,7 +571,7 @@ exceptionMessage e = case e of
       ++ "' refers to a library which is defined within the same "
       ++ "package. To use this feature the package must specify at "
       ++ "least 'cabal-version: >= 1.8'."
-  ReportFailedDependencies failed hackageUrl -> (intercalate "\n\n" (map reportFailedDependency failed))
+  ReportFailedDependencies failed hackageUrl -> intercalate "\n\n" (map reportFailedDependency failed)
     where
       reportFailedDependency (DependencyNotExists pkgname) =
         "there is no version of "
@@ -616,15 +616,16 @@ exceptionMessage e = case e of
   NoWorkingGcc ->
     unlines
       [ "No working gcc"
-      , "This package depends on foreign library but we cannot "
+      , "This package depends on a foreign library but we cannot "
           ++ "find a working C compiler. If you have it in a "
           ++ "non-standard location you can use the --with-gcc "
           ++ "flag to specify it."
       ]
-  NoOSSupport os ->
+  NoOSSupport os what ->
     "Operating system: "
       ++ prettyShow os
-      ++ ", does not support relocatable builds"
+      ++ ", does not support "
+      ++ what
   NoCompilerSupport comp ->
     "Compiler: "
       ++ comp
@@ -672,8 +673,8 @@ exceptionMessage e = case e of
           ++ "where it is."
           ++ "If the library file does exist, it may contain errors that "
           ++ "are caught by the C compiler at the preprocessing stage. "
-          ++ "In this case you can re-run configure with the verbosity "
-          ++ "flag -v3 to see the error messages."
+          ++ "In this case you can re-run 'Setup configure' with the "
+          ++ "verbosity flag -v3 to see the error messages."
       messagePlural =
         "This problem can usually be solved by installing the system "
           ++ "packages that provide these libraries (you may need the "
@@ -683,18 +684,18 @@ exceptionMessage e = case e of
           ++ "where they are."
           ++ "If the library files do exist, it may contain errors that "
           ++ "are caught by the C compiler at the preprocessing stage. "
-          ++ "In this case you can re-run configure with the verbosity "
-          ++ "flag -v3 to see the error messages."
+          ++ "In this case you can re-run 'Setup configure' with the "
+          ++ "verbosity flag -v3 to see the error messages."
       headerCppMessage =
         "If the header file does exist, it may contain errors that "
           ++ "are caught by the C compiler at the preprocessing stage. "
-          ++ "In this case you can re-run configure with the verbosity "
-          ++ "flag -v3 to see the error messages."
+          ++ "In this case you can re-run 'Setup configure' with the "
+          ++ "verbosity flag -v3 to see the error messages."
       headerCcMessage =
         "The header file contains a compile error. "
-          ++ "You can re-run configure with the verbosity flag "
+          ++ "You can re-run 'Setup configure' with the verbosity flag "
           ++ "-v3 to see the error messages from the C compiler."
-  CheckPackageProblems errors -> (intercalate "\n\n" $ errors)
+  CheckPackageProblems errors -> intercalate "\n\n" errors
   LibDirDepsPrefixNotRelative l p ->
     "Library directory of a dependency: "
       ++ show l
@@ -725,8 +726,7 @@ exceptionMessage e = case e of
       ++ "\". Did you build the package first?"
   TestCoverageSupportLibV09 -> "Test coverage is only supported for packages with a library component."
   RawSystemStdout errors -> errors
-  FindFileCwd fileName -> fileName ++ " doesn't exist"
-  FindFileEx fileName -> fileName ++ " doesn't exist"
+  FindFile fileName -> fileName ++ " doesn't exist"
   FindModuleFileEx mod_name extensions searchPath ->
     "Could not find module: "
       ++ prettyShow mod_name
@@ -756,7 +756,7 @@ exceptionMessage e = case e of
   RegisMultiplePkgNotSupported -> "Registering multiple package instances is not yet supported for this compiler"
   RegisteringNotImplemented -> "Registering is not implemented for this compiler"
   NoTestSuitesEnabled ->
-    "No test suites enabled. Did you remember to configure with "
+    "No test suites enabled. Did you remember to 'Setup configure' with "
       ++ "\'--enable-tests\'?"
   TestNameDisabled tName ->
     "Package configured with test suite "
@@ -798,3 +798,5 @@ exceptionMessage e = case e of
     "Failed to find the installed unit '"
       ++ prettyShow unitId
       ++ "' in package database stack."
+  SetupHooksException err ->
+    setupHooksExceptionMessage err
