@@ -138,6 +138,126 @@ function h$base_fstat(fd, stat, stat_off, c) {
         h$unsupported(-1, c);
 }
 
+function h$stat(path, path_off, stat, stat_off) {
+#ifndef GHCJS_BROWSER
+  if(h$isNode()) {
+    try {
+      var stats = h$fs.statSync(h$decodeUtf8z(path, path_off));
+      h$base_fillStat(stats, stat, stat_off);
+      return 0;
+    } catch(e) {
+      h$setErrno(e);
+      return -1;
+    }
+  }
+  else
+#endif
+    h$unsupported(-1);
+}
+
+function h$lstat(path, path_off, stat, stat_off) {
+#ifndef GHCJS_BROWSER
+  if(h$isNode()) {
+    try {
+      var stats = h$fs.lstatSync(h$decodeUtf8z(path, path_off));
+      h$base_fillStat(stats, stat, stat_off);
+      return 0;
+    } catch(e) {
+      h$setErrno(e);
+      return -1;
+    }
+  }
+  else
+#endif
+    h$unsupported(-1);
+}
+
+function h$fstatat(dirfd, path, path_off, stat, stat_off, flag) {
+#ifndef GHCJS_BROWSER
+  if(h$isNode()) {
+    var fp = h$calculate_at(dirfd, path, path_off);
+    try {
+      if (flag & h$base_at_symlink_nofollow) {
+        var fs = h$fs.lstatSync(fp);
+        h$base_fillStat(fs, stat, stat_off);
+        return 0;
+      }
+      else {
+        var fs = h$fs.statSync(fp);
+        h$base_fillStat(fs, stat, stat_off);
+        return 0;
+      }
+
+    } catch(e) {
+      h$setErrno(e);
+      return -1;
+    }
+  }
+#endif
+
+  return h$unsupported(-1);
+}
+
+function h$unlinkat(dirfd, path, path_off, flag) {
+#ifndef GHCJS_BROWSER
+  if(h$isNode()) {
+    var fp = h$calculate_at(dirfd, path, path_off);
+    try {
+      if (flag & h$base_at_removedir) {
+        h$fs.rmdirSync(fp);
+        return 0;
+      }
+      else {
+        h$fs.unlinkSync(fp);
+        return 0;
+      }
+
+    } catch(e) {
+      h$setErrno(e);
+      return -1;
+    }
+  }
+#endif
+
+  return h$unsupported(-1);
+}
+
+function h$dup(fd) {
+#ifndef GHCJS_BROWSER
+  if(h$isNode()) {
+    try {
+      // NodeJS doesn't provide "dup" (see
+      // https://github.com/nodejs/node/issues/41733), so we do this hack that
+      // probably only works on Linux.
+      return h$fs.openSync("/proc/self/fd/"+fd);
+    } catch(e) {
+      h$setErrno(e);
+      return -1;
+    }
+  }
+  else
+#endif
+    h$unsupported(-1);
+}
+
+function h$fdopendir(fd) {
+#ifndef GHCJS_BROWSER
+  if(h$isNode()) {
+    try {
+      // NodeJS doesn't provide "fdopendir", so we do this hack that probably
+      // only works on Linux.
+      return h$fs.opendirSync("/proc/self/fd/"+fd);
+    } catch(e) {
+      h$setErrno(e);
+      return -1;
+    }
+  }
+  else
+#endif
+    h$unsupported(-1);
+}
+
+
 function h$base_isatty(fd) {
     TRACE_IO("base_isatty " + fd)
     //  return 1; // fixme debug
@@ -247,23 +367,6 @@ function h$base_lstat(file, file_off, stat, stat_off, c) {
         h$unsupported(-1, c);
 }
 
-function h$lstat(file, file_off, stat, stat_off) {
-  TRACE_IO("lstat")
-#ifndef GHCJS_BROWSER
-  if(h$isNode()) {
-    try {
-      var fs = h$fs.lstatSync(h$decodeUtf8z(file, file_off));
-      h$base_fillStat(fs, stat, stat_off);
-      return 0;
-    } catch(e) {
-      h$setErrno(e);
-      return -1;
-    }
-  } else
-#endif
-    h$unsupported(-1);
-}
-
 function h$rmdir(file, file_off) {
   TRACE_IO("rmdir")
 #ifndef GHCJS_BROWSER
@@ -285,7 +388,7 @@ function h$rename(old_path, old_path_off, new_path, new_path_off) {
 #ifndef GHCJS_BROWSER
   if (h$isNode()) {
     try {
-      fs.renameSync(h$decodeUtf8z(old_path, old_path_off), h$decodeUtf8z(new_path, new_path_off));
+      h$fs.renameSync(h$decodeUtf8z(old_path, old_path_off), h$decodeUtf8z(new_path, new_path_off));
       return 0;
     } catch(e) {
       h$setErrno(e);
@@ -318,7 +421,7 @@ function h$realpath(path,off,resolved,resolved_off) {
 #ifndef GHCJS_BROWSER
   if (h$isNode()) {
     try {
-      var rp = h$encodeUtf8(fs.realpathSync(h$decodeUtf8z(path,off)));
+      var rp = h$encodeUtf8(h$fs.realpathSync(h$decodeUtf8z(path,off)));
       if (resolved !== null) {
         h$copyMutableByteArray(rp, 0, resolved, resolved_off, Math.min(resolved.len - resolved_off, rp.len));
         RETURN_UBX_TUP2(resolved, resolved_off);
@@ -333,25 +436,55 @@ function h$realpath(path,off,resolved,resolved_off) {
     h$unsupported(-1);
 }
 
-function h$base_open(file, file_off, how, mode, c) {
-  return h$open(file,file_off,how,mode,c);
+function h$path_is_abs(path) {
+ return path.charAt(0) === '/';
 }
 
-function h$openat(dirfd, file, file_off, how, mode) {
-  if (dirfd != h$base_at_fdcwd) {
-    // we only support AT_FDWCD (open) until NodeJS provides "openat"
+function h$path_join2(p1,p2) {
+  // Emscripten would normalize the path here. We don't for now.
+ return (p1 + '/' + p2);
+}
+
+// Compute path from a FD and a path
+function h$calculate_at(dirfd, file, file_off) {
+  var path = h$decodeUtf8z(file,file_off);
+
+  if (h$path_is_abs(path)) {
+    return path;
+  }
+
+  // relative path
+  var dir;
+  if (dirfd == h$base_at_fdcwd) {
+    dir = h$process.cwd();
+  }
+#ifndef GHCJS_BROWSER
+  else if (h$isNode()) {
+    // hack that probably only works on Linux with /proc mounted
+    dir = h$fs.readlinkSync("/proc/self/fd/"+dirfd);
+  }
+#endif
+  else {
     return h$unsupported(-1);
   }
-  else {
-    return h$open(file,file_off,how,mode,undefined);
-  }
+
+  return h$path_join2(dir,path);
 }
 
-function h$open(file, file_off, how, mode,c) {
+function h$openat(dirfd, file, file_off, how, mode, c) {
+  var path = h$calculate_at(dirfd, file, file_off);
+  return h$base_open(path, how, mode, c);
+}
+
+function h$open(file, file_off, how, mode, c) {
+  var path = h$decodeUtf8z(file, file_off);
+  return h$base_open(path, how, mode, c);
+}
+
+function h$base_open(fp, how, mode, c) {
 #ifndef GHCJS_BROWSER
     if(h$isNode()) {
         var flags, off;
-        var fp   = h$decodeUtf8z(file, file_off);
         TRACE_IO("open: " + fp)
         var acc  = how & h$base_o_accmode;
         // passing a number lets node.js use it directly as the flags (undocumented)
@@ -586,6 +719,9 @@ const h$base_o_noctty   = 0x20000;
 const h$base_o_nonblock = 0x00004;
 const h$base_o_binary   = 0x00000;
 const h$base_at_fdcwd   = -100;
+const h$base_at_symlink_nofollow = 0x100;
+const h$base_at_removedir        = 0x200;
+const h$base_at_symlink_follow   = 0x400;
 
 
 function h$base_stat_check_mode(mode,p) {
@@ -912,7 +1048,7 @@ if(h$isNode()) {
     }
 } else if(h$isJsCore()) {
     h$base_readStdin = function(fd, fdo, buf, buf_offset, n, c) {
-	c(0);
+        c(0);
     }
     var h$base_stdoutLeftover = { f: print, val: null };
     var h$base_stderrLeftover = { f: debug, val: null };
@@ -1023,7 +1159,7 @@ function h$opendir(path) {
     throw "h$opendir unsupported";
   }
 
-  const d = fs.opendirSync(h$decodeUtf8z(path,0));
+  const d = h$fs.opendirSync(h$decodeUtf8z(path,0));
   RETURN_UBX_TUP2(d,0);
 }
 
@@ -1077,3 +1213,12 @@ function h$mkdir(path, path_offset, mode) {
   }
   return 0;
 }
+
+// It is required by Google Closure Compiler to be at least defined if
+// somewhere it is used
+var h$stg_cloneMyStackzh, h$stg_decodeStackzh
+h$stg_cloneMyStackzh
+  = h$stg_decodeStackzh
+  = function () {
+    throw new Error('Stack Cloning Decoding: Not Implemented Yet')
+  }
