@@ -7,6 +7,7 @@ module Tests.Properties.Substrings
     ( testSubstrings
     ) where
 
+import Prelude hiding (head, tail)
 import Control.Monad.Trans.State (State, state, runState)
 import Data.Char (isSpace)
 import Test.QuickCheck
@@ -14,6 +15,7 @@ import Test.Tasty (TestTree, testGroup)
 import Test.Tasty.QuickCheck (testProperty)
 import Tests.QuickCheckUtils
 import qualified Data.List as L
+import qualified Data.List.NonEmpty as NonEmptyList
 import qualified Data.Text as T
 import qualified Data.Text.Internal.Fusion as S
 import qualified Data.Text.Internal.Fusion.Common as S
@@ -155,9 +157,13 @@ t_groupBy (applyFun2 -> p)
 tl_groupBy (applyFun2 -> p)
                   = L.groupBy p   `eqP` (map unpackS . TL.groupBy p)
 t_inits           = L.inits       `eqP` (map unpackS . T.inits)
+t_initsNE         = NonEmptyList.inits `eqP` (fmap unpackS . T.initsNE)
 tl_inits          = L.inits       `eqP` (map unpackS . TL.inits)
+tl_initsNE        = NonEmptyList.inits `eqP` (fmap unpackS . TL.initsNE)
 t_tails           = L.tails       `eqP` (map unpackS . T.tails)
+t_tailsNE         = NonEmptyList.tails `eqP` (fmap unpackS . T.tailsNE)
 tl_tails          = L.tails       `eqPSqrt` (map unpackS . TL.tails)
+tl_tailsNE        = NonEmptyList.tails `eqP` (fmap unpackS . TL.tailsNE)
 
 spanML :: Monad m => (b -> m Bool) -> [b] -> m ([b], [b])
 spanML p s = go [] s
@@ -218,9 +224,10 @@ tl_split (applyFun -> p) = split p `eqP` (map unpackS . TL.split p)
 split :: (a -> Bool) -> [a] -> [[a]]
 split _ [] =  [[]]
 split p xs = loop xs
-    where loop s | null s'   = [l]
-                 | otherwise = l : loop (tail s')
-              where (l, s') = break p s
+  where
+    loop s = case break p s of
+      (l, []) -> [l]
+      (l, _ : s') -> l : loop s'
 
 t_chunksOf_same_lengths k = conjoin . map ((===k) . T.length) . ini . T.chunksOf k
   where ini [] = []
@@ -237,7 +244,9 @@ tl_lines          = L.lines       `eqP` (map unpackS . TL.lines)
 t_lines_spacy     = (L.lines      `eqP` (map unpackS . T.lines))  . getSpacyString
 tl_lines_spacy    = (L.lines      `eqP` (map unpackS . TL.lines)) . getSpacyString
 
-tl_lines_laziness = TL.head (head (TL.lines (TL.replicate 1000000000000000 (TL.singleton 'a')))) === 'a'
+tl_lines_laziness = case TL.lines (TL.replicate 1000000000000000 (TL.singleton 'a')) of
+  [] -> property False
+  hd : _ -> TL.head hd === 'a'
 
 tl_lines_specialCase = TL.lines (TL.Chunk (T.pack "foo") $ TL.Chunk (T.pack "bar\nbaz\n") $ TL.Empty) === [TL.pack "foobar", TL.pack "baz"]
 
@@ -357,9 +366,13 @@ testSubstrings =
       testProperty "t_groupBy" t_groupBy,
       testProperty "tl_groupBy" tl_groupBy,
       testProperty "t_inits" t_inits,
+      testProperty "t_initsNE" t_initsNE,
       testProperty "tl_inits" tl_inits,
+      testProperty "tl_initsNE" tl_initsNE,
       testProperty "t_tails" t_tails,
+      testProperty "t_tailsNE" t_tailsNE,
       testProperty "tl_tails" tl_tails,
+      testProperty "tl_tailsNE" tl_tailsNE,
       testProperty "t_spanM" t_spanM,
       testProperty "t_spanEndM" t_spanEndM,
       testProperty "tl_spanM" tl_spanM,
