@@ -1,13 +1,13 @@
 {-# LANGUAGE Safe #-}
 
-module Data.Time.Format
-    (
+module Data.Time.Format (
     -- * UNIX-style formatting
-      FormatTime()
-    , formatTime
-    , module Data.Time.Format.Parse
-    ) where
+    FormatTime (),
+    formatTime,
+    module Data.Time.Format.Parse,
+) where
 
 import Data.Time.Format.Format.Class
 import Data.Time.Format.Format.Instances ()
+import Data.Time.Format.ISO8601 ()
 import Data.Time.Format.Parse
