@@ -1,33 +1,28 @@
 {-# LANGUAGE Safe #-}
 
-module Data.Time.LocalTime.Internal.CalendarDiffTime
-    (
-        -- * Calendar Duration
-        module Data.Time.LocalTime.Internal.CalendarDiffTime
-    ) where
-#if !MIN_VERSION_base(4,11,0)
-import Data.Semigroup hiding (option)
-#endif
-import Data.Fixed
-import Data.Typeable
-import Data.Data
+module Data.Time.LocalTime.Internal.CalendarDiffTime (
+    -- * Calendar Duration
+    module Data.Time.LocalTime.Internal.CalendarDiffTime,
+) where
+
 import Control.DeepSeq
+import Data.Data
 import Data.Time.Calendar.CalendarDiffDays
 import Data.Time.Clock.Internal.NominalDiffTime
+import GHC.Generics
 
 data CalendarDiffTime = CalendarDiffTime
     { ctMonths :: Integer
     , ctTime :: NominalDiffTime
-    } deriving (Eq,
-    Data
-#if __GLASGOW_HASKELL__ >= 802
-    -- ^ @since 1.9.2
-#endif
-    ,Typeable
-#if __GLASGOW_HASKELL__ >= 802
-    -- ^ @since 1.9.2
-#endif
-    )
+    }
+    deriving
+        ( Eq
+        , -- | @since 1.9.2
+          Data
+        , -- | @since 1.9.2
+          Typeable
+        , Generic
+        )
 
 instance NFData CalendarDiffTime where
     rnf (CalendarDiffTime m t) = rnf m `seq` rnf t `seq` ()
@@ -41,9 +36,6 @@ instance Monoid CalendarDiffTime where
     mempty = CalendarDiffTime 0 0
     mappend = (<>)
 
-instance Show CalendarDiffTime where
-    show (CalendarDiffTime m t) = "P" ++ show m ++ "MT" ++ showFixed True (realToFrac t :: Pico) ++ "S"
-
 calendarTimeDays :: CalendarDiffDays -> CalendarDiffTime
 calendarTimeDays (CalendarDiffDays m d) = CalendarDiffTime m $ fromInteger d * nominalDay
 
