@@ -281,7 +281,7 @@ initCapability (Capability *cap, uint32_t i)
 #endif
     cap->total_allocated        = 0;
 
-    initCapabilityIOManager(&cap->iomgr);
+    initCapabilityIOManager(cap); /* initialises cap->iomgr */
 
     cap->f.stgEagerBlackholeInfo = (W_)&__stg_EAGER_BLACKHOLE_info;
     cap->f.stgGCEnter1     = (StgFunPtr)__stg_gc_enter_1;
@@ -1322,7 +1322,7 @@ markCapability (evac_fn evac, void *user, Capability *cap,
     }
 #endif
 
-    markCapabilityIOManager(evac, user, cap->iomgr);
+    markCapabilityIOManager(evac, user, cap);
 
     // Free STM structures for this Capability
     stmPreGCHook(cap);
