@@ -8,6 +8,7 @@
 
 #include "rts/PosixSource.h"
 #include "Rts.h"
+#include "RtsFlags.h"
 
 #include "sm/Storage.h"
 #include "Threads.h"
@@ -19,6 +20,7 @@
 #include "sm/Sanity.h"
 #include "Profiling.h"
 #include "Messages.h"
+#include "IOManager.h"
 #if defined(mingw32_HOST_OS)
 #include "win32/MIOManager.h"
 #endif
@@ -238,7 +240,7 @@ throwToMsg (Capability *cap, MessageThrowTo *msg)
     goto check_target;
 
 retry:
-    write_barrier();
+    RELEASE_FENCE();
     debugTrace(DEBUG_sched, "throwTo: retrying...");
 
 check_target:
@@ -266,7 +268,7 @@ check_target:
         return THROWTO_BLOCKED;
     }
 
-    status = target->why_blocked;
+    status = ACQUIRE_LOAD(&target->why_blocked);
 
     switch (status) {
     case NotBlocked:
@@ -702,33 +704,24 @@ removeFromQueues(Capability *cap, StgTSO *tso)
       break;
   }
 
-#if !defined(THREADED_RTS)
   case BlockedOnRead:
   case BlockedOnWrite:
-#if defined(mingw32_HOST_OS)
   case BlockedOnDoProc:
-#endif
-      removeThreadFromDeQueue(cap, &cap->iomgr->blocked_queue_hd,
-                                   &cap->iomgr->blocked_queue_tl, tso);
-#if defined(mingw32_HOST_OS)
-      /* (Cooperatively) signal that the worker thread should abort
-       * the request.
-       */
-      abandonWorkRequest(tso->block_info.async_result->reqID);
-#endif
+      // These blocking reasons are only used by some I/O managers
+      syncIOCancel(cap, tso);
       goto done;
 
   case BlockedOnDelay:
-        removeThreadFromQueue(cap, &cap->iomgr->sleeping_queue, tso);
-        goto done;
-#endif
+      // This blocking reasons is only used by some I/O managers
+      syncDelayCancel(cap, tso);
+      goto done;
 
   default:
       barf("removeFromQueues: %d", tso->why_blocked);
   }
 
  done:
-  tso->why_blocked = NotBlocked;
+  RELAXED_STORE(&tso->why_blocked, NotBlocked);
   appendToRunQueue(cap, tso);
 }
 
@@ -874,9 +867,10 @@ raiseAsync(Capability *cap, StgTSO *tso, StgClosure *exception,
                 ap->payload[i] = (StgClosure *)*sp++;
             }
 
-            write_barrier(); // XXX: Necessary?
             SET_HDR(ap,&stg_AP_STACK_info,
                     ((StgClosure *)frame)->header.prof.ccs /* ToDo */);
+            // N.B. This will be made visible by updateThunk below, which
+            // implies a release memory barrier.
             TICK_ALLOC_UP_THK(AP_STACK_sizeW(words),0);
 
             //IF_DEBUG(scheduler,
