@@ -9,6 +9,7 @@
 #include "rts/PosixSource.h"
 #include "Rts.h"
 #include "RtsAPI.h"
+#include "RtsFlags.h"
 #include "HsFFI.h"
 
 #include "RtsUtils.h"
@@ -19,6 +20,7 @@
 #include "StablePtr.h"
 #include "Threads.h"
 #include "Weak.h"
+#include "sm/NonMoving.h"
 
 /* ----------------------------------------------------------------------------
    Building Haskell objects from C datatypes.
@@ -507,7 +509,7 @@ void rts_evalStableIOMain(/* inout */ Capability **cap,
     SchedulerStatus stat;
 
     p = (StgClosure *)deRefStablePtr(s);
-    w = rts_apply(*cap, &base_GHCziTopHandler_runMainIO_closure, p);
+    w = rts_apply(*cap, &ghczminternal_GHCziInternalziTopHandler_runMainIO_closure, p);
     tso = createStrictIOThread(*cap, RtsFlags.GcFlags.initialStkSize, w);
     // async exceptions are always blocked by default in the created
     // thread.  See #1048.
@@ -709,6 +711,16 @@ Capability *pauseTokenCapability(PauseToken *pauseToken) {
 // See Note [Locking and Pausing the RTS]
 PauseToken *rts_pause (void)
 {
+
+    // Wait for any nonmoving collection to finish before pausing the RTS.
+    // The nonmoving collector needs to synchronise with the mutator,
+    // so pausing the mutator while a collection is ongoing might lead to deadlock or
+    // capabilities being prematurely re-awoken.
+    if (RtsFlags.GcFlags.useNonmoving) {
+        nonmovingBlockConcurrentMark(true);
+    }
+
+
     // It is an error if this thread already paused the RTS. If another
     // thread has paused the RTS, then rts_pause will block until rts_resume is
     // called (and compete with other threads calling rts_pause). The blocking
@@ -771,6 +783,10 @@ void rts_resume (PauseToken *pauseToken)
     releaseAllCapabilities(getNumCapabilities(), NULL, task);
     exitMyTask();
     stgFree(pauseToken);
+
+    if (RtsFlags.GcFlags.useNonmoving) {
+        nonmovingUnblockConcurrentMark();
+    }
 }
 
 // See RtsAPI.h
