@@ -16,6 +16,7 @@
 #include "sm/OSMem.h"
 #include "hooks/Hooks.h"
 #include "Capability.h"
+#include "IOManager.h"
 
 #if defined(HAVE_CTYPE_H)
 #include <ctype.h>
@@ -269,11 +270,7 @@ void initRtsFlagsDefaults(void)
     RtsFlags.MiscFlags.internalCounters        = false;
     RtsFlags.MiscFlags.linkerAlwaysPic         = DEFAULT_LINKER_ALWAYS_PIC;
     RtsFlags.MiscFlags.linkerMemBase           = 0;
-#if defined(DEFAULT_NATIVE_IO_MANAGER)
-    RtsFlags.MiscFlags.ioManager               = IO_MNGR_NATIVE;
-#else
-    RtsFlags.MiscFlags.ioManager               = IO_MNGR_POSIX;
-#endif
+    RtsFlags.MiscFlags.ioManager               = IO_MNGR_FLAG_AUTO;
 #if defined(THREADED_RTS) && defined(mingw32_HOST_OS)
     RtsFlags.MiscFlags.numIoWorkerThreads      = getNumberOfProcessors();
 #else
@@ -300,6 +297,7 @@ void initRtsFlagsDefaults(void)
     RtsFlags.TickyFlags.showTickyStats   = false;
     RtsFlags.TickyFlags.tickyFile        = NULL;
 #endif
+    RtsFlags.HpcFlags.readTixFile        = HPC_YES_IMPLICIT;
     RtsFlags.HpcFlags.writeTixFile       = true;
 }
 
@@ -490,6 +488,7 @@ usage_text[] = {
 "  -Dr  DEBUG: sparks",
 "  -DC  DEBUG: compact",
 "  -Dk  DEBUG: continuation",
+"  -Do  DEBUG: iomanager",
 "",
 "     NOTE: DEBUG events are sent to stderr by default; add -l to create a",
 "     binary event log file instead.",
@@ -533,8 +532,10 @@ usage_text[] = {
 "             fatal error. When symbols are available an attempt will be",
 "             made to resolve addresses to names. (default: yes)",
 #endif
-"  --io-manager=<native|posix>",
-"             The I/O manager subsystem to use. (default: posix)",
+"  --io-manager=<name>",
+"             The I/O manager to use.",
+"             Options available: auto" IOMGRS_ENABLED_STR
+              " (default: " IOMGR_DEFAULT_STR ")",
 #if defined(THREADED_RTS)
 #if defined(mingw32_HOST_OS)
 "  --io-manager-threads=<num>",
@@ -565,6 +566,10 @@ usage_text[] = {
 "             HeapOverflow exception before the exception is thrown again, if",
 "             the program is still exceeding the heap limit.",
 "",
+"  --read-tix-file=<yes|no>",
+"             Whether to initialize HPC datastructures from  <program>.tix "
+"             at the start of execution. (default: yes)",
+"",
 "  --write-tix-file=<yes|no>",
 "             Whether to write <program>.tix at the end of execution.",
 "             (default: yes)",
@@ -1013,15 +1018,23 @@ error = true;
                       OPTION_SAFE;
                       RtsFlags.MiscFlags.internalCounters = true;
                   }
-                  else if (strequal("io-manager=native",
-                               &rts_argv[arg][2])) {
+                  else if (!strncmp("io-manager=",
+                               &rts_argv[arg][2], 11)) {
                       OPTION_UNSAFE;
-                      RtsFlags.MiscFlags.ioManager = IO_MNGR_NATIVE;
-                  }
-                  else if (strequal("io-manager=posix",
-                               &rts_argv[arg][2])) {
-                      OPTION_UNSAFE;
-                      RtsFlags.MiscFlags.ioManager = IO_MNGR_POSIX;
+                      char *iomgrstr = &rts_argv[arg][13];
+                      IO_MANAGER_FLAG iomgrflag;
+                      enum IOManagerAvailability availability;
+                      availability = parseIOManagerFlag(iomgrstr, &iomgrflag);
+                      if (availability == IOManagerAvailable) {
+                          RtsFlags.MiscFlags.ioManager = iomgrflag;
+                      } else {
+                          errorBelch("%s choice '%s' for --io-manager=\n"
+                                     "The choices are: auto%s",
+                                     availability == IOManagerUnavailable ?
+                                     "unavailable" : "unrecognised",
+                                     iomgrstr, IOMGRS_ENABLED_STR);
+                          stg_exit(EXIT_FAILURE);
+                      }
                   }
                   else if (strequal("info",
                                &rts_argv[arg][2])) {
@@ -1060,6 +1073,16 @@ error = true;
                         RtsFlags.GcFlags.nonmovingDenseAllocatorCount = threshold;
                       }
                   }
+                  else if (strequal("read-tix-file=yes",
+                              &rts_argv[arg][2])) {
+                       OPTION_UNSAFE;
+                       RtsFlags.HpcFlags.readTixFile = HPC_YES_EXPLICIT;
+                  }
+                  else if (strequal("read-tix-file=no",
+                              &rts_argv[arg][2])) {
+                       OPTION_UNSAFE;
+                       RtsFlags.HpcFlags.readTixFile = HPC_NO_EXPLICIT;
+                  }
                   else if (strequal("write-tix-file=yes",
                               &rts_argv[arg][2])) {
                        OPTION_UNSAFE;
@@ -1951,6 +1974,9 @@ static void normaliseRtsOpts (void)
     if (RtsFlags.GcFlags.maxHeapSize != 0 &&
         RtsFlags.GcFlags.heapSizeSuggestion >
         RtsFlags.GcFlags.maxHeapSize) {
+        errorBelch("Maximum heap size (-M) is smaller than suggested heap size (-H)\n"
+                   "Setting maximum heap size to suggested heap size ( %" FMT_Word64 " )",
+                   (StgWord64) RtsFlags.GcFlags.maxHeapSize * (StgWord64) BLOCK_SIZE);
         RtsFlags.GcFlags.maxHeapSize = RtsFlags.GcFlags.heapSizeSuggestion;
     }
 
@@ -2253,6 +2279,9 @@ static void read_debug_flags(const char* arg)
         case 'k':
             RtsFlags.DebugFlags.continuation = true;
             break;
+        case 'o':
+            RtsFlags.DebugFlags.iomanager = true;
+            break;
         default:
             bad_option( arg );
         }
@@ -2722,16 +2751,6 @@ files like <progname>.eventlog, not arbitrary files.
    Helper utilities to query state.
    ------------------------------------------------------------------------- */
 
-bool is_io_mng_native_p (void)
-{
-#if defined(mingw32_HOST_OS)
-  return RtsFlags.MiscFlags.ioManager == IO_MNGR_NATIVE;
-#else
-  return false;
-#endif
-}
-
-
 #if defined(PROFILING)
 bool
 doingLDVProfiling( void )
