@@ -26,7 +26,7 @@
  *
  * -------------------------------------------------------------------------- */
 
-#include "PosixSource.h"
+#include "rts/PosixSource.h"
 #include "ghcconfig.h"
 
 // Enable DWARF Call-Frame Information (used for stack unwinding) on Linux.
@@ -36,7 +36,7 @@
 #define ENABLE_UNWINDING
 #endif
 
-#if defined(sparc_HOST_ARCH) || defined(USE_MINIINTERPRETER)
+#if defined(USE_MINIINTERPRETER)
 /* include Stg.h first because we want real machine regs in here: we
  * have to get the value of R1 back from Stg land to C land intact.
  */
@@ -102,18 +102,12 @@ StgFunPtr StgReturn(void)
 
 #if defined(i386_HOST_ARCH)
 
-#if defined(darwin_HOST_OS) || defined(ios_HOST_OS)
-#define STG_GLOBAL ".globl "
-#define STG_HIDDEN ".private_extern "
-#else
 #define STG_GLOBAL ".global "
 #define STG_HIDDEN ".hidden "
-#endif
 
 /*
  * Note [Stack Alignment on X86]
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- *
  * On X86 (both 32bit and 64bit) we keep the stack aligned on function calls at
  * a 16-byte boundary. This is done because on a number of architectures the
  * ABI requires this (e.g. the System V AMD64 ABI, Mac OS X 32-bit/64-bit ABIs,
@@ -150,7 +144,7 @@ StgFunPtr StgReturn(void)
  * for stg_stop_thread in StgStartup.cmm.
  *
  * Note [Windows Stack allocations]
- *
+ * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  * On windows the stack has to be allocated 4k at a time, otherwise
  * we get a segfault.  This is done by using a helper ___chkstk_ms that is
  * provided by libgcc.  The Haskell side already knows how to handle this
@@ -164,7 +158,7 @@ StgFunPtr StgReturn(void)
  * */
 
 
-static void GNUC3_ATTRIBUTE(used)
+static void STG_USED
 StgRunIsImplementedInAssembler(void)
 {
     __asm__ volatile (
@@ -373,7 +367,7 @@ stack unwinding.
 */
 
 
-static void GNUC3_ATTRIBUTE(used)
+static void STG_USED
 StgRunIsImplementedInAssembler(void)
 {
     __asm__ volatile (
@@ -575,79 +569,6 @@ StgRunIsImplementedInAssembler(void)
 
 #endif /* x86-64 */
 
-/* -----------------------------------------------------------------------------
-   Sparc architecture
-
-   --
-   OLD COMMENT from GHC-3.02:
-
-   We want tailjumps to be calls, because `call xxx' is the only Sparc
-   branch that allows an arbitrary label as a target.  (Gcc's ``goto
-   *target'' construct ends up loading the label into a register and
-   then jumping, at the cost of two extra instructions for the 32-bit
-   load.)
-
-   When entering the threaded world, we stash our return address in a
-   known location so that \tr{%i7} is available as an extra
-   callee-saves register.  Of course, we have to restore this when
-   coming out of the threaded world.
-
-   I hate this god-forsaken architecture.  Since the top of the
-   reserved stack space is used for globals and the bottom is reserved
-   for outgoing arguments, we have to stick our return address
-   somewhere in the middle.  Currently, I'm allowing 100 extra
-   outgoing arguments beyond the first 6.  --JSM
-
-   Updated info (GHC 4.06): we don't appear to use %i7 any more, so
-   I'm not sure whether we still need to save it.  Incedentally, what
-   does the last paragraph above mean when it says "the top of the
-   stack is used for globals"?  What globals?  --SDM
-
-   Updated info (GHC 4.08.2): not saving %i7 any more (see below).
-   -------------------------------------------------------------------------- */
-
-#if defined(sparc_HOST_ARCH)
-
-StgRegTable *
-StgRun(StgFunPtr f, StgRegTable *basereg) {
-
-    unsigned char space[RESERVED_C_STACK_BYTES];
-#if 0
-    register void *i7 __asm__("%i7");
-    ((void **)(space))[100] = i7;
-#endif
-    f();
-    __asm__ volatile (
-                 ".align 4\n"
-                 ".global " STG_RETURN "\n"
-                 STG_RETURN ":"
-                 : : "p" (space) : "l0","l1","l2","l3","l4","l5","l6","l7");
-    /* we tell the C compiler that l0-l7 are clobbered on return to
-     * StgReturn, otherwise it tries to use these to save eg. the
-     * address of space[100] across the call.  The correct thing
-     * to do would be to save all the callee-saves regs, but we
-     * can't be bothered to do that.
-     *
-     * We also explicitly mark space as used since gcc eliminates it
-     * otherwise.
-     *
-     * The code that gcc generates for this little fragment is now
-     * terrible.  We could do much better by coding it directly in
-     * assembler.
-     */
-#if 0
-    /* updated 4.08.2: we don't save %i7 in the middle of the reserved
-     * space any more, since gcc tries to save its address across the
-     * call to f(), this gets clobbered in STG land and we end up
-     * dereferencing a bogus pointer in StgReturn.
-     */
-    __asm__ volatile ("ld %1,%0"
-                                : "=r" (i7) : "m" (((void **)(space))[100]));
-#endif
-    return (StgRegTable *)R1.i;
-}
-
-#endif
 
 /* -----------------------------------------------------------------------------
    PowerPC architecture
@@ -669,7 +590,7 @@ StgRun(StgFunPtr f, StgRegTable *basereg) {
 
 // This version is for PowerPC Linux.
 
-static void GNUC3_ATTRIBUTE(used)
+static void STG_USED
 StgRunIsImplementedInAssembler(void)
 {
         __asm__ volatile (
@@ -744,7 +665,7 @@ StgRunIsImplementedInAssembler(void)
    Everything is in assembler, so we don't have to deal with GCC...
    -------------------------------------------------------------------------- */
 
-#if defined(powerpc64_HOST_ARCH)
+#if defined(powerpc64_HOST_ARCH) && (!defined _CALL_ELF || _CALL_ELF == 1)
 /* 64-bit PowerPC ELF ABI 1.9
  *
  * Stack frame organization (see Figure 3-17, ELF ABI 1.9, p 14)
@@ -772,7 +693,7 @@ StgRunIsImplementedInAssembler(void)
  * the TOC pointer from the function descriptor upon a call to StgReturn.
  * That TOC pointer is the same as the TOC pointer in StgRun.
  */
-static void GNUC3_ATTRIBUTE(used)
+static void STG_USED
 StgRunIsImplementedInAssembler(void)
 {
         __asm__ volatile (
@@ -891,7 +812,7 @@ StgRun(StgFunPtr f, StgRegTable *basereg) {
            those regs not used in Thumb mode. Hard to judge if this is
            needed, but certainly Haskell code is using them for
            placing GHC's virtual registers there. See
-           includes/stg/MachRegs.h Please note that Haskell code is
+           rts/include/stg/MachRegs.h Please note that Haskell code is
            compiled by GHC/LLVM into ARM code (not Thumb!), at least
            as of February 2012 */
       : "%r4", "%r5", "%r6", "%r8", "%r9", "%r10", "%11", "%ip", "%lr"
