@@ -357,7 +357,8 @@ checkClosure( const StgClosure* p )
     info = ACQUIRE_LOAD(&p->header.info);
 
     if (IS_FORWARDING_PTR(info)) {
-        barf("checkClosure: found EVACUATED closure %d", info->type);
+        ASSERT(LOOKS_LIKE_CLOSURE_PTR(info));
+        barf("checkClosure: found EVACUATED closure %u", GET_INFO((StgClosure*)UN_FORWARDING_PTR(info))->type);
     }
 
 #if defined(PROFILING)
@@ -833,8 +834,12 @@ checkGlobalTSOList (bool checkTSOs)
               StgUnderflowFrame *frame =
                   (StgUnderflowFrame*) (stack->stack + stack->stack_size
                           - sizeofW(StgUnderflowFrame));
-              if (frame->info != &stg_stack_underflow_frame_info
-                      || frame->next_chunk == (StgStack*)END_TSO_QUEUE) {
+              if (frame->next_chunk == (StgStack*)END_TSO_QUEUE ||
+                  (  frame->info != &stg_stack_underflow_frame_d_info
+                  && frame->info != &stg_stack_underflow_frame_v16_info
+                  && frame->info != &stg_stack_underflow_frame_v32_info
+                  && frame->info != &stg_stack_underflow_frame_v64_info
+                  )) {
                   break;
               }
               stack = frame->next_chunk;
@@ -1238,6 +1243,7 @@ countNonMovingHeap(struct NonmovingHeap *heap)
     }
     ret += countNonMovingSegments(heap->sweep_list);
     ret += countNonMovingSegments(heap->free);
+    ret += countNonMovingSegments(heap->saved_free);
     return ret;
 }
 
