@@ -5,6 +5,8 @@
  *       For the WINIO manager see base in the GHC.Event modules.
  */
 #include "Rts.h"
+#include "MIOManager.h"
+#include "ThreadLabels.h"
 #include <windows.h>
 #include "ConsoleHandler.h"
 #include "Schedule.h"
@@ -230,7 +232,7 @@ static BOOL WINAPI generic_handler(DWORD dwCtrlType)
             stg_pending_buf[stg_pending_events] = dwCtrlType;
             stg_pending_events++;
         }
-        // we need to wake up awaitEvent()
+        // we need to wake up awaitCompletedTimeoutsOrIO()
         abandonRequestWait();
 #endif
         return true;
