@@ -13,6 +13,7 @@ import Data.Char
 import Data.List (union, intersperse, intercalate, nub)
 import Data.Maybe ( catMaybes )
 import System.Environment ( getArgs )
+import System.IO ( hSetEncoding, stdin, stdout, utf8 )
 
 vecOptions :: Entry -> [(String,String,Int)]
 vecOptions i =
@@ -78,6 +79,8 @@ desugarVectorSpec i              = case vecOptions i of
         desugarTy (TyApp SCALAR [])   = TyApp (TyCon repCon) []
         desugarTy (TyApp VECTOR [])   = TyApp (VecTyCon vecCons vecTyName) []
         desugarTy (TyApp VECTUPLE []) = TyUTup (replicate n (TyApp (TyCon repCon) []))
+        desugarTy (TyApp INTVECTUPLE [])
+                                      = TyUTup (replicate n (TyApp (TyCon "Int#") []) )
         desugarTy (TyApp tycon ts)    = TyApp tycon (map desugarTy ts)
         desugarTy t@(TyVar {})        = t
         desugarTy (TyUTup ts)         = TyUTup (map desugarTy ts)
@@ -116,7 +119,9 @@ main = getArgs >>= \args ->
                    ++ unlines (map ("            "++) known_args)
                   )
        else
-       do s <- getContents
+       do hSetEncoding stdin  utf8 -- The input file is in UTF-8. Set the encoding explicitly.
+          hSetEncoding stdout utf8
+          s <- getContents
           case parse s of
              Left err -> error ("parse error at " ++ (show err))
              Right p_o_specs@(Info _ _)
@@ -126,11 +131,6 @@ main = getArgs >>= \args ->
                       "--data-decl"
                          -> putStr (gen_data_decl p_o_specs)
 
-                      "--has-side-effects"
-                         -> putStr (gen_switch_from_attribs
-                                       "has_side_effects"
-                                       "primOpHasSideEffects" p_o_specs)
-
                       "--out-of-line"
                          -> putStr (gen_switch_from_attribs
                                        "out_of_line"
@@ -146,10 +146,15 @@ main = getArgs >>= \args ->
                                        "code_size"
                                        "primOpCodeSize" p_o_specs)
 
-                      "--can-fail"
+                      "--is-work-free"
                          -> putStr (gen_switch_from_attribs
-                                       "can_fail"
-                                       "primOpCanFail" p_o_specs)
+                                       "work_free"
+                                       "primOpIsWorkFree" p_o_specs)
+
+                      "--is-cheap"
+                         -> putStr (gen_switch_from_attribs
+                                       "cheap"
+                                       "primOpIsCheap" p_o_specs)
 
                       "--strictness"
                          -> putStr (gen_switch_from_attribs
@@ -161,6 +166,11 @@ main = getArgs >>= \args ->
                                        "fixity"
                                        "primOpFixity" p_o_specs)
 
+                      "--primop-effects"
+                         -> putStr (gen_switch_from_attribs
+                                       "effect"
+                                       "primOpEffect" p_o_specs)
+
                       "--primop-primop-info"
                          -> putStr (gen_primop_info p_o_specs)
 
@@ -191,19 +201,23 @@ main = getArgs >>= \args ->
                       "--wired-in-docs"
                          -> putStr (gen_wired_in_docs p_o_specs)
 
+                      "--wired-in-deprecations"
+                         -> putStr (gen_wired_in_deprecations p_o_specs)
+
                       _ -> error "Should not happen, known_args out of sync?"
                    )
 
 known_args :: [String]
 known_args
    = [ "--data-decl",
-       "--has-side-effects",
        "--out-of-line",
        "--commutable",
        "--code-size",
-       "--can-fail",
+       "--is-work-free",
+       "--is-cheap",
        "--strictness",
        "--fixity",
+       "--primop-effects",
        "--primop-primop-info",
        "--primop-tag",
        "--primop-list",
@@ -214,7 +228,8 @@ known_args
        "--make-haskell-wrappers",
        "--make-haskell-source",
        "--make-latex-doc",
-       "--wired-in-docs"
+       "--wired-in-docs",
+       "--wired-in-deprecations"
      ]
 
 ------------------------------------------------------------------
@@ -287,7 +302,9 @@ gen_hs_source (Info defaults entries) =
            opt (OptionString n v) = n ++ " = { " ++ v ++ "}"
            opt (OptionInteger n v) = n ++ " = " ++ show v
            opt (OptionVector _)    = ""
-           opt (OptionFixity mf) = "fixity" ++ " = " ++ show mf
+           opt (OptionFixity mf) = "fixity = " ++ show mf
+           opt (OptionEffect eff) = "effect = " ++ show eff
+           opt (OptionCanFailWarnFlag wf) = "can_fail_warning = " ++ show wf
 
            hdr s@(Section {})                                    = sec s
            hdr (PrimOpSpec { name = n })                         = wrapOp n ++ ","
@@ -330,18 +347,16 @@ gen_hs_source (Info defaults entries) =
                  Section { } -> error "Section is not an entity"
              ]
 
-           extra options = case on_llvm_only options ++ can_fail options of
-             [m1,m2] -> "\n\n__/Warning:/__ this " ++ m1 ++ " and " ++ m2 ++ "."
+           extra options = case can_fail options of
              [m] -> "\n\n__/Warning:/__ this " ++ m ++ "."
              _ -> ""
 
-           on_llvm_only options
-             = [ "is only available on LLVM"
-               | Just (OptionTrue _) <- [lookup_attrib "llvm_only" options] ]
-
            can_fail options
              = [ "can fail with an unchecked exception"
-               | Just (OptionTrue _) <- [lookup_attrib "can_fail" options] ]
+               | Just (OptionEffect eff) <- [lookup_attrib "effect" options]
+               , Just (OptionCanFailWarnFlag wflag) <- [lookup_attrib "can_fail_warning" options]
+               , wflag /= DoNotWarnCanFail
+               , wflag == YesWarnCanFail || eff == CanFail ]
 
            prim_deprecated options n
               = [ "{-# DEPRECATED " ++ wrapOp n ++ " \"" ++ msg ++ "\" #-}"
@@ -350,7 +365,7 @@ gen_hs_source (Info defaults entries) =
 
            prim_fixity options n
               = [ pprFixityDir d ++ " " ++ show i ++ " " ++ asInfix n
-                | OptionFixity (Just (Fixity _ i d)) <- options ]
+                | OptionFixity (Just (Fixity i d)) <- options ]
 
            prim_func n t = [ wrapOp n ++ " :: " ++ pprTy t,
                              wrapOp n ++ " = " ++ funcRhs n ]
@@ -457,39 +472,45 @@ gen_wrappers (Info _ entries)
      ++ "import qualified GHC.Prim\n"
      ++ "import GHC.Tuple ()\n"
      ++ "import GHC.Prim (" ++ types ++ ")\n"
-     ++ unlines (concatMap f specs)
+     ++ unlines (concatMap mk_wrapper wrappers)
      where
-        specs = filter (not.dodgy) $
-                filter (not.is_llvm_only) $
-                filter is_primop entries
-        tycons = foldr union [] $ map (tyconsIn . ty) specs
+        wrappers = filter want_wrapper entries
+        tycons = foldr union [] $ map (tyconsIn . ty) wrappers
         tycons' = filter (`notElem` [TyCon "()", TyCon "Bool"]) tycons
         types = concat $ intersperse ", " $ map show tycons'
-        f spec = let args = map (\n -> "a" ++ show n) [1 .. arity (ty spec)]
-                     src_name = wrap (name spec)
-                     lhs = src_name ++ " " ++ unwords args
-                     rhs = wrapQual (name spec) ++ " " ++ unwords args
-                 in ["{-# NOINLINE " ++ src_name ++ " #-}",
-                     src_name ++ " :: " ++ pprTy (ty spec),
-                     lhs ++ " = " ++ rhs]
+        mk_wrapper spec =
+          let args = map (\n -> "a" ++ show n) [1 .. arity (ty spec)]
+              src_name = wrap (name spec)
+              lhs = src_name ++ " " ++ unwords args
+              rhs = wrap_qual (name spec) ++ " " ++ unwords args
+          in ["{-# NOINLINE " ++ src_name ++ " #-}",
+              src_name ++ " :: " ++ pprTy (ty spec),
+              lhs ++ " = " ++ rhs]
         wrap nm | isLower (head nm) = nm
                 | otherwise = "(" ++ nm ++ ")"
-        wrapQual nm | isLower (head nm) = "GHC.Prim." ++ nm
-                    | otherwise         = "(GHC.Prim." ++ nm ++ ")"
+        wrap_qual nm | isLower (head nm) = "GHC.Prim." ++ nm
+                     | otherwise         = "(GHC.Prim." ++ nm ++ ")"
 
-        dodgy spec
-           = name spec `elem`
-             [-- tagToEnum# is really magical, and can't have
+        want_wrapper :: Entry -> Bool
+        want_wrapper entry =
+          and
+            [ is_primop entry
+            , not $ name entry `elem` magical_primops
+            , not $ is_vector entry
+                -- We currently don't generate wrappers for vector primops.
+                --
+                -- SIMD NCG TODO: this was the logic in place when SIMD primops
+                -- were LLVM only; but now that this is no longer the case I
+                -- suppose this choice can be revisited?
+            ]
+
+        magical_primops :: [String]
+        magical_primops =
+          [ "tagToEnum#"
+              -- tagToEnum# is really magical, and can't have
               -- a wrapper since its implementation depends on
               -- the type of its result
-              "tagToEnum#"
-             ]
-
-        is_llvm_only :: Entry -> Bool
-        is_llvm_only entry =
-            case lookup_attrib "llvm_only" (opts entry) of
-              Just (OptionTrue _) -> True
-              _                   -> False
+          ]
 
 gen_primop_list :: Info -> String
 gen_primop_list (Info _ entries)
@@ -608,6 +629,8 @@ gen_switch_from_attribs attrib_name fn_name (Info defaults entries)
          getAltRhs (OptionString _ s) = s
          getAltRhs (OptionVector _) = "True"
          getAltRhs (OptionFixity mf) = show mf
+         getAltRhs (OptionEffect eff) = show eff
+         getAltRhs (OptionCanFailWarnFlag wf) = show wf
 
          mkAlt po
             = case lookup_attrib attrib_name (opts po) of
@@ -621,32 +644,39 @@ gen_switch_from_attribs attrib_name fn_name (Info defaults entries)
             Nothing -> error ("gen_switch_from: " ++ attrib_name)
             Just xx
                -> unlines alternatives
-                  ++ fn_name ++ " _ = " ++ getAltRhs xx ++ "\n"
+                  ++ fn_name ++ " _thisOp = " ++ getAltRhs xx ++ "\n"
 
-{-
-Note [GHC.Prim Docs]
-~~~~~~~~~~~~~~~~~~~~
-For haddocks of GHC.Prim we generate a dummy haskell file (gen_hs_source) that
-contains the type signatures and the commends (but no implementations)
-specifically for consumption by haddock.
-
-GHCi's :doc command reads directly from ModIface's though, and GHC.Prim has a
-wired-in iface that has nothing to do with the above haskell file. The code
-below converts primops.txt into an intermediate form that would later be turned
-into a proper DeclDocMap.
-
-We output the docs as a list of pairs (name, docs). We use stringy names here
-because mapping names to "Name"s is difficult for things like primtypes and
-pseudoops.
--}
+-- See Note [GHC.Prim Docs] in GHC.Builtin.Utils
 gen_wired_in_docs :: Info -> String
 gen_wired_in_docs (Info _ entries)
   = "primOpDocs =\n  [ " ++ intercalate "\n  , " (catMaybes $ map mkDoc $ concatMap desugarVectorSpec entries) ++ "\n  ]\n"
     where
       mkDoc po | Just poName <- getName po
-               , not $ null $ desc po = Just $ show (poName, desc po)
+               , not $ null $ desc po = Just $ "(fsLit " ++ show poName ++ "," ++ show (desc po) ++ ")"
                | otherwise = Nothing
 
+-- See Note [GHC.Prim Deprecations] in GHC.Builtin.Utils
+gen_wired_in_deprecations :: Info -> String
+gen_wired_in_deprecations (Info _ entries)
+  = "primOpDeprecations =\n  [ "
+    ++ intercalate "\n  , " (catMaybes $ map mkDep $ concatMap desugarVectorSpec entries)
+    ++ "\n  ]\n"
+    where
+      mkDep po
+        | Just poName <- getName po
+        , Just (OptionString _ depMsg) <- lookup_attrib "deprecated_msg" (opts po)
+        = let mkOcc =
+                case po of
+                  PrimOpSpec{}      -> "mkVarOcc"
+                  PrimVecOpSpec{}   -> "mkVarOcc"
+                  PseudoOpSpec{}    -> "mkVarOcc"
+                  PrimTypeSpec{}    -> "mkTcOcc"
+                  PrimVecTypeSpec{} -> "mkTcOcc"
+                  Section{}         -> error "impossible(Section)"
+          in Just $ "(" ++ mkOcc ++ " " ++ show poName ++ ", fsLit " ++ show depMsg ++ ")"
+        | otherwise = Nothing
+
+
 ------------------------------------------------------------------
 -- Create PrimOpInfo text from PrimOpSpecs -----------------------
 ------------------------------------------------------------------
@@ -712,7 +742,9 @@ nonDepTyVarBinder bndr
 -- This assumes that such a re-ordering makes sense: the kinds of the inferred
 -- type variables may not depend on any of the other type variables.
 ppTyVarBinders :: [TyVar] -> ([TyVarBinder], [TyVarBinder])
-ppTyVarBinders names = case go names of { (infs, bndrs) -> (nub infs, nub bndrs) }
+ppTyVarBinders names =
+  case go names of
+    { (infs, bndrs) -> (nub infs, nub bndrs) }
   where
      go [] = ([], [])
      go (tv:tvs)
@@ -728,21 +760,24 @@ ppTyVar "a" = nonDepTyVarBinder "alphaTyVarSpec"
 ppTyVar "b" = nonDepTyVarBinder "betaTyVarSpec"
 ppTyVar "c" = nonDepTyVarBinder "gammaTyVarSpec"
 ppTyVar "s" = nonDepTyVarBinder "deltaTyVarSpec"
-ppTyVar "o" = PrimOpTyVarBinder
-              { inferredTyVarBinders = ["runtimeRep1TyVarInf"]
-              , primOpTyVarBinder    = "openAlphaTyVarSpec" }
-ppTyVar "p" = PrimOpTyVarBinder
-              { inferredTyVarBinders = ["runtimeRep2TyVarInf"]
-              , primOpTyVarBinder    = "openBetaTyVarSpec" }
-ppTyVar "v" = PrimOpTyVarBinder
-              { inferredTyVarBinders = ["levity1TyVarInf"]
-              , primOpTyVarBinder    = "levPolyAlphaTyVarSpec" }
-ppTyVar "w" = PrimOpTyVarBinder
-              { inferredTyVarBinders = ["levity2TyVarInf"]
-              , primOpTyVarBinder    = "levPolyBetaTyVarSpec" }
-ppTyVar _   = error "Unknown type var"
--- o, p, v and w have a special meaning. See primops.txt.pp
--- Note [Levity and representation polymorphic primops]
+-- See Note [Levity and representation polymorphic primops] in primops.txt.pp
+ppTyVar "a_reppoly"
+  = PrimOpTyVarBinder
+  { inferredTyVarBinders = ["runtimeRep1TyVarInf"]
+  , primOpTyVarBinder    = "openAlphaTyVarSpec" }
+ppTyVar "b_reppoly"
+  = PrimOpTyVarBinder
+  { inferredTyVarBinders = ["runtimeRep2TyVarInf"]
+  , primOpTyVarBinder    = "openBetaTyVarSpec" }
+ppTyVar "a_levpoly"
+  = PrimOpTyVarBinder
+  { inferredTyVarBinders = ["levity1TyVarInf"]
+  , primOpTyVarBinder    = "levPolyAlphaTyVarSpec" }
+ppTyVar "b_levpoly"
+  = PrimOpTyVarBinder
+  { inferredTyVarBinders = ["levity2TyVarInf"]
+  , primOpTyVarBinder    = "levPolyBetaTyVarSpec" }
+ppTyVar var = error $ "Unknown type variable name '" ++ var ++ "'"
 
 ppType :: Ty -> String
 ppType (TyApp (TyCon "Any")         []) = "anyTy"
@@ -775,12 +810,11 @@ ppType (TyVar "a")                      = "alphaTy"
 ppType (TyVar "b")                      = "betaTy"
 ppType (TyVar "c")                      = "gammaTy"
 ppType (TyVar "s")                      = "deltaTy"
-ppType (TyVar "o")                      = "openAlphaTy"
-ppType (TyVar "p")                      = "openBetaTy"
-ppType (TyVar "v")                      = "levPolyAlphaTy"
-ppType (TyVar "w")                      = "levPolyBetaTy"
--- o, p, v and w have a special meaning. See primops.txt.pp
--- Note [Levity and representation polymorphic primops]
+-- See Note [Levity and representation polymorphic primops] in primops.txt.pp
+ppType (TyVar "a_reppoly")              = "openAlphaTy"
+ppType (TyVar "b_reppoly")              = "openBetaTy"
+ppType (TyVar "a_levpoly")              = "levPolyAlphaTy"
+ppType (TyVar "b_levpoly")              = "levPolyBetaTy"
 
 ppType (TyApp (TyCon "State#") [x])             = "mkStatePrimTy " ++ ppType x
 ppType (TyApp (TyCon "MutVar#") [x,y])          = "mkMutVarPrimTy " ++ ppType x
