@@ -32,6 +32,7 @@ import GHC.Toolchain.Tools.Nm
 import GHC.Toolchain.Tools.MergeObjs
 import GHC.Toolchain.Tools.Readelf
 import GHC.Toolchain.NormaliseTriple (normaliseTriple)
+import Text.Read (readMaybe)
 
 data Opts = Opts
     { optTriple    :: String
@@ -43,6 +44,8 @@ data Opts = Opts
     , optCxx       :: ProgOpt
     , optCpp       :: ProgOpt
     , optHsCpp     :: ProgOpt
+    , optJsCpp     :: ProgOpt
+    , optCmmCpp    :: ProgOpt
     , optCcLink    :: ProgOpt
     , optAr        :: ProgOpt
     , optRanlib    :: ProgOpt
@@ -88,6 +91,8 @@ emptyOpts = Opts
     , optCxx       = po0
     , optCpp       = po0
     , optHsCpp     = po0
+    , optJsCpp     = po0
+    , optCmmCpp    = po0
     , optCcLink    = po0
     , optAr        = po0
     , optRanlib    = po0
@@ -106,13 +111,15 @@ emptyOpts = Opts
   where
     po0 = emptyProgOpt
 
-_optCc, _optCxx, _optCpp, _optHsCpp, _optCcLink, _optAr, _optRanlib, _optNm,
-    _optReadelf, _optMergeObjs, _optWindres, _optLd
+_optCc, _optCxx, _optCpp, _optHsCpp, _optJsCpp, _optCmmCpp, _optCcLink, _optAr,
+    _optRanlib, _optNm, _optReadelf, _optMergeObjs, _optWindres, _optLd
     :: Lens Opts ProgOpt
 _optCc      = Lens optCc      (\x o -> o {optCc=x})
 _optCxx     = Lens optCxx     (\x o -> o {optCxx=x})
 _optCpp     = Lens optCpp     (\x o -> o {optCpp=x})
 _optHsCpp   = Lens optHsCpp   (\x o -> o {optHsCpp=x})
+_optJsCpp   = Lens optJsCpp   (\x o -> o {optJsCpp=x})
+_optCmmCpp  = Lens optCmmCpp   (\x o -> o {optCmmCpp=x})
 _optCcLink  = Lens optCcLink  (\x o -> o {optCcLink=x})
 _optAr      = Lens optAr      (\x o -> o {optAr=x})
 _optRanlib  = Lens optRanlib  (\x o -> o {optRanlib=x})
@@ -167,6 +174,8 @@ options =
     [ progOpts "cc" "C compiler" _optCc
     , progOpts "cpp" "C preprocessor" _optCpp
     , progOpts "hs-cpp" "Haskell C preprocessor" _optHsCpp
+    , progOpts "js-cpp" "JavaScript C preprocessor" _optJsCpp
+    , progOpts "cmm-cpp" "C-- C preprocessor" _optCmmCpp
     , progOpts "cxx" "C++ compiler" _optCxx
     , progOpts "cc-link" "C compiler for linking" _optCcLink
     , progOpts "ar" "ar archiver" _optAr
@@ -249,9 +258,12 @@ doFormat args = do
   case errs of
     [] -> do
       let opts = foldr (.) id opts0 emptyFormatOpts
-      tgt <- read @Target <$> System.IO.readFile (view _formatOptInput opts)
-      let file = formatOptOutput opts
-      System.IO.writeFile file (show tgt)
+      tgtFile <- System.IO.readFile (view _formatOptInput opts)
+      case readMaybe @Target tgtFile of
+        Nothing -> error $ "Failed to read a valid Target value from " ++ view _formatOptInput opts ++ ":\n" ++ tgtFile
+        Just tgt -> do
+          let file = formatOptOutput opts
+          System.IO.writeFile file (show tgt)
     _ -> do
       mapM_ putStrLn errs
       putStrLn $ usageInfo "ghc-toolchain" formatOpts
@@ -362,6 +374,7 @@ archHasNativeAdjustors :: Arch -> Bool
 archHasNativeAdjustors = \case
   ArchX86    -> True
   ArchX86_64 -> True
+  ArchJavaScript -> True
   _          -> False
 
 
@@ -389,6 +402,10 @@ mkTarget opts = do
     cxx <- findCxx archOs tgtLlvmTarget (optCxx opts)
     cpp <- findCpp (optCpp opts) cc0
     hsCpp <- findHsCpp (optHsCpp opts) cc0
+    -- TODO: same case as ranlib below
+    -- TODO: we need it really only for javascript target (maybe wasm target as well)
+    jsCpp <- Just <$> findJsCpp (optJsCpp opts) cc0
+    cmmCpp <- findCmmCpp (optCmmCpp opts) cc0
     cc <- addPlatformDepCcFlags archOs cc0
     readelf <- optional $ findReadelf (optReadelf opts)
     ccLink <- findCcLink tgtLlvmTarget (optLd opts) (optCcLink opts) (ldOverrideWhitelist archOs && fromMaybe True (optLdOverride opts)) archOs cc readelf
@@ -444,6 +461,8 @@ mkTarget opts = do
                    , tgtCxxCompiler = cxx
                    , tgtCPreprocessor = cpp
                    , tgtHsCPreprocessor = hsCpp
+                   , tgtJsCPreprocessor = jsCpp
+                   , tgtCmmCPreprocessor = cmmCpp
                    , tgtAr = ar
                    , tgtCCompilerLink = ccLink
                    , tgtRanlib = ranlib
