@@ -5,7 +5,6 @@
 {-# LANGUAGE OverloadedStrings   #-}
 {-# LANGUAGE Rank2Types          #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-# OPTIONS_GHC -Wwarn           #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Haddock
@@ -30,6 +29,45 @@ module Haddock (
   withGhc
 ) where
 
+import Control.DeepSeq (force)
+import Control.Monad hiding (forM_)
+import Control.Monad.IO.Class (MonadIO(..))
+import Data.Bifunctor (second)
+import Data.Foldable (forM_)
+import Data.Traversable (for)
+import qualified Data.List as List
+import Control.Exception
+import Data.Maybe
+import Data.IORef
+import Data.Map.Strict (Map)
+import Data.Version (makeVersion)
+import GHC.Parser.Lexer (ParserOpts)
+import qualified GHC.Driver.Config.Parser as Parser
+import qualified Data.Map.Strict as Map
+import System.IO
+import System.Exit
+import System.FilePath
+#ifdef IN_GHC_TREE
+import System.Environment (getExecutablePath)
+#else
+import qualified GHC.Paths as GhcPaths
+import Paths_haddock_api (getDataDir)
+#endif
+import System.Directory (doesDirectoryExist, getTemporaryDirectory)
+import Text.ParserCombinators.ReadP (readP_to_S)
+import GHC hiding (verbosity)
+import GHC.Settings.Config
+import GHC.Driver.Config.Logger (initLogFlags)
+import GHC.Driver.Env
+import GHC.Driver.Session hiding (projectVersion, verbosity)
+import qualified GHC.Driver.Session as DynFlags (DynFlags(..))
+import GHC.Utils.Error
+import GHC.Utils.Logger
+import GHC.Types.Name.Cache
+import GHC.Unit
+import GHC.Utils.Panic (handleGhcException)
+import GHC.Data.FastString
+
 import Haddock.Backends.Xhtml
 import Haddock.Backends.Xhtml.Meta
 import Haddock.Backends.Xhtml.Themes (getThemes)
@@ -44,46 +82,8 @@ import Haddock.Version
 import Haddock.InterfaceFile
 import Haddock.Options
 import Haddock.Utils
-import Haddock.GhcUtils (modifySessionDynFlags, setOutputDir)
-
-import Control.DeepSeq (force)
-import Control.Monad hiding (forM_)
-import Control.Monad.IO.Class (MonadIO(..))
-import Data.Bifunctor (second)
-import Data.Foldable (forM_, foldl')
-import Data.Traversable (for)
-import Data.List (find, isPrefixOf, nub)
-import Control.Exception
-import Data.Maybe
-import Data.IORef
-import Data.Map.Strict (Map)
-import Data.Version (makeVersion)
-import qualified Data.Map.Strict as Map
-import System.IO
-import System.Exit
-import System.FilePath
-
-#ifdef IN_GHC_TREE
-import System.Environment (getExecutablePath)
-#else
-import qualified GHC.Paths as GhcPaths
-import Paths_haddock_api (getDataDir)
-#endif
-import System.Directory (doesDirectoryExist, getTemporaryDirectory)
-
-import Text.ParserCombinators.ReadP (readP_to_S)
-import GHC hiding (verbosity)
-import GHC.Settings.Config
-import GHC.Driver.Config.Logger (initLogFlags)
-import GHC.Driver.Env
-import GHC.Driver.Session hiding (projectVersion, verbosity)
-import qualified GHC.Driver.Session as DynFlags (DynFlags(..))
-import GHC.Utils.Error
-import GHC.Utils.Logger
-import GHC.Types.Name.Cache
-import GHC.Unit
-import GHC.Utils.Panic (handleGhcException)
-import GHC.Data.FastString
+import Haddock.GhcUtils (modifySessionDynFlags, setOutputDir, getSupportedLanguagesAndExtensions)
+import Haddock.Compat (getProcessID)
 
 --------------------------------------------------------------------------------
 -- * Exception handling
@@ -167,12 +167,14 @@ haddockWithGhc ghc args = handleTopExceptions $ do
   qual <- rightOrThrowE (qualification flags)
   sinceQual <- rightOrThrowE (sinceQualification flags)
 
+  let noCompilation = isJust (optOneShot flags) || Flag_NoCompilation `elem` flags
+
   -- Inject dynamic-too into ghc options if the ghc we are using was built with
-  -- dynamic linking
+  -- dynamic linking (except when not doing any compilation)
   flags'' <- ghc flags $ do
         df <- getDynFlags
         case lookup "GHC Dynamic" (compilerInfo df) of
-          Just "YES" -> return $ Flag_OptGhc "-dynamic-too" : flags
+          Just "YES" | not noCompilation -> return $ Flag_OptGhc "-dynamic-too" : flags
           _ -> return flags
 
   -- Inject `-j` into ghc options, if given to Haddock
@@ -190,7 +192,7 @@ haddockWithGhc ghc args = handleTopExceptions $ do
   -- Output dir needs to be set before calling 'depanal' since 'depanal' uses it
   -- to compute output file names that are stored in the 'DynFlags' of the
   -- resulting 'ModSummary's.
-  let withDir | Flag_NoTmpCompDir `elem` flags = id
+  let withDir | Flag_NoTmpCompDir `elem` flags || noCompilation = id
               | otherwise = withTempOutputDir
 
   -- Output warnings about potential misuse of some flags
@@ -201,8 +203,14 @@ haddockWithGhc ghc args = handleTopExceptions $ do
       hPutStrLn stderr noCheckWarning
 
   ghc flags' $ withDir $ do
-    dflags <- getDynFlags
-    logger <- getLogger
+    dflags' <- getDynFlags
+    let unicode = Flag_UseUnicode `elem` flags
+    let dflags
+          | unicode = gopt_set dflags' Opt_PrintUnicodeSyntax
+          | otherwise = dflags'
+    logger' <- getLogger
+    let logger = setLogFlags logger' (initLogFlags dflags)
+    let parserOpts = Parser.initParserOpts dflags
     !unit_state <- hsc_units <$> getSession
 
     -- If any --show-interface was used, show the given interfaces
@@ -213,13 +221,14 @@ haddockWithGhc ghc args = handleTopExceptions $ do
         putMsg logger $ renderJson (jsonInterfaceFile ifaceFile)
 
     -- If we were given source files to generate documentation from, do it
-    if not (null files) then do
+    if not (null files) || isJust (optOneShot flags) then do
       (packages, ifaces, homeLinks) <- readPackagesAndProcessModules flags files
       let packageInfo = PackageInfo { piPackageName =
                                         fromMaybe (PackageName mempty) (optPackageName flags)
                                     , piPackageVersion =
                                         fromMaybe (makeVersion []) (optPackageVersion flags)
                                     }
+          languagesAndExtensions = getSupportedLanguagesAndExtensions ifaces
 
       -- Dump an "interface file" (.haddock file), if requested.
       forM_ (optDumpInterfaceFile flags) $ \path -> liftIO $ do
@@ -230,7 +239,7 @@ haddockWithGhc ghc args = handleTopExceptions $ do
           }
 
       -- Render the interfaces.
-      liftIO $ renderStep logger dflags unit_state flags sinceQual qual packages ifaces
+      liftIO $ renderStep dflags languagesAndExtensions parserOpts logger unit_state flags sinceQual qual packages ifaces
 
     -- If we were not given any input files, error if documentation was
     -- requested
@@ -243,7 +252,7 @@ haddockWithGhc ghc args = handleTopExceptions $ do
       packages <- liftIO $ readInterfaceFiles name_cache (readIfaceArgs flags) noChecks
 
       -- Render even though there are no input files (usually contents/index).
-      liftIO $ renderStep logger dflags unit_state flags sinceQual qual packages []
+      liftIO $ renderStep dflags [] parserOpts logger unit_state flags sinceQual qual packages []
 
 -- | Run the GHC action using a temporary output directory
 withTempOutputDir :: Ghc a -> Ghc a
@@ -256,7 +265,7 @@ withTempOutputDir action = do
 
 -- | Create warnings about potential misuse of -optghc
 optGhcWarnings :: [String] -> [String]
-optGhcWarnings = map format . filter (isPrefixOf "-optghc")
+optGhcWarnings = map format . filter (List.isPrefixOf "-optghc")
   where
     format arg = concat ["Warning: `", arg, "' means `-o ", drop 2 arg, "', did you mean `-", arg, "'?"]
 
@@ -295,9 +304,20 @@ readPackagesAndProcessModules flags files = do
     return (packages, ifaces, homeLinks)
 
 
-renderStep :: Logger -> DynFlags -> UnitState -> [Flag] -> SinceQual -> QualOption
-           -> [(DocPaths, Visibility, FilePath, InterfaceFile)] -> [Interface] -> IO ()
-renderStep logger dflags unit_state flags sinceQual nameQual pkgs interfaces = do
+renderStep
+  :: DynFlags
+  -> [String]
+  -- ^ Supported languages and extensions based on architecture and OS
+  -> ParserOpts
+  -> Logger
+  -> UnitState
+  -> [Flag]
+  -> SinceQual
+  -> QualOption
+  -> [(DocPaths, Visibility, FilePath, InterfaceFile)]
+  -> [Interface]
+  -> IO ()
+renderStep dflags languagesAndExtensions parserOpts logger unit_state flags sinceQual nameQual pkgs interfaces = do
   updateHTMLXRefs (map (\(docPath, _ifaceFilePath, _showModules, ifaceFile) ->
                           ( case baseUrl flags of
                               Nothing  -> docPathsHtml docPath
@@ -313,7 +333,7 @@ renderStep logger dflags unit_state flags sinceQual nameQual pkgs interfaces = d
       (DocPaths {docPathsSources=Just path}, _, _, ifile) <- pkgs
       iface <- ifInstalledIfaces ifile
       return (instMod iface, path)
-  render logger dflags unit_state flags sinceQual nameQual interfaces installedIfaces extSrcMap
+  render dflags languagesAndExtensions parserOpts logger unit_state flags sinceQual nameQual interfaces installedIfaces extSrcMap
   where
     -- get package name from unit-id
     packageName :: Unit -> String
@@ -323,10 +343,21 @@ renderStep logger dflags unit_state flags sinceQual nameQual pkgs interfaces = d
         Just pkg -> unitPackageNameString pkg
 
 -- | Render the interfaces with whatever backend is specified in the flags.
-render :: Logger -> DynFlags -> UnitState -> [Flag] -> SinceQual -> QualOption -> [Interface]
-       -> [(FilePath, PackageInterfaces)] -> Map Module FilePath -> IO ()
-render log' dflags unit_state flags sinceQual qual ifaces packages extSrcMap = do
-
+render
+  :: DynFlags
+  -> [String]
+  -- ^ Supported languages and extensions based on architecture and OS
+  -> ParserOpts
+  -> Logger
+  -> UnitState
+  -> [Flag]
+  -> SinceQual
+  -> QualOption
+  -> [Interface]
+  -> [(FilePath, PackageInterfaces)]
+  -> Map Module FilePath
+  -> IO ()
+render dflags languagesAndExtensions parserOpts logger unit_state flags sinceQual qual ifaces packages extSrcMap = do
   let
     packageInfo = PackageInfo { piPackageName    = fromMaybe (PackageName mempty)
                                                  $ optPackageName flags
@@ -345,10 +376,6 @@ render log' dflags unit_state flags sinceQual qual ifaces packages extSrcMap = d
     opt_latex_style      = optLaTeXStyle     flags
     opt_source_css       = optSourceCssFile  flags
     opt_mathjax          = optMathjax        flags
-    dflags'
-      | unicode          = gopt_set dflags Opt_PrintUnicodeSyntax
-      | otherwise        = dflags
-    logger               = setLogFlags log' (initLogFlags dflags')
 
     visibleIfaces    = [ i | i <- ifaces, OptHide `notElem` ifaceOptions i ]
 
@@ -394,11 +421,14 @@ render log' dflags unit_state flags sinceQual qual ifaces packages extSrcMap = d
       | Flag_HyperlinkedSource `elem` flags = Just hypSrcModuleUrlFormat
       | otherwise = srcModule
 
+    -- These urls have a template for the module %M
     srcMap = Map.union
-      (Map.map SrcExternal extSrcMap)
+      (Map.map (SrcExternal . hypSrcPkgUrlToModuleFormat) extSrcMap)
       (Map.fromList [ (ifaceMod iface, SrcLocal) | iface <- ifaces ])
 
-    pkgSrcMap = Map.mapKeys moduleUnit extSrcMap
+    -- These urls have a template for the module %M and the name %N
+    pkgSrcMap = Map.map (hypSrcModuleUrlToNameFormat . hypSrcPkgUrlToModuleFormat)
+              $ Map.mapKeys moduleUnit extSrcMap
     pkgSrcMap'
       | Flag_HyperlinkedSource `elem` flags
       , Just k <- pkgKey
@@ -408,15 +438,17 @@ render log' dflags unit_state flags sinceQual qual ifaces packages extSrcMap = d
       = Map.insert k srcNameUrl pkgSrcMap
       | otherwise = pkgSrcMap
 
-    -- TODO: Get these from the interface files as with srcMap
+    pkgSrcLMap = Map.map (hypSrcModuleUrlToLineFormat . hypSrcPkgUrlToModuleFormat)
+               $ Map.mapKeys moduleUnit extSrcMap
+    -- These urls have a template for the module %M and the line %L
     pkgSrcLMap'
       | Flag_HyperlinkedSource `elem` flags
       , Just k <- pkgKey
-      = Map.singleton k hypSrcModuleLineUrlFormat
+      = Map.insert k hypSrcModuleLineUrlFormat pkgSrcLMap
       | Just path <- srcLEntity
       , Just k <- pkgKey
-      = Map.singleton k path
-      | otherwise = Map.empty
+      = Map.insert k path pkgSrcLMap
+      | otherwise = pkgSrcLMap
 
     sourceUrls' = (srcBase, srcModule', pkgSrcMap', pkgSrcLMap')
 
@@ -440,12 +472,12 @@ render log' dflags unit_state flags sinceQual qual ifaces packages extSrcMap = d
       _ -> warn' ("Cannot parse reexported module flag '" ++ mod_str ++ "'") >> return [])
 
   libDir   <- getHaddockLibDir flags
-  !prologue <- force <$> getPrologue dflags' flags
+  !prologue <- force <$> getPrologue parserOpts flags
   themes   <- getThemes libDir flags >>= either bye return
 
   let withQuickjump = Flag_QuickJumpIndex `elem` flags
       withBaseURL = isJust
-                  . find (\flag -> case flag of
+                  . List.find (\flag -> case flag of
                            Flag_BaseURL base_url ->
                              base_url /= "." && base_url /= "./"
                            _ -> False
@@ -457,7 +489,6 @@ render log' dflags unit_state flags sinceQual qual ifaces packages extSrcMap = d
       _ <- {-# SCC ppHtmlIndex #-}
            ppHtmlIndex odir title pkgStr
                   themes opt_mathjax opt_contents_url sourceUrls' opt_wiki_urls
-                  withQuickjump
                   (concatMap piInstalledInterfaces allVisiblePackages) pretty
       return ()
 
@@ -469,7 +500,6 @@ render log' dflags unit_state flags sinceQual qual ifaces packages extSrcMap = d
       _ <- {-# SCC ppHtmlContents #-}
            ppHtmlContents unit_state odir title pkgStr
                      themes opt_mathjax opt_index_url sourceUrls' opt_wiki_urls
-                     withQuickjump
                      allVisiblePackages True prologue pretty
                      sincePkg (makeContentsQual qual)
       return ()
@@ -479,7 +509,7 @@ render log' dflags unit_state flags sinceQual qual ifaces packages extSrcMap = d
             ppJsonIndex odir sourceUrls' opt_wiki_urls
                         unicode Nothing qual
                         ifaces
-                        ( nub
+                        ( List.nub
                         . map fst
                         . filter ((== Visible) . piVisibility . snd)
                         $ packages)
@@ -493,7 +523,7 @@ render log' dflags unit_state flags sinceQual qual ifaces packages extSrcMap = d
                   opt_contents_url opt_index_url unicode sincePkg packageInfo
                   qual pretty withQuickjump
       return ()
-    unless withBaseURL $ do
+    unless (withBaseURL || isJust (optOneShot flags)) $ do
       copyHtmlBits odir libDir themes withQuickjump
       writeHaddockMeta odir withQuickjump
 
@@ -508,7 +538,7 @@ render log' dflags unit_state flags sinceQual qual ifaces packages extSrcMap = d
 
             pkgVer =
               fromMaybe (makeVersion []) mpkgVer
-          in ppHoogle dflags' unit_state pkgNameStr pkgVer title (fmap _doc prologue)
+          in ppHoogle dflags unit_state pkgNameStr pkgVer title (fmap _doc prologue)
                visibleIfaces odir
       _ -> putStrLn . unlines $
           [ "haddock: Unable to find a package providing module "
@@ -530,7 +560,7 @@ render log' dflags unit_state flags sinceQual qual ifaces packages extSrcMap = d
   when (Flag_HyperlinkedSource `elem` flags && not (null ifaces)) $ do
     withTiming logger "ppHyperlinkedSource" (const ()) $ do
       _ <- {-# SCC ppHyperlinkedSource #-}
-           ppHyperlinkedSource (verbosity flags) odir libDir opt_source_css pretty srcMap ifaces
+           ppHyperlinkedSource (verbosity flags) (isJust (optOneShot flags)) languagesAndExtensions  odir libDir opt_source_css pretty srcMap ifaces
       return ()
 
 
@@ -610,7 +640,7 @@ withGhc' libDir needHieFiles flags ghcActs = runGhc (Just libDir) $ do
               ++ if needHieFiles
                     then [Opt_WriteHie] -- Generate .hie-files
                     else []
-          dynflags' = (foldl' gopt_set dynflags extra_opts)
+          dynflags' = (List.foldl' gopt_set dynflags extra_opts)
                         { backend = noBackend
                         , ghcMode = CompManager
                         , ghcLink = NoLink
@@ -624,7 +654,7 @@ withGhc' libDir needHieFiles flags ghcActs = runGhc (Just libDir) $ do
 
 unsetPatternMatchWarnings :: DynFlags -> DynFlags
 unsetPatternMatchWarnings dflags =
-  foldl' wopt_unset dflags pattern_match_warnings
+  List.foldl' wopt_unset dflags pattern_match_warnings
   where
     pattern_match_warnings =
       [ Opt_WarnIncompletePatterns
@@ -802,15 +832,15 @@ updateHTMLXRefs packages = do
     mapping' = [ (moduleName m, html) | (m, html) <- mapping ]
 
 
-getPrologue :: DynFlags -> [Flag] -> IO (Maybe (MDoc RdrName))
-getPrologue dflags flags =
+getPrologue :: ParserOpts -> [Flag] -> IO (Maybe (MDoc RdrName))
+getPrologue parserOpts flags =
   case [filename | Flag_Prologue filename <- flags ] of
     [] -> return Nothing
     [filename] -> do
       h <- openFile filename ReadMode
       hSetEncoding h utf8
       str <- hGetContents h -- semi-closes the handle
-      return . Just $! second (fmap rdrName) $ parseParas dflags Nothing str
+      return . Just $! second (fmap rdrName) $ parseParas parserOpts Nothing str
     _ -> throwE "multiple -p/--prologue options"
 
 
