@@ -13,18 +13,18 @@
 #define ASSERT(predicate) /*nothing*/
 
 #else
-void _ghcAssert PROTO((char *, unsigned int));
+void _stgAssert PROTO((char *, unsigned int));
 
 #define ASSERT(predicate)			\
 	if (predicate)				\
 	    /*null*/;				\
 	else					\
-	    _ghcAssert(__FILE__, __LINE__)
+	    _stgAssert(__FILE__, __LINE__)
 #endif
 
 /* partain: some ubiquitous types: floatish & intish.
    Dubious to use float/int, but that is what it used to be...
-   (WDP 95/03)   
+   (WDP 95/03)
 */
 typedef double	floatish;
 typedef double  doublish; /* higher precision, if anything; little used */
